% CONSTRAINT REPLACEMENT for EXISTENTIAL VARIABLES ELIMINATION
%
% we assume that we are required to eliminate ONE existential variable at a time
/*
Rewrite c as  (d, e)  where: e is a conjunction	of equalities
and \not\models d -> r1=r2 for any r1,r2 occurring in c.	

Apply the following rewritings until there is at most one
Z=<a (or Z<a) and at most one a=<Z (or a<Z):

1) 
Z=<a, Z=<b, d --> (Z=<a, a=<b, d) or (Z=<b, b<a,  d)
Z=<a, Z<b,  d --> (Z=<a, a<b,  d) or (Z<b,  b=<a, d)
Z<a,  Z<b,  d --> (Z<a,  a<b,  d) or (Z<b,  b=<a, d)
a=<Z, b=<Z, d --> (a=<Z, b=<a, d) or (b=<Z, a<b,  d)
a=<Z, b<Z,  d --> (a=<Z, b<a,  d) or (b<Z,  a=<b, d)
a<Z,  b<Z,  d --> (a<Z,  b<a,  d) or (b<Z,  a=<b, d)

2) 
Delete inconsistent constraints.

3) 
d1, d2 --> d1  if  \forall(d1 -> d2)  

where d2 is a constraint such that Z \in vars(d2)
(that is, d2 is of the form:  Z=<a  or  Z<a  or  a=<Z  or  a<Z)

Note that 2 and 3 are optional.

Termination: At each replacement an occurrence of Z is deleted.
(When there are many existential variables, termination is more 
involved. Multiset ordering is needed.) */

% MODE: constraint_replacement(+Def,-NewDef)
% TYPE: constraint_replacement(def,def)
% SEMANTICS: we are given a definition made of a SET of clauses with
%   some (>1) existential variables. Suppose Z is one of these existential
%   variables, then each (some??) clause will be
%   of the form H :- C(Z), B(Z) and we want to abtain a new set of clauses
%   of the form H :- D, C2(Z), B(Z) where the constraint D does not contain
%   the existential variable Z and C2(Z) is of the form Wi{=<,<}Z{=<,<}Ui
constraint_replacement((P,[]),(P,[])).
constraint_replacement((P,[C|Cs]),(P,NewClauses)) :-
	constraint_replace_one_clause(C,NewCs),
	append(NewCs,Clauses,NewClauses),
	constraint_replacement((P,Cs),(P,Clauses)).

% MODE: constraint_replace_one_clause(+Clause,-NewCs)
% TYPE: constraint_replace_one_clause(clause,list(clause))
% SEMANTICS: we are given ONE clause
%   of the form H :- C(Z), B(Z) and we want to abtain a new clause
%   of the form H :- D, C2(Z), B(Z) where the constraint D does not contain
%   the existential variable Z and C2(Z) is of the form Wi{=<,<}Z{=<,<}Ui
constraint_replace_one_clause((H,bd(C,[])),[(H,bd(NewC,[],[]))]) :-
	term_variables(H,Vars),
	solve(clpr,C,Vars,NewC).
constraint_replace_one_clause((H,bd(C,B)),NewCs) :-
	setof(X,existential_variable((H,bd(C,B)),X),ExVars),
	separe_equalities(C,CnoEq,Eq),
	bind_equal_variables(Eq), 	% ????		% eliminate exist.vars by binding
	rewrite_constraint([CnoEq],ExVars,NewClist),	% eliminate exist.vars by constraint-rewriting
	generate_clauses((H,bd(C,B)),NewClist,ExVars,NewCs).

%
% constraint_replace_one_clause((p(A,[X|Xs]),bd([A=<M,X=<M],[q(Xs,M)])),NewCs).
%NewCs = [(p(A,[X|Xs]),bd([X=<A],[M>=A],[q(Xs,M)])),(p(A,[X|Xs]),bd([A<X],[M>=X],[q(Xs,M)]))]

% MODE: separe_equalities(+C,-CnoEq,-CEq)
% TYPE: separe_equalities(list(constraint),list(constraint),list(constraint))
% SEMANTICS: 
separe_equalities([],[],[]).
separe_equalities([C|Cs],[C|Rest],Eq) :-
	C\=(_=_), C\=(_=:=_),
	separe_equalities(Cs,Rest,Eq).
separe_equalities([C|Cs],Rest,[C|Eq]) :-
	(C=(_=_) ; C=(_=:=_)),
	separe_equalities(Cs,Rest,Eq).

% ???????????????????
% MODE: bind_equal_variables(+List)
% TYPE: bind_equal_variables(list(var))
% SEMANTICS: 
bind_equal_variables([]).
bind_equal_variables([Eq|Eqs]) :-
	(Eq=(A=:=B) -> A=B ; Eq),
	bind_equal_variables(Eqs).

% MODE: rewrite_constraint(+Dnf,+ExVars,-NewDnf)
% TYPE: rewrite_constraint(list(list(constraint)),list(var),list(list(constraint)))
% SEMANTICS:
rewrite_constraint(Dnf,[],Dnf).
rewrite_constraint(Dnf,[ExV|ExVars],NewDnf) :-
	rewrite_wrt_one_variable(Dnf,ExV,NewDnfTmp),
	rewrite_constraint(NewDnfTmp,ExVars,NewDnf).


% MODE: rewrite_wrt_one_variable(+Dnf,+ExVar,-NewDnf)
% TYPE: rewrite_wrt_one_variable(list(list(constraint)),var,list(list(constraint)))
% SEMANTICS:
rewrite_wrt_one_variable([],_,[]).
rewrite_wrt_one_variable([C|Cs],ExV,NewC) :-
	normalize_constraint(C,NomalC),
	rewrite_rule(NomalC,ExV,NewCtmp),			% try to apply one rewrite rule
	append(NewCtmp,Cs,NewCs),
	rewrite_wrt_one_variable(NewCs,ExV,NewC).
rewrite_wrt_one_variable([C|Cs],ExV,[C|NewCs]) :-	% if finished applying rewrite rules
	rewrite_wrt_one_variable(Cs,ExV,NewCs).		% to one conjunct, then go to the next one

normalize_constraint([],[]).
normalize_constraint([C|Cs],[NewC|NewCs]) :-
	normalize_c(C,NewC),
	normalize_constraint(Cs,NewCs).
normalize_constraint([C|Cs],[C|NewCs]) :-
	normalize_constraint(Cs,NewCs).
	
normalize_c(X>(0.0),A>B) :- subsumes((A-B),X).
normalize_c(X<(0.0),A<B) :- subsumes((A-B),X).
normalize_c(X>=(0.0),A>=B) :- subsumes((A-B),X).
normalize_c(X=<(0.0),A=<B) :- subsumes((A-B),X).
normalize_c(X>(-0.0),A>B) :- subsumes((A-B),X).
normalize_c(X<(-0.0),A<B) :- subsumes((A-B),X).
normalize_c(X>=(-0.0),A>=B) :- subsumes((A-B),X).
normalize_c(X=<(-0.0),A=<B) :- subsumes((A-B),X).
%---
normalize_c((0.0)<X,A>B) :- subsumes((A-B),X).
normalize_c((0.0)>X,A<B) :- subsumes((A-B),X).
normalize_c((0.0)=<X,A>=B) :- subsumes((A-B),X).
normalize_c((0.0)>=X,A=<B) :- subsumes((A-B),X).
normalize_c((-0.0)<X,A>B) :- subsumes((A-B),X).
normalize_c((-0.0)>X,A<B) :- subsumes((A-B),X).
normalize_c((-0.0)=<X,A>=B) :- subsumes((A-B),X).
normalize_c((-0.0)>=X,A=<B) :- subsumes((A-B),X).



% ---------------------------------------------------------------------------------------------
% MODE: rewrite_rule(C,ExV,NewC)
% TYPE: rewrite_rule(list(constraint),var,list(list(constraint)))
% SEMANTICS: FAILS if no rule can be applied
%  -- RULE 1 --
rewrite_rule(C,ExV,NewC) :-		% Z=<A, Z=<B, d ---> (Z=<A, A=<B, d) or (Z=<B, B<A, d) 
	del_constraint(ExV,=<,A,C,Cs),
	del_constraint(ExV,=<,B,Cs,D),
	C1=[ExV=<A,A=<B], C2=[ExV=<B,B<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,>=,ExV,C,Cs),
	del_constraint2(B,>=,ExV,Cs,D),
	C1=[ExV=<A,A=<B], C2=[ExV=<B,B<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
%  -- RULE 2 --
rewrite_rule(C,ExV,NewC) :-		% Z=<A, Z<B, d ---> (Z=<A, A<B, d) or (Z<B, B=<A, d)
	del_constraint(ExV,=<,A,C,Cs),
	del_constraint(ExV,<,B,Cs,D),
	C1=[ExV=<A,A<B], C2=[ExV<B,B=<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,>=,ExV,C,Cs),
	del_constraint2(B,>,ExV,Cs,D),
	C1=[ExV=<A,A<B], C2=[ExV<B,B=<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
%  -- RULE 3 --
rewrite_rule(C,ExV,NewC) :-		% Z<A, Z<B, d ---> (Z<A, A=<B, d) or (Z<B, B<A, d)
	del_constraint(ExV,<,A,C,Cs),
	del_constraint(ExV,<,B,Cs,D),
	C1=[ExV<A,A=<B], C2=[ExV<B,B<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,>,ExV,C,Cs),
	del_constraint2(B,>,ExV,Cs,D),
	C1=[ExV<A,A=<B], C2=[ExV<B,B<A],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
%  -- RULE 4 --
rewrite_rule(C,ExV,NewC) :-		% A=<Z, B=<Z, d ---> (A=<Z, B=<A, d) or (B=<Z, A<B, d)
	del_constraint(ExV,>=,A,C,Cs),
	del_constraint(ExV,>=,B,Cs,D),
	C1=[ExV>=A,B=<A], C2=[ExV>=B,A<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,=<,ExV,C,Cs),
	del_constraint2(B,=<,ExV,Cs,D),
	C1=[ExV>=A,B=<A], C2=[ExV>=B,A<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
%  -- RULE 5 --
rewrite_rule(C,ExV,NewC) :-		% A=<Z, B<Z, d ---> (A=<Z, B<A, d) or (B<Z, A=<B, d)
	del_constraint(ExV,>=,A,C,Cs),
	del_constraint(ExV,>,B,Cs,D),
	C1=[ExV>=A,B<A], C2=[ExV>B,A=<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,=<,ExV,C,Cs),
	del_constraint2(B,<,ExV,Cs,D),
	C1=[ExV>=A,B<A], C2=[ExV>B,A=<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
%  -- RULE 6 --
rewrite_rule(C,ExV,NewC) :-		% A<Z, B<Z, d ---> (A<Z, B=<A, d) or (B<Z, A<B, d)
	del_constraint(ExV,>,A,C,Cs),
	del_constraint(ExV,>,B,Cs,D),
	C1=[ExV>A,B=<A], C2=[ExV>B,A<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
rewrite_rule(C,ExV,NewC) :-		% simmetry
	del_constraint2(A,<,ExV,C,Cs),
	del_constraint2(B,<,ExV,Cs,D),
	C1=[ExV>A,B=<A], C2=[ExV>B,A<B],
	append(C1,D,NewC1), append(C2,D,NewC2),
	NewC=[NewC1,NewC2].
% ---------------------------------------------------------------------------------------------

% the existential variable occurs as the LHS of the constraint
del_constraint(ExV,Constr,Var,[C|Cs],Cs) :-
	C=..[Constr,A,B],ExV==A,Var=B.
del_constraint(C1,[C|Cs],[C|NewCs]) :- del_constraint(C1,Cs,NewCs).
% % the existential variable occurs as the RHS of the constraint
del_constraint2(Var,Constr,ExV,[C|Cs],Cs) :-
	C=..[Constr,A,B],ExV==B,Var=A.
del_constraint2(C1,[C|Cs],[C|NewCs]) :- del_constraint2(C1,Cs,NewCs).

% MODE: generate_clauses(+Clause,+NewClist,-NewClauses)
% TYPE: 
% SEMANTICS: 
generate_clauses(_,[],_,[]).
generate_clauses((H,bd(_,B)),[NewC|NewCs],ExVars,[(H,bd(CnoEx,CEx,B))|Clauses]) :-
	separe_cons_with_ex_variables(NewC,ExVars,CnoEx,CEx), !,
	generate_clauses((H,bd(_,B)),NewCs,ExVars,Clauses).

% MODE: separe_cons_with_ex_variables(+NewConstraint,+ExVariables,-CnoEx,-CEx)
% TYPE: separe_cons_with_ex_variables(list(constraint),list(var),list(constraint),list(constraint))
% SEMANTICS:
separe_cons_with_ex_variables([],_,[],[]).
separe_cons_with_ex_variables([C|NewCs],ExVars,[C|CnoEx],CEx) :-
	term_variables(C,CVars),
	common_vars2(CVars,ExVars,[]),	% the given constraint does not contain an existential variable
	separe_cons_with_ex_variables(NewCs,ExVars,CnoEx,CEx).
separe_cons_with_ex_variables([C|NewCs],ExVars,CnoEx,[C|CEx]) :-	
	separe_cons_with_ex_variables(NewCs,ExVars,CnoEx,CEx).
%
% separe_cons_with_ex_variables([A>=4,C<B,A>B+1],[C],NoEx,Ex).
% Ex = [C<B],
% NoEx = [A>=4,A>B+1]
%
% separe_cons_with_ex_variables([A>=4,C<B,A>B+1],[B],NoEx,Ex).
% Ex = [C<B,A>B+1],
% NoEx = [A>=4]
%
% separe_cons_with_ex_variables([A>=4,C<B,A>B+1],[A],NoEx,Ex).
% Ex = [A>=4,A>B+1],
% NoEx = [C<B]
%

