% basic_op.pl
%
% This file contains low level predicates used in other files.

:- use_module(library(system), [ directory_files/2 ]).
:- use_module(library(lists), [ append/3, memberchk/2, member/2, permutation/2 ]).
:- use_module(library(terms), [ variant/2]). %, unify_with_occurs_check/2 ]).




same_functor(T,T2) :- functor(T,A,N), functor(T2,A,N).

unifiable(Atom,Atom2) :-
    \+ \+ unify_with_occurs_check(Atom2,Atom).

add_new([],L,L).
add_new([E|L1],L2,L3) :-
   belongs(E,L2), !,
   add_new(L1,L2,L3).
add_new([E|L1],L2,[E|L3]) :-
   add_new(L1,L2,L3).

% delete(L1,L2,L): L is the result of deleting each element which occurs both in L1 in L2. 
% Ex: delete([a,b],[c,a,b,a,c,b],[c,a,c,b]) is true.
delete([],L,L).
delete([X|R],L,L1) :- del(X,L,L2), delete(R,L2,L1).

del(X,[X|L],L).
del(X,[Y|L],[Y|L1]) :- \+(X=Y), del(X,L,L1).

% delete_insert(L1,L2,L3,L4): L4 is the result of replacing in L2 the sublist L1 by L3.
delete_insert( [E|L], P, Ns, P1):-
  delete( L, P, P2), replace_elt_in_list(E, P2, Ns, P1).
replace_elt_in_list(E,[E|P2], Ns, P):-	!, append(Ns,P2,P).
replace_elt_in_list(E,[F|P2], Ns, [F|P]):- replace_elt_in_list(E, P2, Ns, P).

member_rest(X,[X|L],L).
member_rest(X,[Y|L],[Y|L2]) :- member_rest(X,L,L2).

append(L1,L2,L3,R) :- append(L2,L3,Z), append(L1,Z,R).

append([], []).
append([L|Ls], List0) :- append(L, List1, List0), append(Ls, List1).

no_dupplicates([]).
no_dupplicates([_]) :- !.
no_dupplicates([X|L]) :- no_dupplicates(L), \+ belongs(X,L).

remove_dupplicates([],[]).
remove_dupplicates([X|L],[X|L2]) :-
    del_3(X,L,L3),
    remove_dupplicates(L3,L2).

% This sort does not remove duplicates.
sort_list([],[]).
sort_list([A],[A]) :- !.
sort_list([A|L],L2) :- sort_list(L,L3), ord_add_elt_list(L3,A,L2).

% changed on the 13/2/98
%ord_add_elt_list([], Element, [Element]).
%ord_add_elt_list([Head|Tail], Element, Set) :-
%   compare(Order, Head, Element), ord_add_elt_list(Order, Head, Tail, Element, Set).

ord_add_elt_list([], Element, [Element]).
ord_add_elt_list([Head|Tail], Element, Set) :-
   Head=.. [FHead|_],    Element=.. [FElement|_], 
   compare(Order, FHead, FElement), ord_add_elt_list(Order, Head, Tail, Element, Set).
 
ord_add_elt_list(<, Head, Tail, Element, [Head|Set]) :- !,
        ord_add_elt_list(Tail, Element, Set).
ord_add_elt_list(>, Head, Tail, Element, [Element,Head|Tail]) :- !.
ord_add_elt_list(_, Head, Tail, Element, [Head|Set]) :-
        ord_add_elt_list(Tail, Element, Set).
ord_add_elt_list(_, Head, Tail, Element, [Element,Head|Tail]).

% split(L,N,L1,E,L2): N is a positive integer. E is the element having position N
% in L, and L is the concatenation of L1, [E], and L2.

split(L,N,L1,E,L2) :- M is N-1, length1(L1,M), append(L1,[E|L2],L), nonvar(E),!.
split(L,N,L1,E,L2) :- M is N+1, split(L,M,L1,E,L2).

% length1(L,N): N is a not negative integer. L is a list made out of N variables.
length1([],0).
length1([_|L],N) :- N>0, N1 is N-1, length1(L,N1).

% vars_subset(T1,T2): each variable occurring in the term T1 occurs also in the term T2.
vars_subset(T1,T2) :- free_variables(T2,[],[],V1), free_variables(T1,V1,[],[]).

intersection_vars_empty(L1,L2) :- intersection_vars(L1,L2,L3), L3==[].
intersection_vars_non_empty(L1,L2) :- intersection_vars(L1,L2,L3), L3=[_|_].

intersection_vars([],_,[]).
intersection_vars([E|L],L2,[E|L3]) :- belongs(E,L2), !, intersection_vars(L,L2,L3).
intersection_vars([_|L],L2,L3) :- intersection_vars(L,L2,L3).

% belongs(E,L): E is an element of list L.
belongs(X,[Y|_]) :- X==Y, !.
belongs(X,[_|T]) :- belongs(X,T).

% conj_to_list(S,L): S a conjunction of the form (A1,...,An), and L is the list [A1,...,An].
conj_to_list(X,[]) :- X==true, !.
conj_to_list(B,L) :-
   (nonvar(B), functor(B,',',_) -> (B = (B1,B2), L=[B1|H], conj_to_list(B2,H)); L=[B]).

last(Last, [Head|Tail]) :- last_1(Tail, Head, Last).		
last_1([], Last, Last).
last_1([Head|Tail], _, Last) :- last_1(Tail, Head, Last).

last([E],[],E).
last([F|L],[F|L2],E):- last(L,L2,E).

my_last(L1,L2,X) :- last(L1,L3,X),!, my_last(L3,L2,X).
my_last(L,L,_).

% positions_to_elts(L,Pos,L1): the elements of list L1 are all elements of list L, and they
% occur at positions Pos . Ex: positions_to_elts([a,b,c,d],[2,4],[b,d]) is true.
positions_to_elts(_,[],[]).
positions_to_elts(B,[H|T],[A|L]) :- pos1(A,B,H), positions_to_elts(B,T,L).

pos1(X,[X|_],1).
pos1(X,[_|L],M) :-  pos1(X,L,N), M is N+1.

% pos(E,L,N): N is the position of the first occurrence of `E' in the list L. 
% Ex: pos (E,[a,b,E,E],3) is true.
pos(X,[Y|_],1) :- X==Y, !.
pos(X,[_|L],M) :- pos(X,L,N), M is N+1.

% delete_2(L1,L2,L3): each element of list L1 is an element of list L2. L3 is obtained by
% deleting each element of L1 from L2. Ex: delete_2([a,X],[c,a,X,a,c,X],[c,a,c,X]) is true.
delete_2([],L,L).
delete_2([X|R],L,L1) :- del_2(X,L,L2), delete_2(R,L2,L1).
del_2(X,[Y|L],L) :- X==Y, !.
del_2(X,[Y|L],[Y|L1]) :- del_2(X,L,L1).

delete_3([],L,L).
delete_3([X|R],L,L1) :- del_3(X,L,L2), delete_3(R,L2,L1).
del_3(_,[],[]).
del_3(X,[Y|L],L1) :- X==Y, !, del_3(X,L,L1).
del_3(X,[Y|L],[Y|L1]) :- del_3(X,L,L1).

% del_1(E,N,L,L1): L1 is obtained by deleting the element at position N in list L. 
% Ex: del_1(E,3,[E,b,E,c],[E,b,c]) is true.
del_1(X,1,[Y|L],L) :- X==Y.
del_1(X,N,[Y|L],[Y|L1]) :- N > 1, M is N-1, del_1(X,M,L,L1).

% int_var(T,TL,V): V is the list of variables occurring in TL and not occurring in T.
int_var(T,TL,V) :- free_variables(T,[],[],V1), free_variables(TL,V1,[],V).

del_und(X,1,[X|L],L).
del_und(X,N,[Y|L],[Y|L1]) :- del_und(X,N2,L,L1), N is N2 + 1.

conc_position(B2,1,Right,B1):- !, append(Right,B2,B1).
conc_position([E|B2],N,Right,[E|B1]):- N2 is N-1, conc_position(B2,N2,Right,B1).

my_directory_files(Dir,Files) :-
   directory_files(Dir,Files1), sort(Files1,Files2), filter_dot_pl(Files2,Files).
%   directory_files(Dir,Files1), sort(Files1,Files2), remove_saved_versions(Files2,Files).

filter_dot_pl([],[]).
filter_dot_pl([F|L1],L) :-
    name(F,NameF), append(_,[46,112,108],NameF),!, L=[F|L2], filter_dot_pl(L1,L2).
filter_dot_pl([_F|L],L2) :-
    filter_dot_pl(L,L2).


remove_saved_versions([],[]).
remove_saved_versions([F|L],L2) :-
    name(F,NameF), last(126,NameF),!, remove_saved_versions(L,L2).
remove_saved_versions([F|L],[F|L2]) :-
    remove_saved_versions(L,L2).

non_empty_directory([E,'..'|_],[E]) :- !.
non_empty_directory([E|L],[E|LL]) :- non_empty_directory(L,LL).

flatten([],[]).
flatten([[]|L], L2) :- !, flatten(L,L2).
flatten([[E|L]|L2], L3) :-
  !, flatten([E|L],EL), flatten(L2,L4), append(EL,L4,L3).
flatten([E|L], [E|L2]) :- flatten(L,L2).

strict_subterm(X,T) :-
   nonvar(T),
   functor(T,_,N),
   subterm(X,T,N).

subterm(X,T,N) :- N > 0,
   arg(N,T,Arg),
   subterm(X,Arg).
subterm(X,T,N) :- N > 0, N2 is N-1, subterm(X,T,N2).

subterm(X,Y) :- var(Y), !, X==Y.
subterm(X,T) :-
   T=..[_|L],
   subterm_l(X,L).
subterm_l(X,[Y|_]) :- subterm(X,Y).
subterm_l(X,[_|L]) :- subterm_l(X,L).


check_is_a_variable_argument([Lit|_],ArgVar) :-
  (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
   Atom=..[_|ArgsH], belongs(ArgVar,ArgsH), var(ArgVar).
check_is_a_variable_argument([_|Lits],ArgVar) :-
  check_is_a_variable_argument(Lits,ArgVar).

is_a_variable_argument([Lit|_],ArgVar) :-
  (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
   Atom=..[_|ArgsH], member(ArgVar,ArgsH), var(ArgVar).
is_a_variable_argument([_|Lits],ArgVar) :-
  is_a_variable_argument(Lits,ArgVar).


is_an_output_argument([Lit|_],ArgVar) :-
  (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
   Atom=..[F|ArgsH], same_length(ArgsH,Modes), 
   mode(F,_,Modes), pos1(ArgVar,ArgsH,Pos), pos1(-,Modes,Pos).
is_an_output_argument([_|Lits],ArgVar) :-
  is_an_output_argument(Lits,ArgVar).


% From CIAO module "metaterm"
% bag_of_vars(+Term,-Vars) :- Vars is the bag of variables, in
% reversed order of occurrence, contained in Term

bag_of_vars(Term,Vars) :- varsbag(Term,Vars,[]).

varsbag(X,Vars,Tail) :-
	var(X),!,
	Vars = [X|Tail].
varsbag([H|T],Vars,Tail) :- !,
	varsbag(H,Vars,Tail0),
	varsbag(T,Tail0,Tail).
varsbag(Term,Vars,Tail) :-
	functor(Term,_,A),
	go_inside(A,Term,Vars,Tail).

go_inside(0,_,Tail,Tail) :- !.
go_inside(N,T,Bag,Tail) :-
	Nth is N-1,
	arg(N,T,ARG),
	varsbag(ARG,Bag,Tail0),
	go_inside(Nth,T,Tail0,Tail).



/*----------------
The following predicates have been borrowed from the quintus libraries.
------------------*/

%From library(sets): 
seteq(Set1, Set2) :- subset(Set1, Set2), subset(Set2, Set1).
subset([], _).
subset([Element|Elements], Set) :- memberchk(Element, Set),  subset(Elements, Set).
subtract([], _, []).
subtract([Element|Elements], Set, Dif) :-
     memberchk(Element, Set), !, subtract(Elements, Set, Dif).
subtract([Element|Elements], Set, [Element|Dif]) :- subtract(Elements, Set, Dif).

%From library(lists): 
subseq([], [], []).
subseq([Head|Tail], Sbsq, [Head|Cmpl]) :- subseq(Tail, Sbsq, Cmpl).
subseq([Head|Tail], [Head|Sbsq], Cmpl) :- subseq(Tail, Sbsq, Cmpl).

%From library(anti_unify): 
/* Anti-unification is the mathematical dual of unification:  given two terms T1 and T2
it returns the most specific term which generalises them, T.  T is the most specific term
which unifies with both T1 and T2.  A common use for this is in learning; the idea of using
it that way comes from Gordon Plotkin. The code here is based on a routine called
'generalize/5' written by Fernando Pereira. The name was changed because there are other
ways of generalising things, but there is nly one dual of unification.
*/
%   anti_unify(+Term1, +Term2, -Term)
%   binds Term to a most specific generalisation of Term1 and Term2.
%   When you call it, Term should be a variable.
anti_unify(Term1, Term2, Term) :- anti_unify(Term1, Term2, _, _, Term).

%   anti_unify(+Term1, +Term2, -Subst1, -Subst2, -Term)
%   binds Term to a most specific generalisation of Term1 and Term2,
%   and Subst1 and Subst2 to substitutions such that
%       Subst1(Term) = Term1
%       Subst2(Term) = Term2
%   Substitutions are represented as lists of Var=Term pairs, where
%   Var is a Prolog variable, and Term is the term to substitute for Var.
%   When you call it, Subst1, Subst2, and Term should be variables.
anti_unify(Term1, Term2, Phi1, Phi2, Term) :-
	anti_unify(Term1, Term2, [], [], Phi1, Phi2, Term).
anti_unify(Term1, Term2, Phi1, Phi2, Theta1, Theta2, Term) :-
   (nonvar(Term1), nonvar(Term2), functor(Term1, F, N), functor(Term2, F, N) ->
    functor(Term, F, N), anti_unify(N, Term1, Term2, Phi1, Phi2, Theta1, Theta2, Term)
   ; Term1 == Term2 -> Theta1 = Phi1, Theta2 = Phi2, Term = Term1
   ; already_covered(Phi1, Phi2, Term1, Term2, V) -> Theta1 = Phi1, Theta2 = Phi2, Term = V
   ;   /* Not same functor, not identical, not covered by an existing V */
    Theta1 = [V = Term1|Phi1], Theta2 = [V = Term2|Phi2], Term = V ).

anti_unify(0, _, _, SS, ST, SS, ST, _) :- !.
anti_unify(N, S, T, SS0, ST0, SS, ST, U) :-
   arg(N, S, Sn), arg(N, T, Tn), arg(N, U, Un),
   anti_unify(Sn, Tn, SS0, ST0, SS1, ST1, Un),
   M is N-1, anti_unify(M, S, T, SS1, ST1, SS, ST, U).

already_covered([V=S0|_], [V=T0|_], S, T, V) :- S0 == S, T0 == T, !.
already_covered([_|Phi1], [_|Phi2], S, T, V) :-	already_covered(Phi1, Phi2, S, T, V).

%From library(occurs): 
%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.
free_of_var(Variable, Term) :-	Term == Variable, !, fail.
free_of_var(Variable, Term) :- 
   compound(Term), !, functor(Term, _, Arity), free_of_var(Arity, Term, Variable).
free_of_var(_, _).
free_of_var(1, Term, Var) :- !, arg(1, Term, Arg), free_of_var(Var, Arg).
free_of_var(N, Term, Var) :-
   arg(N, Term, Arg), free_of_var(Var, Arg), M is N-1, !, free_of_var(M, Term, Var).

%   occurrences_of_var(+Term, +Variable, ?Tally)
%   is true when the given Variable occurs exactly Tally times in
%   Term.  It can only be used to calculate or test Tally;
%   to enumerate Variables you'll have to use sub_term/2 and then
%   test them with this routine.  If you just want to find out
%   whether Variable occurs in Term or not, use contains_var/2
%   or free_of_var/2.  This used to be called occ/3.
occurrences_of_var(Variable, Term, Occurrences) :-
   occurrences_of_var(Term, Variable, 0, Tally), Occurrences = Tally.
occurrences_of_var(Term, Var, SoFar, Tally) :- Term == Var, !, Tally is SoFar+1.
occurrences_of_var(Term, Variable, SoFar, Tally) :-
   compound(Term), !, functor(Term, _, Arity),
   occurrences_of_var(Arity, Term, Variable, SoFar, Tally).
occurrences_of_var(_, _, Tally, Tally).
occurrences_of_var(1, Term, Variable, SoFar, Tally) :- !,
   arg(1, Term, Argument), occurrences_of_var(Argument, Variable, SoFar, Tally).
occurrences_of_var(N, Term, Variable, SoFar, Tally) :-
   arg(N, Term, Argument), occurrences_of_var(Argument, Variable, SoFar, Accum),
   M is N-1, occurrences_of_var(M, Term, Variable, Accum, Tally).

%   free_variables(+Goal, +Bound, +Vars0, -Vars) binds Vars to the union of Vars0 with
%   the set of *FREE* variables in Goal, that is the set of variables which are captured
%   neither by Bound nor by any internal quantifiers or templates in Goal. 
free_variables(Term, Bound, Vars0, Vars) :-
   ( nonvar(Term) -> free_variables_1(Term, Bound, Vars0, Vars) 
   ; term_is_free_of(Bound, Term), list_is_free_of(Vars0, Term) -> Vars = [Term|Vars0]
   ;   Vars = Vars0 ).

free_variables_1((Conjunct,Conjuncts), Bound) --> !,
   free_variables(Conjunct,  Bound),	free_variables(Conjuncts, Bound).
free_variables_1((Disjunct ; Disjuncts), Bound) --> !,
   free_variables(Disjunct,  Bound), free_variables(Disjuncts, Bound).
free_variables_1(NormalGoal, Bound) --> data_variables(NormalGoal, Bound).

%   data_variables(+Term, +Bound, +Vars0, -Vars) binds Vars to the union of Vars0 with the
%   set of variables in Term which do not occur in (are not "captured by") Bound.
%   When the Bound contains no variables, it delivers the same results as
% term_variables(Term, Vars0, Vars).
data_variables(Term, Bound, Vars0, Vars) :-
   ( nonvar(Term) -> functor(Term, _, N), data_variables(N, Term, Bound, Vars0, Vars)
   ; term_is_free_of(Bound, Term), list_is_free_of(Vars0, Term) ->  Vars = [Term|Vars0]
   ;   Vars = Vars0 ).

data_variables(N, Term, Bound) -->
   (  { N =:= 0 } -> [] ; { arg(N, Term, Arg), M is N-1 },
   data_variables(Arg, Bound), data_variables(M, Term, Bound)).

%   term_is_free_of(+Term, +Var) is a meta-logical predicate which is true when the variable
%   Var does not occur anywhere in the term Term.  It is used when the Term is a tree
% built from all the existential quantifiers and Templates dominating
%   (the goal containing) this variable.
term_is_free_of(Term, Var) :-
   ( var(Term) -> Term \== Var ; functor(Term, _, N), term_is_free_of(N, Term, Var) ).
term_is_free_of(N, Term, Var) :-
   ( N =:= 0 -> true ; arg(N, Term, Arg), term_is_free_of(Arg, Var), M is N-1,
    term_is_free_of(M, Term, Var) ).

%   list_is_free_of(+Vars0, +Var) is a meta-logical predicate which is true when the
%   variable Var is not an element of Vars0, which is known to be a list of variables.
%   It is used when Vars0 is the set of free variables which have been built up so far,
%   and we are considering whether Var should be added to this set.
list_is_free_of([], _).
list_is_free_of([Head|Tail], Var) :- Head \== Var, list_is_free_of(Tail, Var).

%From library(?)
%   chars_to_term(+Chars, ?Term) takes a list of characters such as "f(a,b)", lacking a
%   terminal full stop and layout character, and reads a term from it, such as f(a,b).
%   It fails if there is anything left over.
chars_to_term(Chars, Term) :-
   chars_to_stream(Chars, 3, Stream), current_input(OldStream), set_input(Stream),
   ( read(X), get0(-1) -> Flag = 1 ; Flag = 0   ), set_input(OldStream), close(Stream),
   Flag =:= 1, Term = X.

chars_to_stream(Chars, Dotp, Stream) :-
   chars(Chars, 0, Length), !, new_byte_vector(Length, Dotp, Vector), Vector =\= 0,
   chars_to_byte_vector(Chars, Vector, Length), DottedLength is Length+Dotp,
   stream_from_byte_vector(Vector, DottedLength, StreamCode, 0), stream_code(Stream, StreamCode).
chars_to_stream(Chars, _, _) :-
   format(user_error, '~N! chars_to_stream/2: bad Chars argument ~w~n', [Chars]), fail.

%   chars(+Chars, +L0, -L) succeeds when Chars is a proper list of character codes,
%   and L-L0 is its length.
chars(-, _, _) :- !, fail.
chars([], L, L).
chars([Char|Chars], L0, L) :-
   integer(Char), 0 =< Char, Char =< 255, L1 is L0+1, chars(Chars, L1, L).

%   Store the characters of Chars into the byte vector Vector. The current length Length
%   is checked for overflow.
chars_to_byte_vector([], _, _).
chars_to_byte_vector([Char|Chars], Vector, Length) :-
   Length > 0,  store_byte(Char, Vector, Next), Left is Length-1,
   chars_to_byte_vector(Chars, Next, Left).

%From library(ask): 
%   ask_default_character(+Spec, ?Char) lets the programmer specify the default character
%   in whatever way s/he finds convenient, either as an integer, as a string, or as a
%   Prolog atom.  The case of the character is preserved.  We export it so your program
%   can use the same convention elsewhere.
ask_default_character(Spec, Char) :-
   (   integer(Spec) -> Char = Spec ; atomic(Spec) -> name(Spec, [Char|_])
   ;   Spec = [Head|_], nonvar(Head), ask_default_character(Head, Char) ).

%   yesno(+Question, +Default). This predicate has been changed by S. Renault
%   wrt its original definition in order to avoid the call to the predicate to_lower.
yesno(Question, Default) :-
    ask_default_character(Default, DefChar),
    repeat,
    prompted_char([Question,' [',{DefChar},']? '], Answer),
%    to_lower(Char, Answer),
    (   is_endline(Answer), !,    (DefChar = 89 ; DefChar= 121)
%    to_lower(DefChar, 0'y)
    ;   Answer =:= "y" -> !, true
    ;   Answer =:= "Y" -> !, true
    ;   Answer =:= "n" -> !, fail
    ;   Answer =:= "N" -> !, fail
    ;   prompt('Please answer Yes or No followed by RETURN'), ttynl, fail ).

%   prompt(+ListOrPrompt) writes a Prompt (or a List of terms making up a Prompt) to the
%   terminal, starting at the beginning of a new line, and forcing the output out without
%   adding a newline.  It is used for writing prompts. Note that a list of character codes
%   will NOT be printed as if it were a string.  This may be reconsidered.
prompt(ListOrPrompt) :-
   current_output(OldTell), set_output(user_output), format('~N', []),
   prompt_(ListOrPrompt), ttyflush, set_output(OldTell).

prompt_([]) :- !.
prompt_([Head|Tail]) :- !, prompt_(Head), prompt_(Tail).
prompt_({Ch}) :- integer(Ch), Ch >= 0, Ch =< 255, !, put(Ch).
prompt_(Other) :- write(Other).	% must be write, not writeq

%   asks a question where the answer should be an integer in the range Lower..Upper.
%   If the user enters an empty line, the Default (which need not be in range, or even
%   an integer) is returned instead. A typical use might be
%   ask_between('Item number', 1, N, 0, Index).
ask_between(Prompt, Lower, Upper, Default, Answer) :-
   repeat,
   prompted_numeric_line([Prompt,' [',Default,']: '], Chars),
   (   Chars = "", X = Default
   ; number_chars(X, Chars), integer(X), X >= Lower, X =< Upper
   ; format(user_output,'~NPlease enter an integer between ~d and ~d followed
      by RETURN.~n', [Lower,Upper]), fail ), !, Answer = X.

%   ask_oneof(+Prompt, +Constants, +Default, ?Selected) expects an answer to be (a prefix of)
%   one of the constants in the given list of constants. If the user enters an empty line,
%   the Default (which need not be among the Constants) is returned.
ask_oneof(Prompt, Constants, Default, Selected) :-
   repeat,
   prompted_line([Prompt,' [',Default,']: '], Chars),
   (   Chars = "", X = Default
   ;   prefix_of_unique(Constants, Chars, X)
   ;   format(user_output, '~NPlease enter one of these constants~n ~p followed by a RETURN.
   If you just type RETURN,~n ~p will be taken as your answer. Do not add a full stop.~n',
   [Constants,Default]), fail ), !, Selected = X.

%   prefix_of_unique(Constants, Chars, Selected) is true when (1) Chars is the name/2
%   of Selected which is a member of the list of constants Constants, or (2) Chars is
%   a prefix of that name, and there is only 1 such element.
prefix_of_unique(_, [0'?|_], _) :- !, fail.
prefix_of_unique(Constants, Chars, Sel) :- prefix_of_unique(Constants, Chars, _, Sel).
prefix_of_unique([], _, Flag, Selected) :- !, nonvar(Flag), Selected = Flag.
prefix_of_unique([Constant|_], Chars, _, Selected) :-
   name(Constant, Chars), !, Selected = Constant.
prefix_of_unique([Constant|Constants], Chars, Flag, Selected) :-
   \+ (name(Constant,L), \+ append(Chars,_,L)), !, Flag = Constant,
   prefix_of_unique(Constants, Chars, Flag, Selected).
prefix_of_unique([_|Constants], Chars, Flag, Selected) :-
   prefix_of_unique(Constants, Chars, Flag, Selected).

%   ask_number(+Prompt, +Default, ?Answer) asks a question where the answer should be
%   some sort of number, either integer or float.  The answer is left as whatever type
%   the user supplied:  integers are not converted to floats.  If the user supplies an
%   empty line, the Default (which need not be a number) is returned.
%   Layout, "+", and "_" are ignored.
ask_number(Prompt, Default, Answer) :-
   repeat,
   prompted_numeric_line([Prompt,' [',Default,']: '], Chars),
   (   Chars = "", X = Default
   ;   number_chars(X, Chars)
   ;   format(user_output, '~NPlease enter a number followed by RETURN~n', []),
       fail  ), !, Answer = X.

%   ask_number(+Prompt, +Lower, +Upper, +Default, ?Answer) asks a question where the
%   answer should be a number in the range Lower..Upper.  If the user enters an empty
%   line, the Default (which need not be in range, or even a number) is returned.
%    A typical use might be ask_number('Percentage', 0, 100, 50, Percentage).
ask_number(Prompt, Lower, Upper, Default, Answer) :-
   repeat,
   prompted_numeric_line([Prompt,' [',Default,']: '], Chars),
   (   Chars = "", X = Default
   ;   number_chars(X, Chars), X >= Lower, X =< Upper
   ;	format(user_output,'~NPlease enter a number between ~w and ~w
        followed by RETURN~n', [Lower,Upper]), fail ), Answer = X.

prompted_char(Prompt, Char) :-
   prompt(Prompt),                % write the prompt to the terminal
   ttyget0(C1),                   % read the first character
   (   is_endline(C1)             % either it was an end of line
   ;   repeat,	                  % or else we have to skip the
   ttyget0(C2),	                  % rest of the terminal line
   is_endline(C2)), !, Char = C1. % we have to consume the end of line

%   prompted_numeric_line(+Prompt, ?Chars) is a version of prompted_line/2 which strips
%   away layout characters, underscores, and leading "+" signs from its result.  It
%   also strips any "."s which are not followed by a digit, so that it does no harm
%   to enter "123." as if you were typing to read/1.
prompted_numeric_line(Prompt, Neat) :-
   prompted_line(Prompt, Chars), filter_numeric(Chars, 0, Neat).

filter_numeric([], _, []).
filter_numeric([Char|Raw], State, Neat) :-
   (   Char =< " " -> true
   ;   Char =:= "_" -> true
   ;   Char =:= "+" -> State =:= 0
   ;   Char =:= "." -> not_before_a_digit(Raw)
   ;   fail ), !, filter_numeric(Raw, State, Neat).
filter_numeric([Char|Raw], _, [Char|Neat]) :- filter_numeric(Raw, 1, Neat).

not_before_a_digit([]).
not_before_a_digit([X|_]) :- ( X < "0" -> true ; X > "9" -> true ).

%   prompted_line(+Prompt, -Chars) writes a prompt to the terminal and reads a line of
%   characters from the terminal.  It fails if the user ended the line with the EOF
%   character. This might move into ask.pl eventually.
prompted_line(Prompt, Chars) :- prompt(Prompt),	fget_line(user, Chars).

%   fget_line(+Stream, ?Chars) reads a line from the given input Stream, and returns the
%   characters in the list Chars.  It does NOT return the line terminating character,
%   so it is useful for portable programming.  If the terminator was the end of the
%   file, it simply fails and later calls will abort.
fget_line(Str, Chars) :- fget_line(Str, Line, Terminator), Terminator >= 0, Chars = Line.

%   fget_line(+Stream, ?Chars, ?Terminator) reads a line from the given input Stream,
%   and returns the characters in the list Chars, and the line terminating character
%   in Terminator.  If the terminator was end of file, it just returns it like always.
%   When you use this routine, the last line will often be ignored if not properly ended.
fget_line(Stream, Chars, Term) :-
   current_input(OldInput), see(Stream), get_line_1(Line, Terminator),
   set_input(OldInput), Chars = Line, Term = Terminator.

get_line_1(Line, Terminator) :-	     % 9 is the ASCII TAB character.
   get0(Char),                       % 5 is the EBCDIC TAB character.
   (   Char < " ", Char =\= 9, Char =\= 5 ->
       Line = [], Terminator = Char
   ;   Line = [Char|Chars], get_line_1(Chars, Terminator) ).

%   get_line(?Chars) reads a line from the current input stream, and returns the characters
%   in the list Chars. It does NOT return the line terminating character, so it is useful
%   for portable programming.  If the terminator was the end of the file, it simply
%   fails and later calls will abort.
get_line(Chars) :- get_line_1(Line, Terminator), Terminator >= 0, Chars = Line.

get_file(File, Lines) :-
   current_input(OldStream), open(File, read, NewStream), set_input(NewStream),
   get_file(Ls), set_input(OldStream), close(NewStream), Lines = Ls.

get_file(Lines) :-
   get0(Char),
   (   Char  <   0 -> Lines = []
   ;   Char =:= 10 -> Lines = [[]|Rest], get_file(Rest)
   ;                  Lines = [Line|Rest], get_file(Char, Line, Rest) ).

get_file(Char, [Char|Line], Lines) :-
   get0(Next),
   (   Next  >  10 -> get_file(Next, Line, Lines)
   ;   Next =:= 10 -> Line = [], get_file(Lines)
   ;   Next  <   0 -> Line = [], Lines = []
   ;                  get_file(Next, Line, Lines) ).

%   ask_chars(+Prompt, +MinLength, +MaxLength, ?Answer) expects an answer which is a
%   sequence of between MinLength and MaxLength characters, not beginning with a '?'.
ask_chars(Prompt, MinLength, MaxLength, Answer) :-
   repeat,
   prompted_line([Prompt,': '], Chars), length(Chars, Length),
   (   Length >= MinLength, Length =< MaxLength, \+ (Chars = [0'?|_])
    ;   format(user_output, '~NPlease enter between ~d and ~d characters followed by RETURN.
   Do not add a full stop unless it is part of the answer.~n',
   [MinLength,MaxLength]), fail ), !, Answer = Chars.

print_length(Command, Length) :-
   open_null_stream(Stream), 'print length'(Command, user, Stream),
   character_count(Stream, CharsWritten), line_count(Stream, LineNumber), close(Stream),
%	LineNumber = 1,
	LineNumber = 0,      %%%%% Quintus -> Sicstus
	Length = CharsWritten.

'print length'(Module:Command, _, Stream) :- !, 'print length'(Command, Module, Stream).
'print length'(write(X), _, Stream) :- !, write(Stream, X).
'print length'(writeq(X), _, Stream) :- !, writeq(Stream, X).
'print length'(display(X), _, Stream) :- !, write_canonical(Stream, X).
'print length'(write_canonical(X), _, Stream) :- !, write_canonical(Stream, X).
'print length'(print(X), _, Stream) :- !, print(Stream, X).
'print length'(format(F,L), _, Stream) :- !, format(Stream, F, L).
'print length'(Command, Module, Stream) :-
   current_output(OldOutput), set_output(Stream),
   (   Module:Command -> true
   ;   set_output(OldOutput), close(Stream), fail ), set_output(OldOutput).

%   is_endline(+Char) succeeds when Char is the code of some character which terminates
%   a line.  Char must already be instantiated.
is_endline(C) :-                % line terminator, NOT just newline
   C < 32,                      % covers EOF, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_
   C =\= 9 /* ^I */.            % not ^I (TAB) though.
%\end{verbatim}
%\end{small}
%\end{document}