% case_split.pl
%
% This file contains the predicates for the case split rule used in the strategy determ.
% It uses the notion of predicate modes. There are two options for the rule:
% binary case split and case split by global analysis. Predicate modes are assumed to be
% set in the file determ_strat_parameters.pl. If there are missing, there are requested
% to the user via a menu.

:- dynamic modes/3.

:- use_module(library(lists), [ append/3,
                                member/2 ]).
:- use_module(library(terms), [ variant/2,
                                subsumes_chk/2]).
                                %unify_with_occurs_check/2 ]).

/*-----------  CASE SPLIT BY GLOBAL ANALYSIS ------------------*/
% Clauses is a list of clause numbers, NewCNS is a list of list of clause numbers (packets),
% NewCN is the flattening of NewCNS.

apply_case_split_global([], [[]],_).
apply_case_split_global(Clauses,NewCNS,NewCN) :-
    split_clauses_into_packets(Clauses,Packets),
    (Packets=[_,_|_] ->
    (add_clauses_packets(Packets,NewCNS), append(NewCNS,NewCN),
    update_programs((Clauses, case_split, NewCN), _),
    (derivation_trace_on -> new_clauses(case_split, Clauses, NewCNS);true))
    ; NewCNS=[Clauses], NewCN=Clauses).

split_clauses_into_packets(Clauses,Packets) :-
    restrict_clauses_to_input_heads(Clauses,I_Heads),
    construct_general_structure_of_packets(I_Heads,G_Struct),
    fill_general_structure_of_packets(Clauses,G_Struct,Struct),
    resulting_packets(Struct,Packets).

add_clauses_packets([],[]).
add_clauses_packets([P|Packets],[NewCN|NewCNS]) :-
    add_clauses(NewCN,P,nil), add_clauses_packets(Packets,NewCNS).

restrict_clauses_to_input_heads([],[]).
restrict_clauses_to_input_heads([C|L],LRH) :-
    head_of_clause(C,H), functor(H,F,N), read_modes(F,N,Modes), H=..[_|ArgsF],
    restrict_arg_to_input(ArgsF,Modes,NewArgsF), RH=.. [F|NewArgsF],
    restrict_clauses_to_input_heads(L,Modes,[RH],LRH).

restrict_clauses_to_input_heads([],_,L,L).
restrict_clauses_to_input_heads([C|L],Modes,L_i, LRH):-
    head_of_clause(C,H), H=..[F|ArgsF],
    restrict_arg_to_input(ArgsF,Modes,NewArgsF), RH=.. [F|NewArgsF],
    ((member(E,L_i), variant(E,RH)) -> restrict_clauses_to_input_heads(L,Modes,L_i,LRH)
    ; restrict_clauses_to_input_heads(L,Modes,[RH|L_i],LRH)).

restrict_arg_to_input([],[],[]).
restrict_arg_to_input([X|ArgsF],['+'|Modes],[X|NewArgsF]) :- !,
    restrict_arg_to_input(ArgsF,Modes,NewArgsF).
restrict_arg_to_input([_|ArgsF],[_|Modes],[_|NewArgsF]) :-
    restrict_arg_to_input(ArgsF,Modes,NewArgsF).

construct_general_structure_of_packets(L,L2):-
    infer_missing_heads(L,LL), construct_general_structure_of_packets(LL,[],L2).
construct_general_structure_of_packets([],_,[]).
construct_general_structure_of_packets([H|Heads], Headsi, [ExtH|ExtHeads]) :-
    append(Heads,Headsi,HeadsT), copy_term(HeadsT,HeadsT_C), copy_term(H,H_C),
    infer_extended_heads(H_C,HeadsT_C, ExtH),
    construct_general_structure_of_packets(Heads, [H|Headsi], ExtHeads).

infer_missing_heads([],[]).
infer_missing_heads([H|Heads], All_Heads) :-
    infer_missing_heads(H, Heads, Heads1),
    infer_missing_heads(Heads,Heads2), append(Heads1,Heads2,All_Heads).

infer_missing_heads(H, [], [H]).
infer_missing_heads(H, [H2|Heads], Heads2) :-
   (subsumes_chk(H,H2) ; subsumes_chk(H2,H) 
   ; \+ unify_with_occurs_check(H,H2)),!, infer_missing_heads(H, Heads, Heads2).
infer_missing_heads(H, [H2|Hs], [H2_C|Hs2]) :-
   copy_term(H,H_C), copy_term(H2,H2_C), unify_with_occurs_check(H_C,H2_C),
   infer_missing_heads(H, Hs, Hs2).

infer_extended_heads(H_C,HsT_C, ExtH) :-
    more_general_instances(H_C,HsT_C,L), infer_constraints(H_C,L, (H_C,[]), ExtH).

more_general_instances(H,Hs,L):- all_instances(H,Hs,L2), more_general_instances2(L2,[],L).

all_instances(_,[],[]).
all_instances(H,[H1|Hs],[H1|L]) :- 
    \+ variant(H,H1), subsumes_chk(H,H1), !, all_instances(H,Hs,L).
all_instances(H,[_|Hs],L) :- all_instances(H,Hs,L).

more_general_instances2([],I,I).
more_general_instances2([H|Hs],L,L2):-
   (exist_more_general(H,Hs) ; exist_more_general(H,L)), !, more_general_instances2(Hs,L,L2).
more_general_instances2([H|Hs],L,L2):-	append(L,[H],LL), more_general_instances2(Hs,LL,L2).

exist_more_general(H,Hs) :- member(H2,Hs), \+ variant(H,H2), subsumes_chk(H2,H).

update_constraints([],Const,Const) :- !.
update_constraints(Const1,Const, NewConst) :- append(Const1,Const,NewConst).

infer_constraints(_,[], (HC,I), (HC,I,[])).
infer_constraints(HC,[H|L], (HC,I), ExtH) :-
   functor(HC,F,N), read_modes(F,N,Modes), HC=.. [_|ArgsHC], H=.. [_|ArgsH],
   input_args(ArgsHC, Modes, R_ArgsHC), input_args(ArgsH, Modes, R_ArgsH),
   anti_unify(R_ArgsHC,R_ArgsH,T1,T2,_), recover_constraint(T1,T2,Const),
   update_constraints(I,Const,NewConst), infer_constraints(HC, L, (HC,NewConst), ExtH).

input_args([],[],[]).
input_args([X|ArgsF],['+'|Modes],[X|NewArgsF]) :-!, input_args(ArgsF,Modes,NewArgsF).
input_args([_|ArgsF],[_|Modes],[NewArgsF]) :- input_args(ArgsF,Modes,NewArgsF).

fill_general_structure_of_packets([],Packets, Packets).
fill_general_structure_of_packets([C|L], Packets_i, Packets) :-
   fill_general_structure_with_clause(C, Packets_i, Packets_i2),
   fill_general_structure_of_packets(L, Packets_i2, Packets).
fill_general_structure_with_clause(_, [], []).
fill_general_structure_with_clause(C,[(H,Const,CC)|Hs],[(H,Const,[(HC_C,BdC_C)|CC])|P_i2]) :-
   clause_number_to_clause(C,(HC,BdC)), subsumes_chk_wrt_input_pos(HC,H),!,
   copy_term((HC,BdC), (HC_C,BdC_Ci)), copy_term((H,Const), (H_C,Const_C)), unify_with_occurs_check(HC_C, H_C),
   append(Const_C,BdC_Ci,BdC_C), fill_general_structure_with_clause(C, Hs, P_i2).
fill_general_structure_with_clause(C, [X|Hs], [X|P_i2]) :-
   fill_general_structure_with_clause(C, Hs, P_i2).

subsumes_chk_wrt_input_pos(Head,H1) :-
   functor(Head,F,N), read_modes(F,N,Modes), Head=.. [_|ArgsHead], H1=.. [_|ArgsH1],
   restrict_arg_to_input(ArgsHead, Modes, R_ArgsHead),
   restrict_arg_to_input(ArgsH1, Modes, R_ArgsH1), subsumes_chk(R_ArgsHead,R_ArgsH1).

resulting_packets([],[]).
resulting_packets([(_,_,Cls1)|Packets],[Cls1|New_Clauses]) :-
   resulting_packets(Packets,New_Clauses).

recover_constraint([],[],[]).
recover_constraint([(_=_)|T1],[(_=X)|T2], Const) :-
   var(X), singleton_variable(X,T2),!, recover_constraint(T1,T2, Const).
recover_constraint([(_=A)|T1],[(_=X)|T2],[\+(A=X)|Const]):- recover_constraint(T1,T2, Const).

singleton_variable(X,T) :- free_of_var(X,T).

/*---------------------  BINARY CASE SPLIT -----------------------*/
% The head of clause Clause_Number2 has an input argument which is more general than
% the corresponding argument in the head of clause Clause_Number.
% Clauses is the result of the split of clause Clause_Number and contains two clauses.

apply_case_split_binary([], [[]],_).
apply_case_split_binary(Clauses,NewCNS,NewCN) :-
    split_clauses(Clauses,Split1,Split2), (Split2 \== [] ->
    apply_case_split_binary2(Split1,Split2,NewCNS,NewCN) ;
    apply_case_split_binary1(Clauses,NewCNS,NewCN)).

apply_case_split_binary1(Clauses,NewCNS,NewCls) :-
    anti_unify_heads(Clauses,Head,U),
    (exist_clause_to_case_split(Clauses,Head,U,CA) ->
    apply_case_split_binary1_possible(Clauses,CA,NewCNS,NewCls) ;
    apply_case_split_binary1_impossible(Clauses,NewCNS,NewCls)).

exist_clause_to_case_split([_],_,_,_):- !, fail.
exist_clause_to_case_split(Clauses,Head,U,CA) :-
    exist_clause_to_case_split(Clauses,Head,U,CA,Clauses).
exist_clause_to_case_split([C1|_],Head,[(H1,_)|_],C1,AllClauses) :-
    \+ variant(Head,H1), subsumes_chk_wrt_input_mode(Head,H1),
    verify_suitable_clause(H1,Head,AllClauses), !.
exist_clause_to_case_split([_|Clauses],Head,[_|U],C,AllC) :-
    exist_clause_to_case_split(Clauses,Head,U,C,AllC).

subsumes_chk_wrt_input_mode(Head,H1):-
    functor(Head,F,N),
    (read_modes(F,N,Modes) -> subsumes_chk_wrt_input_mode(Modes,Head,H1) 
    ; (nl, write('No modes for '), write(F/N), subsumes_chk(Head,H1))).
subsumes_chk_wrt_input_mode(Modes,Head,H1) :-
    Head=.. [_|ArgsHead], H1=.. [_|ArgsH1],
    subsumes_chk_wrt_input_mode_args(Modes,ArgsHead,ArgsH1).
subsumes_chk_wrt_input_mode_args(['+'|_],[A|_],[A2|_]):-
    \+ variant(A,A2), subsumes_chk(A,A2).
subsumes_chk_wrt_input_mode_args([_|Modes],[_|ArgsHead],[_|ArgsH1]):-
    subsumes_chk_wrt_input_mode_args(Modes,ArgsHead,ArgsH1).

verify_suitable_clause(H,Head,[C|_]) :-
    clause_number_to_clause(C,(H1,B1)), anti_unify(H,H1,U), variant(U,Head),
    subsumes_chk_wrt_input_mode(H1,H), copy_term(H,C_H), copy_term((H1,B1),(C_H1,C_B1)),
    unify_wrt_input_mode(C_H,C_H1), true_equalities(C_B1,_), !.
verify_suitable_clause(C1,Head,[_|AllClauses]) :- verify_suitable_clause(C1,Head,AllClauses).

case_split_clauses(Cls,C,Cls2):- case_split_clauses(Cls,C,[],Cls2).
case_split_clauses([],_,Cls,Cls).
case_split_clauses([C|Cls],C,Cls_i,Cls2) :-	!,
    append(Cls_i,[C],Cls_i2),
    case_split_clauses(Cls,C,Cls_i2,Cls2).
case_split_clauses([C|Cls],CA,Cls_i,Cls2):-
    head_of_clause(C,H),
    head_of_clause(CA,H2),
    ((\+ unify_wrt_input_mode(H,H2) ; variant_wrt_input_mode(H,H2)) -> 
    (append(Cls_i,[C],Cls_i2), case_split_clauses(Cls,CA,Cls_i2,Cls2))
    ; (apply_case_split(C,CA,Cls1_bef_simpl), check_equalities(Cls1_bef_simpl,Cls1),
    add_clauses(NewCNs,Cls1,nil),
    update_programs((C,(case_split,CA), NewCNs),_),
    (derivation_trace_on -> new_clauses(case_split,C,CA,NewCNs);true),
    append(Cls_i,NewCNs,Cls_i2),
    case_split_clauses(Cls,CA,Cls_i2,Cls2))).

apply_case_split(Clause_Number,Clause_Number2,Clauses) :- 
    clause_number_to_clause(Clause_Number,(H,B)),
    head_of_clause(Clause_Number2,H2),
    anti_unify(H,H2, Theta_H, Theta_H2,_),
    recover_case_split_arg(H,Theta_H, Theta_H2,A_H,A_H2),
    case_split_arg((H,B),A_H,A_H2,Clauses).

recover_case_split_arg(H,[V1=A_H|_],[V1=A_H2|_],A_H,A_H2):-
    functor(H,F,N), read_modes(F,N,Modes), H =.. [_|Args],
    var(A_H), nonvar(A_H2), occur_in_input(A_H,Args,Modes), !.
recover_case_split_arg(H,[_|L1],[_|L2],A_H,A_H2):-
    recover_case_split_arg(H,L1,L2,A_H,A_H2).

apply_case_split_binary1_impossible(Clauses,[Clauses],Clauses).

apply_case_split_binary1_possible(Cs,CA,NewCNS,NewCls) :-
   case_split_clauses(Cs,CA,NewCs),
   (strat(apply_subsumption(during_case_split)) -> apply_subsumption_clauses(NewCs,Cs2)
   ; Cs2=NewCs), apply_case_split_binary(Cs2,NewCNS,NewCls).

apply_case_split_binary2(Split1,Split2,NewCNS,NewCN) :-
   apply_case_split_binary(Split1,NewCNS1,NewCN1),
   apply_case_split_binary(Split2,NewCNS2,NewCN2),
   append(NewCNS1,NewCNS2,NewCNS), append(NewCN1,NewCN2,NewCN).

split_clauses(L,L1,L2):-
   label_arc(_,(_,(case_split,X),_),_), member(X,L), !,
   split_clauses_after_case_split(L,X,L1,L2).
split_clauses([A|C],L1,L2):- split_clauses(C,[A],[],L1,L2,[A|C]).

split_clauses_after_case_split([],_,[],[]).
split_clauses_after_case_split([X|L],X,[X|L1],L2):- !,
   split_clauses_after_case_split(L,X,L1,L2).
split_clauses_after_case_split([C|L],X,[C|L1],L2):-
    label_arc(_,(_,(case_split,X),[C,_]),_), !,
    split_clauses_after_case_split(L,X,L1,L2).
split_clauses_after_case_split([C|L],X,L1,[C|L2]):-
    label_arc(_,(_,(case_split,X),[_,C]),_), !,
    split_clauses_after_case_split(L,X,L1,L2).
split_clauses_after_case_split([C|L],X,[C|L1],L2):-
    head_of_clause(C,H), head_of_clause(X,H2), unify_wrt_input_mode(H,H2), !,
    split_clauses_after_case_split(L,X,L1,L2).
split_clauses_after_case_split([C|L],X,L1,[C|L2]):-
    split_clauses_after_case_split(L,X,L1,L2).

split_clauses([C1|L],I1,I2,L1,L2,InitSet):-
    no_common_instance_wrt_input_mode(C1,I1),!, append(I2,[C1],I22),
    split_clauses(L,I1,I22,L1,L2,InitSet).
split_clauses([C1|L],I1,I2,L1,L2,InitSet):-
    no_common_instance_wrt_input_mode(C1,I2),!, append(I1,[C1],I11),
    split_clauses(L,I11,I2,L1,L2,InitSet).
split_clauses([],I1,I2,I1,I2,_):-!.
split_clauses(_,_,_,InitSet,[],InitSet).

no_common_instance_wrt_input_mode(N,L):- common_instance_wrt_input_mode(N,L), !, fail.
no_common_instance_wrt_input_mode(_,_).

common_instance_wrt_input_mode(N,L):-
    head_of_clause(N,H1), common_instance_with_head_wrt_input_mode(N,H1,L).

common_instance_with_head_wrt_input_mode(N,H,[N3|L]):-
    (label_arc(_,(_,(case_split,_),[N,N2])) ; label_arc(_,(_,(case_split,_),[N2,N]))),
    variant_heads(N3,N2), !, common_instance_with_head_wrt_input_mode(N,H,L).
common_instance_with_head_wrt_input_mode(N,H,[N2|L]):-
    label_arc(_,(_,(case_split,N3),[_,N2])), variant_heads(N3,N), !,
    common_instance_with_head_wrt_input_mode(N,H,L).	    
common_instance_with_head_wrt_input_mode(N,H,[N3|L]):-
    label_arc(_,(_,(case_split,N2),[_,N])), variant_heads(N2,N3), !,
    common_instance_with_head_wrt_input_mode(N,H,L).	    
common_instance_with_head_wrt_input_mode(_,H1,[N|_]):-
    head_of_clause(N,H2), \+ \+ unify_wrt_input_mode(H1,H2).
common_instance_with_head_wrt_input_mode(N,H1,[_|L]):-
    common_instance_with_head_wrt_input_mode(N,H1,L).

variant_heads(N2,N3) :- head_of_clause(N2,H2), head_of_clause(N3,H3), variant(H2,H3), !.

unify_wrt_input_mode(H1,H2) :-
   functor(H1,F,N),
   (read_modes(F,N,Modes) -> 
       unify_wrt_input_mode(H1,H2,Modes) ; \+ \+ unify_with_occurs_check(H1,H2)).
unify_wrt_input_mode(H1,H2,Modes) :-
   H1=..[_|ArgsH1], H2=..[_|ArgsH2], unify_wrt_input_mode_args(ArgsH1,ArgsH2,Modes).

unify_wrt_input_mode_args([A|ArgsH1],[A2|ArgsH2],['+'|Modes]) :- !,
    unify_with_occurs_check(A,A2), unify_wrt_input_mode_args(ArgsH1,ArgsH2,Modes).
unify_wrt_input_mode_args([_|ArgsH1],[_|ArgsH2],[_|Modes]) :-
    unify_wrt_input_mode_args(ArgsH1,ArgsH2,Modes).
unify_wrt_input_mode_args([],[],[]).

variant_wrt_input_mode(H1,H2) :-
    functor(H1,F,N), (read_modes(F,N,Ms) -> variant_wrt_input_mode(H1,H2,Ms); variant(H1,H2)).
variant_wrt_input_mode(H1,H2,Modes) :-
    H1=..[_|ArgsH1], H2=..[_|ArgsH2], variant_wrt_input_mode_args(ArgsH1,ArgsH2,Modes).

variant_wrt_input_mode_args(ArgsH1,ArgsH2,Modes) :-
   find_input_mode_args(ArgsH1,ArgsH2,Modes,S_ArgsH1,S_ArgsH2), variant(S_ArgsH1,S_ArgsH2).

find_input_mode_args([A|ArgsH1],[A2|ArgsH2],['+'|M],[A|S_ArgsH1], [A2|S_ArgsH2]) :- !,
    find_input_mode_args(ArgsH1,ArgsH2,M,S_ArgsH1,S_ArgsH2).
find_input_mode_args([_|ArgsH1],[_|ArgsH2],[_|Modes],S_ArgsH1,S_ArgsH2) :-
    find_input_mode_args(ArgsH1,ArgsH2,Modes,S_ArgsH1,S_ArgsH2).
find_input_mode_args([],[],[],[],[]).

