% clause_replacement_rules.pl
%
% This file contains the predicates which specify the effects of applying the
% transformation rules associated with the predicates restriction, delete_clauses,
% delete_failing_body_clauses, add_clause, replace_clause, move_clause,
% subsumption, split_clause, which are defined in top_level.pl, after selection
% and checking of the input parameters.

:- use_module(library(terms), [ variant/2, subsumes_chk/2 ]).

/*------------
apply_restriction(L,Cls,Cls2): Eliminates transient definitions for predicates in L, 
that is, clauses that are not used in the definition of the predicates in L.
--------------*/
apply_restriction([P/A], Cls, Cls2) :- apply_restriction([P/A], Cls, [], Cls2).
apply_restriction(_, [], Cls, Cls) :- !.
apply_restriction(L, Cls, Inter, Cls2) :-
   member_rest(C,Cls,Cls3), head_of_clause(C,Head), predname_and_arity(Head,P,A),
   member(P/A,L),!, get_predicates_body(L,C,L2), append(Inter, [C], Inter2),
   apply_restriction(L2, Cls3, Inter2, Cls2).
apply_restriction(_, Elim, Cls, Cls) :-
    update_programs((Elim,elimclauses,[]), _),
    (derivation_trace_on -> new_clause(elimclauses,Elim);true).

/*-----------
apply_delete_clause(C): delete clause C from current program.
------------*/
apply_delete_clause(Clause_Number) :-
  current_prog(PN,CNs), (label_arc(PN,(Clause_Number,delclause,[]), PNSucc) -> 
  (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));           
  update_programs((Clause_Number,delclause,[]), PNSucc)),
  (derivation_trace_on -> new_clause(delclause,Clause_Number);true).

/*-----------
apply_del_cl_failing_bd(C): delete clause C from current program. This predicate is used
to delete a clause by application of the delete_failing_body_clauses rule or in the context
of an unfold_d_ing step, i.e. by application of the unfold_d, unfold_d_p or unfold_d_i rules.
------------*/
apply_del_cl_failing_bd(C) :- 
     current_prog(PN,CNs), (label_arc(PN,(C,delclause,[]), PNSucc) -> 
     (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));           
     update_programs((C,delclause,[]), PNSucc)),
     (derivation_trace_on -> new_clause(delclause_failure_body,C);true).

/*-----------
apply_add_clause(C,Pos): add clause C to current program at position Pos.
------------*/
apply_add_clause(Clause,Pos) :-
        current_prog(PN,CNs), rearrange_definition(Clause, (H,Body_list)),
        ((label_arc(PN,(Pos,addclause,NewCN),PNSucc), NewCN=[N],
        clause_number_to_clause(N,(Head,Body)), 
        variant((H,Body_list),(Head,Body))) -> (label_node(PNSucc,CNsSucc),
        update_current_prog((PN,CNs),(PNSucc,CNsSucc)));
        (add_clauses(NewCN,[(H,Body_list)],nil),
        update_programs((Pos,addclause,NewCN),  PNSucc))),
        (derivation_trace_on -> new_clause(addclause,NewCN);true).
/*-----------
apply_replace_clause(C1,C2): replace clause C1 by clause C2.
------------*/
apply_replace_clause(Cl,Clause) :-
    current_prog(PN,CNs), rearrange_definition(Clause, (H,Body_list)),
    ((label_arc(PN,(Cl,repclause,NewCN),PNSucc), NewCN=[N],
    clause_number_to_clause(N,(Head,Body)),
    variant((H,Body_list),(Head,Body))) ->
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)));
    (add_clauses(NewCN,[(H,Body_list)],nil),
    update_programs((Cl,repclause,NewCN),  PNSucc))), NewCN=[N],
    (derivation_trace_on -> new_clause(repclause,Cl,N);true).

/*-----------
apply_move_clause(C,Pos): move clause C to position Pos
------------*/
apply_move_clause(Clause,Pos) :-
     update_programs((Pos,moveclause,Clause),_),
     (derivation_trace_on -> new_clause(moveclause,Clause);true).

/*-----------
apply_substitution_clauses(L1,L2): L2 is the list L1 in which the clauses which are subsumed
by another clause in L1 are deleted.
------------*/
apply_subsumption_clauses(L1,L2):- apply_subsumption_clauses(L1,[],L2).
apply_subsumption_clauses([],L,L).
apply_subsumption_clauses([E|L],I,F) :-
   is_subsumed(E,L),!, update_programs((E,delclause,[]),_),
   apply_subsumption_clauses(L,I,F).
apply_subsumption_clauses([E|L],I,F) :-
   is_subsumed(E,I),!, update_programs((E,delclause,[]),_),
   apply_subsumption_clauses(L,I,F).
apply_subsumption_clauses([E|L],Inter,F) :-
    append(Inter,[E],Inter2), apply_subsumption_clauses(L,Inter2,F).

is_subsumed(E,[F|_]):- is_subsumed_clause(E,F),!.
is_subsumed(E,[_|L]):- is_subsumed(E,L).
% N is subsumed by M
is_subsumed_clause(N,M):-
    clause_number_to_clause(N,(H1,B1)), clause_number_to_clause(M,(H2,B2)),
    (subsumes_chk( (H2,B2), (H1,B1)) ; (append(F,_,B1), subsumes_chk((H2,B2), (H1,F)))),
   (derivation_trace_on ->
        (write('clause '), write(N), write(' subsumed by '), write(M), nl);
        true).

apply_delete_tautologies([],[]).
apply_delete_tautologies([T|L],L2) :-
    is_a_tautology(T), !,
    apply_delete_tautology(T),
    apply_delete_tautologies(L,L2).
apply_delete_tautologies([T|L],[T|L2]) :-
    apply_delete_tautologies(L,L2).


/*-----------
apply_delete_tautology(C): delete clause C from current program if C is of the
form A :- A.
------------*/
apply_delete_tautology(Clause_Number) :-
  update_programs((Clause_Number,deltautology,[]),_),
  (derivation_trace_on -> new_clause(deltautology,Clause_Number);true).

/*-----------
apply_delete_bool_simplification(C): delete clause C from current program if C
contains A and \+ A in its body.
------------*/

apply_delete_bool_simplification(Clause_Number) :-
  update_programs((Clause_Number,delboolsimpl,[]),_),
  (derivation_trace_on -> new_clause(delboolsimpl,Clause_Number);true).

/*-----------
apply_split_clause(Clause_N,A,Cls,Applied):
Cls is the list of the two clauses which results from the split of clause Clause_N wrt atom A.
Case split is only possible if A and the head of clause Clause_N are unify. In this case
Applied is set to constant "applied", which will indicate at some later stage that a
transformation step has been applied. Otherwise it is left as a variable which will indicate
that the transformation tree is unchanged.
-----------*/
apply_split_clause(Clause_N,A,Cls,Applied) :-
   clause_number_to_clause(Clause_N,(H,B)),
   (variant(H,A) -> msg_unchanged_by_case_split ;
   (\+ \+  H=A -> apply_split_clause(H,B,A,Cls,Applied)
   ; msg_unchanged_by_case_split)).
apply_split_clause(H,B,A,Cls,Applied) :-
    anti_unify(H,A, Theta_H, Theta_H2,_),
    recover_split_clause_arg(Theta_H,Theta_H2,A_H,A_H2,Applied),
    case_split_arg((H,B),A_H,A_H2,Cls).

recover_split_clause_arg([V1=A_H|_],[V1=A_H2|_],A_H,A_H2,applied):-
    var(A_H), nonvar(A_H2), !.
recover_split_clause_arg([_|L1],[_|L2],A_H,A_H2,Applied):-
    recover_split_clause_arg(L1,L2,A_H,A_H2,Applied).
recover_split_clause_arg([],[],_,_,_):- msg_unchanged_by_case_split.

case_split_arg((H,B),A_H,K, [(H,B3), (H2, [\+(A_H2=K)| B2])]):-
   copy_term((A_H,H,B),(A_H2,H2,B2)), A_H=K,
   singletons_vars_body(B,Singletons),
   simpl_equalities_body(Singletons,B,B3).

simpl_equalities_body(_,[],[]).
simpl_equalities_body(Singletons,[([A|B]=[C|D])|Bd1],[B=D|Bd2]):-
   A==C,!, simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[\+([A|B]=[C|D])|Bd1],[\+(A=C)|Bd2]):-
   (belongs(D,Singletons);belongs(B,Singletons)),
   simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[P|Bd1],[P|Bd2]):-
   simpl_equalities_body(Singletons,Bd1,Bd2).

singletons_vars_body(B,Sings):-
   free_variables(B,[],[],Vars), occurrences_of_vars(Vars,B,Sings).

occurrences_of_vars([V|Vars],B,[V|Sings]):-
   occurrences_of_var(V,B,1),!, occurrences_of_vars(Vars,B,Sings).
occurrences_of_vars([_|Vars],B,Sings):- occurrences_of_vars(Vars,B,Sings).
occurrences_of_vars([],_,[]).

apply_instantiate_clause(CN,[V|Vs],[T|Ts]) :- !,
   apply_instantiate_clause(CN,[V|Vs],[T|Ts],_).
apply_instantiate_clause(CN,[V|Vs],[T|Ts],NewCNs) :-
   clause_number_to_clause(CN,C),
   apply_instantiate_clause2(C,[V|Vs],[T|Ts],Clauses),
   add_clauses(NewCNs,Clauses,nil),
   update_programs((CN,(instantiation,[V|Vs],[T|Ts]), NewCNs), _), 
   (derivation_trace_on ->
        new_clauses(instantiation,CN,[V|Vs],[T|Ts],NewCNs);true).

apply_instantiate_clause2(C,[V],[T],Clauses) :- !,
   my_free_variables(C,L),
   pos1(Var,L,V),
   current_clauses(Cls),
   type_pattern(Cls,T,Patterns),
   apply_instantiate_clause_(C,Var,Patterns,Clauses).

apply_instantiate_clause2(C,Vs,Ts,Clauses) :- 
   not_ordered(Vs), !,
   reorder(Vs,Ts,Vs2,Ts2),
   apply_instantiate_clause2(C,Vs2,Ts2,Clauses).

apply_instantiate_clause2(C,[V|Vs],[T|Ts],Clauses) :- !,
   C = (Hd,Bd),
   my_free_variables([Hd|Bd],L),
   pos1(Var,L,V),
   current_clauses(Cls),
   type_pattern(Cls,T,Patterns),
   apply_instantiate_clause_(C,Var,Patterns,Clauses1),
   new_variable_positions(Vs,Patterns,LNewVs),
   apply_instantiate_clauses(Clauses1,LNewVs,Ts,Clauses).

apply_instantiate_clauses([],_,_,[]).
apply_instantiate_clauses([C|Cs],[LVs|LsVs],Ts,Clauses) :-
   apply_instantiate_clause2(C,LVs,Ts,Cls),
   apply_instantiate_clauses(Cs,LsVs,Ts,LCls),
   append(Cls,LCls,Clauses).

new_variable_positions(_,[],[]).
new_variable_positions(Vs,[P|Patterns],[NewVs|LNewVs]) :-
   free_variables(P,[],[],VarsP),
   length(VarsP,N),
   new_variable_positions1(Vs,N,NewVs),
   new_variable_positions(Vs,Patterns,LNewVs).

new_variable_positions1(L,1,L) :- !.
new_variable_positions1([],_,[]).
new_variable_positions1([V|Vs],N,[NewV|NewVs]) :-
   NewV is V + N - 1,
   new_variable_positions1(Vs,N,NewVs).


apply_instantiate_clause_one_var(CN,V,T) :-
   apply_instantiate_clause_one_var(CN,V,T,_).
apply_instantiate_clause_one_var(CN,V,T,NewCNs) :-
   clause_number_to_clause(CN,C),
   C = (Hd,Bd),
   my_free_variables([Hd|Bd],L),
   pos1(Var,L,V),
   current_clauses(Cls),
   type_pattern(Cls,T,Patterns),
   apply_instantiate_clause_(C,Var,Patterns,Clauses),
   add_clauses(NewCNs,Clauses,nil),
   update_programs((CN,(instantiation,V,T), NewCNs), _), 
   (derivation_trace_on -> new_clauses(instantiation,CN,V,T,NewCNs);true).

apply_instantiate_clause_(_,_,[],[]).
apply_instantiate_clause_(C,V,[P|Ps],[C|Cls]) :-
   copy_term((C,V), (C2,V2)),
   unify_with_occurs_check(V,P),
   apply_instantiate_clause_(C2,V2,Ps,Cls).

type_pattern([],_,[]).
type_pattern([N|Cls],T,[P|Patterns]) :-
   clause_number_to_clause(N,(Head,_)),
   predname_and_arity(Head,T,1), !,
   arg(1,Head,P),
   type_pattern(Cls,T,Patterns).
type_pattern([_|Cls],T,Patterns) :-
   type_pattern(Cls,T,Patterns).

my_all_free_variables(X,[X]) :- var(X), !.
my_all_free_variables(X,[]) :- ground(X), !.
my_all_free_variables(X,Vars) :-
   X =.. [_|L],
   my_all_free_variables_args(L,[],Vars).

my_all_free_variables_args([],Vars,Vars).
my_all_free_variables_args([T|Ts],Vars_i,Vars) :-
   my_all_free_variables(T,Vars_T),
   append(Vars_T,Vars_i,Vars2),
   my_all_free_variables_args(Ts,Vars2,Vars).

my_free_variables(X,[X]) :- var(X), !.
my_free_variables(X,[]) :- ground(X), !.
my_free_variables(X,Vars) :-
   X =.. [_|L],
   my_free_variables_args(L,[],Vars).

my_free_variables_args([],Vars,Vars).
my_free_variables_args([T|Ts],Vars_i,Vars) :-
   my_free_variables(T,Vars_T),
   add_new_variables(Vars_T,Vars_i,Vars2),
   my_free_variables_args(Ts,Vars2,Vars).

add_new_variables([],V,V).
add_new_variables([V|Vs],Vars, NewVars) :-
    belongs(V,Vars), !,
    add_new_variables(Vs,Vars, NewVars).
add_new_variables([V|Vs],Vars, NewVars) :-
    append(Vars,[V],Vars2),
    add_new_variables(Vs,Vars2, NewVars).



/*********

 lloyd_topor

**********/

lloyd_topor((Hd,Bd),Cs) :-
   lloyd_topor_bd(Hd,[],Bd,Cs).

lloyd_topor_bd(Hd,Bd_1,[],[(Hd,Bd_1)]):- !.

%lloyd_topor_bd(Hd,Bd_1,[\+(Phi_1,Phi_2)|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[\+(Phi_1)|Phi],C1),
%   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
%   (is_a_conjunction(Phi_2) ->
%    lloyd_topor_bd(_Hd,_Bd_1,[\+(and(_Phi_2))|_Phi],C2) ;
%    lloyd_topor_bd(_Hd,_Bd_1,[\+(_Phi_2)|_Phi],C2) ),
%   append(C1,C2,Cs).

%lloyd_topor_bd(Hd,[not(Phi_1,Phi_2)|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[not(Phi_1)|Phi],C1),
%   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
%   (is_a_conjunction(Phi_2) ->
%    lloyd_topor_bd(_Hd,_Bd_1,[not(and(_Phi_2))|_Phi],C2) ;
%    lloyd_topor_bd(_Hd,_Bd_1,[not(_Phi_2)|_Phi],C2) ),
%   append(C1,C2,Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(and(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[\+(Phi_1)|Phi],C1),
   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
 %  (is_a_conjunction(Phi_2) ->
%    lloyd_topor_bd(_Hd,_Bd_1,[\+(and(_Phi_2))|_Phi],C2) ;
    lloyd_topor_bd(_Hd,_Bd_1,[\+(_Phi_2)|_Phi],C2), % ),
   append(C1,C2,Cs).

lloyd_topor_bd(Hd,Bd_1,[not(and(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[not(Phi_1)|Phi],C1),
   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
%   (is_a_conjunction(Phi_2) ->
%    lloyd_topor_bd(_Hd,_Bd_1,[not(and(_Phi_2))|_Phi],C2) ;
    lloyd_topor_bd(_Hd,_Bd_1,[not(_Phi_2)|_Phi],C2), % ),
   append(C1,C2,Cs).

%lloyd_topor_bd(Hd,Bd_1,[\+(and(Phi_1))|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[\+(Phi_1)|Phi],Cs).

%lloyd_topor_bd(Hd,Bd_1,[not(and(Phi_1))|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[not(Phi_1)|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[a(L1,Phi_1)|Phi], Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[\+(e(L1,\+(Phi_1)))|Phi], Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(a(L1,Phi_1))|Phi], Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[e(L1,\+(Phi_1))|Phi], Cs).

lloyd_topor_bd(Hd,Bd_1,[not(a(L1,Phi_1))|Phi], Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[e(L1,not(Phi_1))|Phi], Cs).

lloyd_topor_bd(Hd,Bd_1,[->(Phi_1,Phi_2)|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[\+(Phi_1)|Phi],C1),
   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
   lloyd_topor_bd(_Hd,_Bd_1,[_Phi_2|_Phi],C2),
   append(C1,C2,Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(->(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1,\+(Phi_2)|Phi], Cs).

lloyd_topor_bd(Hd,Bd_1,[not(->(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1,not(Phi_2)|Phi],Cs).

%lloyd_topor_bd(Hd,Bd_1,[or(Phi_1,Phi_2,Phi_3)|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],C1),
%   copy_term((Hd,Bd_1,Phi_2,Phi_3,Phi),(_Hd,_Bd_1,_Phi_2,_Phi_3,_Phi)),
%   lloyd_topor_bd(_Hd,_Bd_1,[or(_Phi_2,_Phi_3)|_Phi],C2),
%   append(C1,C2,Cs).

lloyd_topor_bd(Hd,Bd_1,[or(Phi_1,Phi_2)|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],C1),
   copy_term((Hd,Bd_1,Phi_2,Phi),(_Hd,_Bd_1,_Phi_2,_Phi)),
   lloyd_topor_bd(_Hd,_Bd_1,[_Phi_2|_Phi],C2),
   append(C1,C2,Cs).

%lloyd_topor_bd(Hd,Bd_1,[or(Phi_1)|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).

%lloyd_topor_bd(Hd,Bd_1,[\+(or(Phi_1,Phi_2,Phi_3))|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd, Bd_1,[\+(Phi_1),\+(or(Phi_2,Phi_3))|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(or(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd, Bd_1,[\+(Phi_1),\+(Phi_2)|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(or(Phi_1))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd, Bd_1,[\+(Phi_1)|Phi],Cs).

%lloyd_topor_bd(Hd,Bd_1,[not(or(Phi_1,Phi_2,Phi_3))|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd, Bd_1,[not(Phi_1),not(or(Phi_2,Phi_3))|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[not(or(Phi_1,Phi_2))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd, Bd_1,[not(Phi_1),not(Phi_2)|Phi],Cs).

%lloyd_topor_bd(Hd,Bd_1,[not(or(Phi_1))|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[not(Phi_1)|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(\+(Phi_1))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).
lloyd_topor_bd(Hd,Bd_1,[\+(not(Phi_1))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).
lloyd_topor_bd(Hd,Bd_1,[not(\+(Phi_1))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).
lloyd_topor_bd(Hd,Bd_1,[not(not(Phi_1))|Phi],Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[and(Phi_1,Phi_2)|Phi],Cs) :- !,
%   (is_a_conjunction(Phi_2) ->
%    lloyd_topor_bd(Hd,Bd_1,[Phi_1,and(Phi_2)|Phi],Cs) ;
    lloyd_topor_bd(Hd,Bd_1,[Phi_1,Phi_2|Phi],Cs). % ).

%lloyd_topor_bd(Hd,Bd_1,[and(Phi_1)|Phi],Cs) :- !,
%   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs).

lloyd_topor_bd(Hd,Bd_1,[e(_,Phi_1)|Phi], Cs) :- !,
   lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi], Cs).

lloyd_topor_bd(Hd,Bd_1,[\+(e(L,Phi_1))|Phi],Cs) :- !,
   gen_new_name_functor(NewPred),
   vars(Phi_1,L,L2),
   A =.. [NewPred|L2],
   append(Bd_1,[\+(A)],NewBd_1),
   lloyd_topor_bd(Hd,NewBd_1,Phi,C1),
   copy_term((A,Phi_1),(_A,_Phi_1)),
   conj_to_list(_Phi_1,_LPhi_1),
   lloyd_topor_bd(_A,[],_LPhi_1,C2), append(C1,[C2],Cs).

lloyd_topor_bd(Hd,Bd_1,[not(e(L,Phi_1))|Phi], Cs) :- !,
   gen_new_name_functor(NewPred),
   vars(Phi_1,L,L2),
   A =.. [NewPred|L2],
   append(Bd_1,[not(A)],NewBd_1),
   lloyd_topor_bd(Hd,NewBd_1,Phi,C1),
   copy_term((A,Phi_1),(_A,_Phi_1)),
   conj_to_list(_Phi_1,_LPhi_1),
   lloyd_topor_bd(_A,[],_LPhi_1,C2), append(C1,[C2],Cs).

lloyd_topor_bd(Hd,Bd_1,[Phi_1|Phi],Cs) :- !,
   (is_a_conjunction(Phi_1) -> 
        (conj_to_list(Phi_1,L_Phi_1), append(L_Phi_1,Phi,L), 
	 lloyd_topor_bd(Hd,Bd_1,L,Cs) ) ;
   ((is_a_negative_literal(Phi_1,Conj), is_a_conjunction(Conj)) ->
    (conj_to_list(Conj,ConjList), NewConj=..[and|ConjList], NewNeg=..[\+,NewConj],
   lloyd_topor_bd(Hd,Bd_1,[NewNeg|Phi],Cs)) ;
   append(Bd_1,[Phi_1],NewBd_1),
   lloyd_topor_bd(Hd,NewBd_1,Phi,Cs))).




%%%%%%%%% FOR NEW TOP LEVEL PREDICATES

/*-----------
apply_r_substitution_clauses(L1,L2):
 L2 is the list L1 in which the clauses which are r_subsumed
 by another clause in L1 are deleted.
------------*/
apply_r_subsumption_clauses(L1,L2):- apply_r_subsumption_clauses(L1,[],L2).
apply_r_subsumption_clauses([],L,L).
apply_r_subsumption_clauses([E|L],I,F) :-
   is_r_subsumed(E,L),!, update_programs((E,delclause,[]),_),
   apply_r_subsumption_clauses(L,I,F).
apply_r_subsumption_clauses([E|L],I,F) :-
   is_r_subsumed(E,I),!, update_programs((E,delclause,[]),_),
   apply_r_subsumption_clauses(L,I,F).
apply_r_subsumption_clauses([E|L],Inter,F) :-
    append(Inter,[E],Inter2), apply_r_subsumption_clauses(L,Inter2,F).

is_r_subsumed(E,[F|_]):- is_r_subsumed_clause(E,F),!.
is_r_subsumed(E,[_|L]):- is_r_subsumed(E,L).
% N is r_subsumed by M
is_r_subsumed_clause(N,M):-
    clause_number_to_clause(N,(H1,B1)),
    clause_number_to_clause(M,(H2,B2)),
    permutation(B1,PermB1),
    (subsumes_chk( (H2,B2), (H1,PermB1)) ;
     (append(F,_,PermB1), subsumes_chk((H2,B2), (H1,F)))),
   write('clause '), write(N), write(' subsumed by '), write(M), nl.

