% determ_prog.pl
%
% This file contains the predicates which are used by the "determ"
% strategy which aims at making programs (semi) deterministic.
% The notion of semi-determinism uses the concept of modes.
% In order to apply this strategy on an initial program, the user
% is assumed to have given modes to all predicates that occur in it.
%
% The strategy is parametric wrt to subsidiary strategies which 
% give information about how to unfold, if and when subsumption must
% be applied, is and how case split must be applied. Possible parameters
% that might be chosen by the user at the moment are the following:
%
% apply_case_split(global_analysis)
% apply_case_split(binary) 
% apply_subsumption(during_case_split)
% apply_subsumption(after_case_split) 
% apply_simplification_eq(separate)
% after_first_step_unfold_leftmost_det
% after_first_step_unfold_leftmost_consumer
% after_first_step_unfold_leftmost_weak_consumer
% unfold_only_non_instantiated_head
% delete_clauses_with_failing_body
%
% Added 12/2/98
%folding_with_clause_permutation
%folding_with_body_rearrangement
%simple_folding
%generalize_output_in_bodies_before_new_def(all)
%generalize_output_in_bodies_before_new_def(only_ground)
%generalize_head_output_in_continuation

% Parameters must be set before the goal "determ_strat" is run. The setting of
% parameters is done in the file "determ_prog_parameters.pl".

:- dynamic current_new_def_functor/1,  initial_def/3.
:- use_module(library(terms), [ variant/2 ]).

/*-------
determ_strat (or simply determ) is the top level goal which applies the
strategy which transform an initial program into a semi-deterministic
specialized one, as described in:
     Pettorossi, A., Proietti, M., Renault, S.:
     Reducing Nondeterminism while Specializing Logic Programs
     24th ACM Symposium on Principles of Programming Languages (POPL'97)
     La Sorbonne, Paris, France, January 15-17, 1997, p. 414-427.
The initial program, the initial specialized clause, predicate modes and
other parameters are set in the file determ_strat_parameters.
--------*/
determ_strat :-
      statistics(runtime,[_,T1]),
      erase_environment2,
      consult(['src_sics/determ_strat_parameters.pl']),
      show_strat_parameters, %show_mode_predicates,
      file_input(Input), input_determ(Input), 
      init_environment, assert_initial_def(X),
      (determ_def(X) -> (
           (max_depth(stop) -> fail
           ; (nl, write('Final program : '), nl, nl,
           pp, retract_parameters_strat)))
      ; (retract_parameters_strat, msg_determ_strat_fails)),
      statistics(runtime,[_,T2]), nl,
      T is T2-T1, write('Elapsed time '), write(T), write(ms), nl.

introduce_continuation_in_clause_if_necessary((H,B),(H2,B2)) :-
	functor(H,F,N),	functor(H2,F,N),
	H=..[_|Args],
	mode(F,N,Modes), 
	introduce_continuation_in_clause(Args,Modes,NewArgs,Cont),
        (Cont=true -> (H,B)=(H2,B2) ;
	    (Gen=..['@',Cont]),
    	    H2=..[_|NewArgs],
	    append(B,[Gen],B2)).

introduce_continuation_in_clause((H,B),(H2,B2)) :-
	functor(H,F,N),	functor(H2,F,N),
	H=..[_|Args],
	mode(F,N,Modes), 
	introduce_continuation_in_clause(Args,Modes,NewArgs,Cont),
%        (Cont=[] -> Gen=..['@',true] ; Gen=..['@',Cont]),
        Gen=..['@',Cont],
	H2=..[_|NewArgs],
	append(B,[Gen],B2).
	
introduce_continuation_in_clause([],[],[],true).
introduce_continuation_in_clause([A|Args],[+|Modes],[A|NewArgs],Cont) :- !,
       introduce_continuation_in_clause(Args,Modes,NewArgs,Cont).
introduce_continuation_in_clause([A|Args],[_|Modes],[A|NewArgs],Cont) :-
	var(A), !,
       introduce_continuation_in_clause(Args,Modes,NewArgs,Cont).
introduce_continuation_in_clause([A|Args],[_|Modes],[X|NewArgs],Cont2) :- %[X=A|Cont]) :-
       introduce_continuation_in_clause(Args,Modes,NewArgs,Cont), append_conjunction(X=A,Cont,Cont2).

/*---------------
determ_def(Clauses) tries to derive from Clauses a new program
where heads of clauses are mutually exclusive, w.r.t some
input arguments that are given with the initial program or
infer at run time, for the new definitions.
When this goal terminates, it gives a new program
equivalent to Clauses. It might not terminate, due to a non
termination in the unfolding strategy, or an introduction
of a new definition an infinite number of times.
----------------*/
determ_def(_) :-
    max_depth(stop), !.
determ_def([]) :- !.
determ_def(Cl0) :-
    unfolding_steps(Cl0,Cl_0), check_trace_session(unf),
    (strat(apply_subsumption(after_unfolding)) -> 
     apply_subsumption_clauses(Cl_0,Cl1_0) ; Cl1_0=Cl_0),
     omit_unit_clauses(Cl1_0,Cl1),
     (strat(apply_case_split(global_analysis)) -> 
     apply_case_split_global(Cl1,P1,_) ; 
     (strat(apply_case_split(binary)) -> 
     apply_case_split_binary(Cl1,P1,_) ; 
     (strat(simple_folding) -> %P1=[Cl1])),  changed 27/02/98
	       packets_of_one_clause(Cl1,P1) ;       
               split_into_packets_of_mut_excl_heads(Cl1,P1)))),
     check_trace_session(cs),
     (strat(apply_simplification_eq(separate)) ->
     apply_simplification_eq_on_packets(P1,P2) ; P1=P2),
     (strat(apply_subsumption(after_case_split)) -> 
     apply_subsumption_on_packets(P2,P) ; P2=P),
     folding_steps_on_packets(P,NewDefs), check_trace_session(fold),
     increment_depth, determ_defs(NewDefs).

determ_defs([]).
determ_defs([NewDef|NewDefs]) :-
     determ_def(NewDef),!, 
     determ_defs(NewDefs).

omit_unit_clauses([],[]).
omit_unit_clauses([C|Cls],Cls2) :-
     is_a_unit_clause(C), !, omit_unit_clauses(Cls,Cls2).
omit_unit_clauses([C|Cls],Cls2) :-
     continuation_body(C), !,
     fold_conjunction_continuations([C],_), omit_unit_clauses(Cls,Cls2).
omit_unit_clauses([C|Cls],[C|Cls2]) :- omit_unit_clauses(Cls,Cls2).

folding_steps_on_packets([Packet|Packets],[NewDef|NewDefs]) :-
    remove_clauses_non_subject_to_folding(Packet,Packet2),
    (Packet2=[] -> NewDef=[] ; folding_one_packet(Packet2,NewDef)),
     folding_steps_on_packets(Packets,NewDefs).
folding_steps_on_packets([],[]).

% Removal of clauses with empty bodies or bodies made of
% inequalities only.
remove_clauses_non_subject_to_folding([],[]).
%remove_clauses_non_subject_to_folding([E|L],L2):-
%	is_a_unit_clause(E),	!,
%        remove_clauses_non_subject_to_folding(L,L2).
remove_clauses_non_subject_to_folding([E|L],L2) :-
	body_of_clause(E,B), body_without_inequalities(B,_,B2),
	(B2==[] -> true
	;   (is_a_list_of_continuations(B2) -> fold_conjunction_continuations([E],_))), !,
        remove_clauses_non_subject_to_folding(L,L2).
remove_clauses_non_subject_to_folding([E|L],[E|L2]) :-
        remove_clauses_non_subject_to_folding(L,L2).

folding_one_packet(Packet2,NewDef) :-
    (strat(generalize_head_output_in_continuation) ->
               (fold_conjunction_continuations(Packet2,Packet_i),
        	fold_disjunction_continuations(Packet_i,Packet))
                    ; Packet2=Packet),
    ( try_to_fold_with_defs(Packet) -> NewDef=[] ;
	introduce_new_def_and_fold(Packet,NewDef)).

/*----------------------- FOLDING STRATEGY ---------------------------*/

fold_disjunction_continuations([],[]).
fold_disjunction_continuations([C],[C]) :- !.
fold_disjunction_continuations(Cls,Cls2) :-
     member_rest(C,Cls,Rest),
     member_rest(C2,Rest,Rest2),
     fold_disjunction_continuations_(C,C2,C3),!,
     add_clauses([NewCN],[C3],nil),
     update_programs(([C,C2],fold_continuations,[NewCN]), _), 
     (derivation_trace_on ->
        new_clauses(fold_continuations,[C,C2],[NewCN]);true),
        %changed by Maurizio 3 Oct 2002 to switch derivation trace on/off
fold_disjunction_continuations([NewCN|Rest2],Cls2).
fold_disjunction_continuations(Cls,Cls).

fold_disjunction_continuations_(C,C2,(H,B3)) :-
	clause_number_to_clause(C,(H,B_Cont)),
	clause_number_to_clause(C2,(H2,B2_Cont)),
        split_body_continuation(B_Cont,B,Cont),
        split_body_continuation(B2_Cont,B2,Cont2),
%	variant(B,B2),
	variant((H,B),(H2,B2)),
	(H,B)=(H2,B2), make_disjunction(Cont,Cont2,Cont3),
	append(B,['@'(Cont3)],B3).
	
fold_conjunction_continuations([],[]).
fold_conjunction_continuations([C|Cs],[C2|Cs2]) :-
     fold_conjunction_continuations_cl(C,C2),
     fold_conjunction_continuations(Cs,Cs2).

fold_conjunction_continuations_cl(C,NewCN) :-
	clause_number_to_clause(C,(HC,BC)),
	introduce_continuation_in_clause_if_necessary((HC,BC),(H,B)),
	fold_conjunction_continuations_body_cl(B,B2,Changed),
	(Changed==changed -> 
	        (add_clauses([NewCN],[(H,B2)],nil),
        	update_programs((C,fold_continuations,[NewCN]), _), 
                (derivation_trace_on ->
                 new_clauses(fold_continuations,C,[NewCN]);true)) ; NewCN=C).

fold_conjunction_continuations_body_cl([],[],_).
%fold_conjunction_continuations_body_cl(['@'(G1),'@'(G2)|As],['@'(NewG)],changed) :- !,
%	fold_conjunction_continuations_body_cl(['@'(G2)|As],['@'(G3)],_),
%	append_conjunction(G1,G3,NewG).
fold_conjunction_continuations_body_cl(['@'(G1),'@'(G2)|As],B,changed) :- !,
	append_conjunction(G1,G2,G),
	fold_conjunction_continuations_body_cl(['@'(G)|As],B,_).
fold_conjunction_continuations_body_cl(['@'(G1),L|As],[L|Bs],X) :- !,
	fold_conjunction_continuations_body_cl(['@'(G1)|As],Bs,X).
fold_conjunction_continuations_body_cl([A|As],[A|Bs],X) :-
        fold_conjunction_continuations_body_cl(As,Bs,X).


try_to_fold_with_defs(Clauses) :-
        same_length(LDefs,Clauses),
	node(Pi,_,_,(gen_def,LDefs), _), 
%        length(LDefs,N2), N2 >= N,
% Modified for optimization 13/2/98	
%        length(LDefs,N2), N2 = N,
	(strat(folding_with_clause_permutation) ->
	    (permutation(Clauses,PermClauses),
	    try_to_fold_with_defs_whole_bd(Pi,PermClauses,LDefs));
            try_to_fold_with_defs_whole_bd(Pi,Clauses,LDefs)).

try_to_fold_with_defs(Clauses) :-
	strat(folding_partial_body), !,
        try_to_fold_with_defs_partial_bd(Clauses,Clauses,[]).

try_to_fold_with_defs_partial_bd(ClausesIni,Clauses,DefsI) :-
        same_length(LDefs,Clauses),
	node(_,_,_,(gen_def,LDefs), _), 
        try_to_fold_with_defs_partial_bd(ClausesIni,Clauses,DefsI,LDefs).


try_to_fold_with_defs_partial_bd(ClausesIni,Clauses,DefsI,Defs):-
        heads_unifiable(Defs,_,_), 
	general_fold_cond2(Clauses,Defs,_,_,NewClause),
        add_clauses(NewCN,[NewClause],nil), 
	append(DefsI,Defs,Defs2),
	(completely_folded(NewClause) -> 
	    (update_programs((ClausesIni,(fold, Defs2),NewCN), _), 
	    (derivation_trace_on ->
                new_clauses(fold,ClausesIni,Defs2,NewCN) ; true))
	    ;
	    try_to_fold_with_defs_partial_bd(ClausesIni,NewCN,Defs2)).


try_to_fold_with_defs_whole_bd(Pi,Clauses,Defs) :-
	heads_unifiable(Defs,_,A), 
        (strat(folding_with_body_rearrangement) ->
	    	    general_fold_cond2_whole_bd(Clauses,Defs,_,_,NewClause) ;
    	    fold_cond2_whole_bd(Clauses,Defs,_,_,NewClause)),
        definition_pred_used_for_folding(Defs,A),
        (derivation_trace_on -> msg_user_successful_folding(Pi,A);true),
	add_clauses(NewCN,[NewClause],nil), %NewCN=[X],
	update_programs((Clauses,(fold, Defs),NewCN), _), 
        (derivation_trace_on -> new_clauses(fold,Clauses,Defs,NewCN);true).

% Added the 12/2/98
% Continuation arguments of the bodies of the clauses to be folded
% are generalized to a new variable

new_defs_general_case3([],_,_,[]).
new_defs_general_case3([(H,B1)],X,L2,[(NewH,B2)]) :-!,
	split_body_continuation(B1,B,_),
	free_variables(B,[],[],VarsB),
	new_defs_general_case3_body(B1,VarsB,B2,L_Var), H =..[_|L1],
	new_defs_general_case3_args(L1,L2,L3),
	append(L3,L_Var,L4), NewH =..[X|L4].

new_defs_general_case3(LC1,X,L2,LC2) :-
	new_defs_general_case3_bodies(LC1,L_BDs2,LGenVars),
	new_defs_general_case3_heads(L_BDs2,X,L2,LGenVars,LC2).

new_defs_general_case3_bodies([],[],[]).
new_defs_general_case3_bodies([(H,B1)|LL1],[(H,B2)|LL2],LGenVars) :-
	split_body_continuation(B1,B,_),
	% changed 17/02/98
%	free_variables(B,[],[],VarsB),
	free_variables(B,H,[],VarsB),
        new_defs_general_case3_body(B1,VarsB,B2,LGenVarsB2),
	new_defs_general_case3_bodies(LL1,LL2,LGenVars2),
	append(LGenVarsB2,LGenVars2,LGenVars).
	
new_defs_general_case3_heads([],_,_,_,[]).
new_defs_general_case3_heads([(H,B)|LL1],X,L2,LGenVars,[(NewH,B)|LL2]) :-
	H =..[_|L1],
	new_defs_general_case3_args(L1,L2,L3),
	append(L3,LGenVars,L4),
	NewH =..[X|L4], new_defs_general_case3_heads(LL1,X,L2,LGenVars,LL2).

new_defs_general_case3_args([],[],[]).
new_defs_general_case3_args([A1|L1],[A2|L2], [A1|L3]) :-
	var(A2), !, new_defs_general_case3_args(L1, L2, L3).
new_defs_general_case3_args([_|L1],[fv|L2], L3) :-
 	new_defs_general_case3_args(L1, L2, L3).
new_defs_general_case3_args([A1|L1],[A2|L2], L3) :-
	 A1=..[F|LL1],  A2=..[F|LL2],
	append(LL1,L1,NewL1), append(LL2,L2,NewL2),
	new_defs_general_case3_args(NewL1, NewL2, L3).

new_defs_general_case3_body([],_,[],[]).
new_defs_general_case3_body([Lit|B],VarsB,B2,L_Var):-
        (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
	functor(Atom,=,_), !,
	new_defs_general_case3_body(B,VarsB,B2,L_Var).
new_defs_general_case3_body([E],VarsB,['@'(GenArgE)],[GenArgE|V]):-
        is_a_continuation(E),
	free_variables(E,[],[],VarsE),
	intersection_vars(VarsE,VarsB,V).
new_defs_general_case3_body([E|B],V,[E|B2],L) :-
	new_defs_general_case3_body(B,V,B2,L).
	

% Added the 12/2/98
% all or only_ground arguments of the bodies of the clauses to be folded
% are generalized to a new variable.

new_defs_general_case2([],_,_,[]).
new_defs_general_case2([(H,B1)],X,L2,[(NewH,B2)]) :-!,
	new_defs_general_case2_body(B1,B2,L_Var), H =..[_|L1],
	new_defs_general_case2_args(L1,L2,L3),
	append(L3,L_Var,L4), NewH =..[X|L4].

new_defs_general_case2(LC1,X,L2,LC2) :-
	new_defs_general_case2_bodies(LC1,L_BDs2,LGenVars),
	new_defs_general_case2_heads(L_BDs2,X,L2,LGenVars,LC2).

new_defs_general_case2_bodies([],[],[]).
new_defs_general_case2_bodies([(H,B1)|LL1],[(H,B2)|LL2],LGenVars) :-
        new_defs_general_case2_body(B1,B2,LGenVarsB2),
	new_defs_general_case2_bodies(LL1,LL2,LGenVars2),
	append(LGenVarsB2,LGenVars2,LGenVars).
	
new_defs_general_case2_heads([],_,_,_,[]).
new_defs_general_case2_heads([(H,B)|LL1],X,L2,LGenVars,[(NewH,B)|LL2]) :-
	H =..[_|L1],
	new_defs_general_case2_args(L1,L2,L3),
	append(L3,LGenVars,L4),
	NewH =..[X|L4], new_defs_general_case2_heads(LL1,X,L2,LGenVars,LL2).

new_defs_general_case2_args([],[],[]).
new_defs_general_case2_args([A1|L1],[A2|L2], [A1|L3]) :-
	var(A2), !, new_defs_general_case2_args(L1, L2, L3).
new_defs_general_case2_args([_|L1],[fv|L2], L3) :-
 	new_defs_general_case2_args(L1, L2, L3).
new_defs_general_case2_args([A1|L1],[A2|L2], L3) :-
	 A1=..[F|LL1],  A2=..[F|LL2],
	append(LL1,L1,NewL1), append(LL2,L2,NewL2),
	new_defs_general_case2_args(NewL1, NewL2, L3).

new_defs_general_case2_body(B,B2,L_var) :-
	(strat(generalize_output_in_bodies_before_new_def(all)) ->
	    vars_to_be_generalized(B,VarsG) ; VarsG=[]),
	new_defs_general_case2_body_(B,B2,VarsG,L_var).

new_defs_general_case2_body_([],[],_,[]).
new_defs_general_case2_body_([Lit|B],B2,VarsG,L_Var):-
        (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
	functor(Atom,=,_), !,
	new_defs_general_case2_body_(B,B2,VarsG,L_Var).
new_defs_general_case2_body_([E|B],[GenE|B2],VarsG,L_Var):-
        (is_a_negative_literal(E,Atom) -> true ; E=Atom),
	Atom=..[F|L_ArgsE], same_length(L_ArgsE,LModes),
	mode(F,_,LModes),
	generalize_output_args(LModes,L_ArgsE,VarsG,L_GenArgsE,L_VarGenE),
	Gen_Atom=..[F|L_GenArgsE],
	(is_a_negative_literal(E,_) -> GenE=(\+(Gen_Atom)) ; GenE=Gen_Atom),
	new_defs_general_case2_body_(B,B2,VarsG,L_Var2),
	append(L_VarGenE,L_Var2,L_Var).
generalize_output_args([],[],_,[],[]).
generalize_output_args([+|LModes],[A1|As],VarsG,[A1|GenAs],L) :- !,
     generalize_output_args(LModes,As,VarsG,GenAs,L).
generalize_output_args([-|LModes],[A1|As],VarsG,[GenA1|GenAs],[GenA1|L]) :-
     strat(generalize_output_in_bodies_before_new_def(all)),
     (nonvar(A1) ; (var(A1), belongs(A1,VarsG))), !,
     generalize_output_args(LModes,As,VarsG,GenAs,L).
generalize_output_args([-|LModes],[A1|As],VarsG,[A1|GenAs],L) :-
     strat(generalize_output_in_bodies_before_new_def(all)), !,
     generalize_output_args(LModes,As,VarsG,GenAs,L).
generalize_output_args([-|LModes],[A1|As],VarsG,[GenA1|GenAs],[GenA1|L]) :-
     strat(generalize_output_in_bodies_before_new_def(only_ground)),
     ground(A1),!,
     generalize_output_args(LModes,As,VarsG,GenAs,L).
generalize_output_args([-|LModes],[A1|As],VarsG,[A1|GenAs],L) :-
     strat(generalize_output_in_bodies_before_new_def(only_ground)),
     generalize_output_args(LModes,As,VarsG,GenAs,L).

introduce_new_def_and_fold(Clauses,NewDefsN) :-
	gen_new_name_functor(X), anti_unify_heads(Clauses,Head,U),
	Head =.. [_|L2],
       (strat(generalize_head_output_in_continuation) ->
   	        new_defs_general_case3(U,X,L2,NewDefs2) ;
	(strat(generalize_output_in_bodies_before_new_def(_)) ->
	        new_defs_general_case2(U,X,L2,NewDefs2) ;
	    	        new_defs_general_case(U,X,L2,NewDefs2))),
        rearrange_definitions(NewDefs2,NewDefs3),
        try_to_drop_arguments(NewDefs3, NewDefs),
	(derivation_trace_on ->
                (nl, write('Introduction of definition '),
                write(X), write(':'));true),
        apply_gen_define(NewDefs,NewDefsN),
	current_prog(PN,_),
        assert(initial_def(X,PN,NewDefsN)),
	f(Clauses,NewDefsN),
        (needs_modes -> assert_modes_new_def(NewDefsN) ; true), !.

new_defs_general_case([],_,_,[]).
new_defs_general_case([(H,B1)],X,L2,[(NewH,B2)]) :-!,
	new_defs_general_case_body(B1,B2), H =..[_|L1],
	new_defs_general_case_args(L1,L2,L3), NewH =..[X|L3].
new_defs_general_case([(H,B1)|LL1],X,L2,[(NewH,B2)|LL2]) :-
	new_defs_general_case_body(B1,B2), H =..[_|L1],
	new_defs_general_case_args(L1,L2,L3),
	NewH =..[X|L3], new_defs_general_case(LL1,X,L2,LL2).
new_defs_general_case_args([],[],[]).
new_defs_general_case_args([A1|L1],[A2|L2], [A1|L3]) :-
	var(A2), !, new_defs_general_case_args(L1, L2, L3).
new_defs_general_case_args([_|L1],[fv|L2], L3) :-
 	new_defs_general_case_args(L1, L2, L3).
new_defs_general_case_args([A1|L1],[A2|L2], L3) :-
	 A1=..[F|LL1],  A2=..[F|LL2],
	append(LL1,L1,NewL1), append(LL2,L2,NewL2),
	new_defs_general_case_args(NewL1, NewL2, L3).
new_defs_general_case_body([],[]).
%new_defs_general_case_body([E|B],B2):-
%	(functor(E, \+,_);functor(E,=,_)), !,
%	new_defs_general_case_body(B,B2).
new_defs_general_case_body([Lit|B],B2):-
        (is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
	functor(Atom,=,_), !,
	new_defs_general_case_body(B,B2).
new_defs_general_case_body([E|B],[E|B2]):-
	new_defs_general_case_body(B,B2).

/*-------------------- UNFOLDING STRATEGY --------------------*/

unfolding_steps(Cls1,Cls) :-
     unfold_leftmost_atom(Cls1,Cls2),    % 1st step always leftmost
     unfolding_after_first_step(Cls2,Cls).

unfold_leftmost_atom(Cls1,Cls) :-
     unfold_leftmost_atom2(Cls1,Cls2), !, append(Cls2,Cls).

unfold_leftmost_atom2([E|Cls1],[NewCNs|Cls2]) :-
     unfold_clause_leftmost_atom2(E,NewCNs),
     unfold_leftmost_atom2(Cls1,Cls2).
unfold_leftmost_atom2([],[]).
unfold_clause_leftmost_atom2(E,[]) :-
    body_of_clause(E,R), strat(delete_clauses_with_failing_body),
    \+ non_failed_body(R), !,
    apply_delete_clause(E).
unfold_clause_leftmost_atom2(E,NewCNs) :-
    	suitable_prog_to_unfold(E,PN),
        bagof((Head,Body),  one_unf_i(PN,E,1,(Head,Body)),Clauses),
	add_clauses(NewCNs,Clauses,nil),
    	update_programs((E,(uunf_i,PN,1),NewCNs),_), 
	(derivation_trace_on ->new_clauses(unfold_d_p,E,1,NewCNs);true), !.
unfold_clause_leftmost_atom2(E,_) :-
        msg_failure_on_unfolding(E), nltab,
        (yesno('Do you want to apply clause deletion ?', n) ->
	apply_delete_clause(E) ; msg_derivation_fails(E)).

suitable_prog_to_unfold(E,PN):-
	body_of_clause(E,[B1|_]), predname_and_arity(B1,F,A),
	predname_and_arity(N2,F,A),
	(pred_in_initial_prog(N2) -> PN= 1 ; initial_def(F,PN,_)).

unfolding_after_first_step([],[]).
% Do not unfold unit-clauses
unfolding_after_first_step([E|L_i],[E|L_f]) :-
    is_a_unit_clause(E),!, unfolding_after_first_step(L_i,L_f).
% Do not unfold continuations
unfolding_after_first_step([E|L_i],[E|L_f]) :-
    continuation_body(E), !, unfolding_after_first_step(L_i,L_f).
unfolding_after_first_step([C|L_i], L_f) :-
    unfolding_clause_after_first_step(C,LC), !,
    unfolding_after_first_step(L_i, L), append(LC,L,L_f).

unfolding_clause_after_first_step(C,Cls) :-
    body_of_clause(C,R),  \+ non_failed_body(R), !,
    (strat(delete_clauses_with_failing_body) ->    
         (apply_delete_clause(C), Cls =[]) ;
        (msg_failure_on_unfolding(C), nltab,    
        yesno('Do you want to apply clause deletion ?', n) ->
       (apply_delete_clause(C), Cls=[]) ;  Cls=[C])).
unfolding_clause_after_first_step(C,[C]) :-
     head_of_clause(C,H), strat(unfold_only_non_instantiated_head),
     \+ non_instantiated_head(H), !.
unfolding_clause_after_first_step(C,LC_f) :-
    strat(after_first_step_unfold_leftmost_det), !,
    (unfold_leftmost_det(C,LC) ->
     unfolding_after_first_step(LC, LC_f) ; LC_f=[C]).
unfolding_clause_after_first_step(C,LC_f) :-
    strat(after_first_step_unfold_leftmost_consumer), !,
    (unfold_leftmost_cons(C,LC) ->
    unfolding_after_first_step(LC, LC_f) ; LC_f=[C]).
unfolding_clause_after_first_step(C,LC_f) :-
    strat(after_first_step_unfold_leftmost_weak_consumer), !,
    (unfold_leftmost_weak_cons(C,LC) ->
     unfolding_after_first_step(LC, LC_f) ; LC_f=[C]).
unfolding_clause_after_first_step(C,[C]).
    
unfold_leftmost_det(E,NewCNs) :-
	suitable_prog_to_unfold_leftmost_det(E,Pos,PN),
       	bagof((Head,Body),  one_unf_i(PN,E,Pos,(Head,Body)),Clauses),
	Clauses=[_],!, add_clauses(NewCNs,Clauses,nil),
    	update_programs((E,(uunf_i,PN,Pos),NewCNs),_), 
	(derivation_trace_on -> new_clauses(unfold_d_p,E,Pos,NewCNs);true).

suitable_prog_to_unfold_leftmost_det(E,Pos,PN) :-
        body_of_clause(E,Bd), leftmost_det_unfolding(Bd,1,Pos,PN).

leftmost_det_unfolding([B1|_],Pos,Pos,PN):-
        predname_and_arity(B1,F,A), predname_and_arity(N2,F,A),
	(pred_in_initial_prog(N2) -> PN= 1 ; initial_def(F,PN,_)).
leftmost_det_unfolding([_|Bd],Pos,Pos2,PN):-
	Posi is Pos + 1, leftmost_det_unfolding(Bd,Posi,Pos2,PN).

apply_subsumption_on_packets([],[]).
apply_subsumption_on_packets([P|Ps],[NewP|NewPs]) :-
    apply_subsumption_clauses(P,NewP), apply_subsumption_on_packets(Ps,NewPs).

apply_simplification_eq_on_packets([],[]).
apply_simplification_eq_on_packets([P|Ps],[NewP|NewPs]) :-
    apply_simplification_eq_on_packet(P,NewP),!,
    apply_simplification_eq_on_packets(Ps,NewPs).
apply_simplification_eq_on_packet([],[]).
apply_simplification_eq_on_packet([C1|C],C2) :-
    clause_number_to_clause(C1, (H,B)),
    apply_unfold_d_eq_(C1, (H,B), removed(_)), !,
    apply_simplification_eq_on_packet(C,C2).
apply_simplification_eq_on_packet([C1|C],[C3|C2]) :-
    clause_number_to_clause(C1, (H,B)),
    apply_unfold_d_eq_(C1, (H,B), C3),
    apply_simplification_eq_on_packet(C,C2).

unfold_leftmost_cons(E,NewCNs) :-
        suitable_prog_to_unfold_leftmost_consumer(E,Pos,PN),
       	(bagof((Head,Body),  one_unf_i(PN,E,Pos,(Head,Body)),Clauses) ->
	(add_clauses(NewCNs,Clauses,nil),
    	update_programs((E,(uunf_i,PN,Pos),NewCNs),_), 
	(derivation_trace_on ->new_clauses(unfold_d_p,E,Pos,NewCNs);true)) ;
        apply_delete_clause(E)).

unfold_leftmost_weak_cons(E,NewCNs) :-
        suitable_prog_to_unfold_leftmost_weak_consumer(E,Pos,PN),
       	(bagof((Head,Body),  one_unf_i(PN,E,Pos,(Head,Body)),Clauses) ->
	(add_clauses(NewCNs,Clauses,nil),
    	update_programs((E,(uunf_i,PN,Pos),NewCNs),_), 
	(derivation_trace_on -> new_clauses(unfold_d_p,E,Pos,NewCNs);true)) ;
        apply_delete_clause(E)).

suitable_prog_to_unfold_leftmost_consumer(E,Pos,PN) :-
        body_of_clause(E,Bd), leftmost_consumer_unfolding(Bd,1,Pos,PN).
suitable_prog_to_unfold_leftmost_weak_consumer(E,Pos,PN) :-
        body_of_clause(E,Bd), leftmost_weak_consumer_unfolding(Bd,1,Pos,PN).

leftmost_consumer_unfolding([B1|_],Pos,Pos,PN):-
        functor(B1,F,A), \+ F =(\+), mode(F,A,Modes),
        consumer(B1,Modes), functor(N2,F,A),
        (pred_in_initial_prog(N2) -> PN= 1 ; initial_def(F,PN,_)).
leftmost_consumer_unfolding([_|Bd],Pos,Pos2,PN):-
        Posi is Pos + 1,
        leftmost_consumer_unfolding(Bd,Posi,Pos2,PN).

leftmost_weak_consumer_unfolding([B1|_],Pos,Pos,PN):-
        functor(B1,F,A), \+ F =(\+), mode(F,A,Modes),
        weak_consumer(B1,Modes), functor(N2,F,A),
        (pred_in_initial_prog(N2) -> PN= 1 ; initial_def(F,PN,_)).
leftmost_weak_consumer_unfolding([_|Bd],Pos,Pos2,PN):-
        Posi is Pos + 1,
        leftmost_weak_consumer_unfolding(Bd,Posi,Pos2,PN).

consumer(_,[]).
consumer(B1,Modes) :- B1 =.. [_|Args], consumer_args(Args,Modes).
weak_consumer(_,[]).
weak_consumer(B1,Modes) :- B1 =.. [_|Args], weak_consumer_args(Args,Modes).

consumer_args([],[]).
consumer_args([A-_|L],[M|L2]) :- !,
        (M== '+' -> nonvar(A) ; true), consumer_args(L,L2).
 consumer_args([A|L],[M|L2]) :-
        (M== '+' -> nonvar(A) ; true), consumer_args(L,L2).

weak_consumer_args([A|_],[X|_]) :- X=='+', nonvar(A).
weak_consumer_args([_|L],[_|L2]) :- weak_consumer_args(L,L2).
 
non_instanciated_head(Hd):- 
	functor(Hd,F,N), functor(Hd2,F,N), variant(Hd,Hd2).
non_instanciated_head2(Hd):-
	functor(Hd,F,N), mode(F,N,Modes),
	Hd=..[_|ArgsF], restrict_arg_to_input(ArgsF,Modes,NewArgsF),
        non_instanciated_head_input_args2(NewArgsF).
non_instanciated_head_input_args2([]).
non_instanciated_head_input_args2([A|L]) :-
    non_instanciated_head_input_args2(L),
    (var(A) ; ( A = [B|C], var(B), var(C))).

get_arity_new_def([(H,_)|_], A) :- functor(H,_,A).

try_to_drop_arguments(Defs, NewDefs) :-
   get_arity_new_def(Defs,Arity),
   try_to_drop_arguments(Defs, Arity,NewDefs).
try_to_drop_arguments(Defs, Arity, NewDefs) :-
   useless_arguments(Defs,Arity,Args), drop_arguments(Defs, Args, NewDefs).

useless_arguments(_,0,[]) :- !.
useless_arguments(Defs,Arity,[Arity|Args]) :-
   is_a_useless_argument(Defs,Arity), !, X is Arity - 1,
   useless_arguments(Defs,X,Args).
useless_arguments(Defs,Arity,Args) :-
   X is Arity - 1, useless_arguments(Defs,X,Args).

% corrected on the 19/2/98
is_a_useless_argument([D],Arity) :- !,
   D = (H,B), H =.. [_|Args],
   del_4(Var,Arity,Args,NewArgs), var(Var),
%   free_of_var(Var,B). % changed 6/3/98
   (free_of_var(Var,B) ; (pos(Var,NewArgs,OtherOcc), OtherOcc<Arity)).

is_a_useless_argument(Defs,A) :- 	is_a_useless_argument2(Defs,A).

is_a_useless_argument2([],_).
is_a_useless_argument2([D|Defs],Arity) :-
   D = (H,B), H =.. [_|Args],
   del_4(Var,Arity,Args,NewArgs), var(Var),
   free_of_var(Var,(NewArgs,B)), is_a_useless_argument2(Defs,Arity).



drop_arguments(Defs, [], Defs).
drop_arguments(Defs, [A|L], NewDefs) :-
   drop_argument(Defs, A, NewDefs1),  % must be in decreasing order
   drop_arguments(NewDefs1, L, NewDefs).
drop_argument([],_,[]).
drop_argument([D|Defs], A, [NewD|NewDefs]) :-
   D = (H,B), H =.. [F|Args], del_4(A,Args,NewArgs),
   NewH =.. [F|NewArgs], NewD = (NewH,B),
   drop_argument(Defs, A, NewDefs).




