% environment.pl
%
% This file contains the predicates that are responsible for the updating of the derivation
% tree when a transformation step is performed, and of other information necessary for
% performing further transformation steps.
% The tree is represented as a set of facts of the form: node(Node,Father,Sons,LA,LN), where:
% Node is positive integer identifying a node, Father is not negative integer identifying
% the father of Node, Sons is a list of positive integers identifying the sons of Node,
% LA and LN are respectively the label of the arc to the node Node and the label of Node.

:- use_module(library(lists), [ append/3 ]).

/*-----------
apply_start(P): initialize a new transformation sequence with P as initial program
-----------*/
apply_start(P) :-
    erase_environment, assert(file_input(P)), see(P), read_list(1), seen,
    show_initial_program, init_environment.

/*-------------
erase_environment: Empty current environment.
init_environment: Initialize a transformation tree.
These predicates are used by the start and load_session at top_level.
--------------*/
erase_environment :-
   retractall(file_input(_)), retractall(new_file_input(_)), 
   retractall(node(_,_,_,_,_)),	retractall(maxnode(_)), retractall(max_law(_)),
   retractall(law(_,_)), retractall(max_clause(_)),
   retractall(type_definition(_,_,_)), 
   retractall(current_new_def_functor(_)), 
   retractall(number_clauses(_)),
   retractall(mode_definition(_,_,_)),    retractall(mode(_,_,_)),
   retractall(clause3(_,_,_)), retractall(current_prog(_,_)),
   retractall(clauses_of_new_definition(_,_)).
init_environment :-
    setof(N,C^clause3(N,C,prog),LN), length(LN,L), assert(max_clause(L)), assert(max_law(0)),
    assert(current_prog(1,LN)), asserta(node(1,0,[],0,LN)), asserta(maxnode(1)).

/*-------------
apply_set_current_program(NP): Set current program to NP. This predicate is used 
by set_current_program at top level.
------------*/
apply_set_current_program(NP):-
   label_node(NP,L), current_prog(N,NCs), retract(current_prog(N,NCs)), assert(current_prog(NP,L)).

/*-------------
apply_add_law(Law): Add a new replacement law Law in the current theory.
------------*/
apply_add_law(Law) :-
    Law=(Left->Right), conj_to_list((Left),Left_List), conj_to_list((Right),Right_List),
    (clause(max_law(_),_) -> max_law(N), retract(max_law(N))  ; N=0),
    M is N+1, assert(law(M,(Left_List,Right_List))), assert(max_law(M)).

apply_add_laws([]).
apply_add_laws([Law|Laws]) :- apply_add_law(Law), apply_add_laws(Laws).


/*-------------
apply_delete_law(Law): Deletea  replacement law Law from the current theory.
------------*/
apply_delete_law(N) :-
    law(N,L), retract(law(N,L)), retract(max_law(K)), K1 is K-1, assert(max_law(K1)).

/*---------------
read_list(N): Read a file which contains a prolog program, and assert each clause of the
program in the environment using the dynamic predicate clause3.
-----------------*/
%read_list(N) :-
%    read(C), C\==end_of_file, !,
%    (C=(Head :- Body) -> (M is N + 1,
%    (conj_to_list(Body, List_Lit) -> assert(clause3(N,(Head,List_Lit),prog));
%     (Body==true -> assert(clause3(N,(Head,[]),prog)) 
%      ; assert(clause3(N,(Head,[Body]),prog))))) ;
%     (C=..[type,Pred,Ar,Type] -> 
%                      (assert(type_definition(Pred,Ar,Type)), M = N) ;
%     (C=..[mode,Pred,Ar,Mode] -> 
%                      (assert(mode_definition(Pred,Ar,Mode)), M = N) ;
%     (M is N + 1, assert(clause3(N,(C,[]),prog)))))), read_list(M).
%read_list(N) :- M is N-1, assert(number_clauses(M)).

% Modified on 13/2/98
read_list(N) :-
    read(C), C\==end_of_file, !,
    (C=(Head :- Body) ->
	(M is N + 1, conj_to_list(Body, List_Lit), assert(clause3(N,(Head,List_Lit),prog)));
     (C=..[type,Pred,Ar,Type] -> 
                      (assert(type_definition(Pred,Ar,Type)), M = N) ;
     (C=..[mode,Pred,Ar,Mode] -> 
                      (assert(mode_definition(Pred,Ar,Mode)),
		      assert(mode(Pred,Ar,Mode)), M = N) ;
     (M is N + 1, assert(clause3(N,(C,[]),prog)))))), read_list(M).
read_list(N) :- M is N-1, assert(number_clauses(M)).


read_list_cont(N) :-
    read(C), C\==end_of_file, !,
    (C=(Head :- Body) -> (M is N + 1, conj_to_list(Body, List_Lit),
                          introduce_continuation_in_clause((Head,List_Lit),C2),
                          assert(clause3(N,C2,prog))) ;
     (C=..[type,Pred,Ar,Type] -> 
                      (assert(type_definition(Pred,Ar,Type)), M = N) ;
     (C=..[mode,Pred,Ar,Mode] -> 
                      (assert(mode_definition(Pred,Ar,Mode)),
		      assert(mode(Pred,Ar,Mode)), M = N) ;
     (M is N + 1, introduce_continuation_in_clause((C,[]),C2), assert(clause3(N,C2,prog)))))),
     read_list_cont(M).

read_list_cont(N) :- M is N-1, assert(number_clauses(M)).

/*---------------
update_current_prog((N,P),(N1,P1)): retract current program N and assert the new one P1.
-----------------*/
update_current_prog((N,P),(N1,P1)) :- retract(current_prog(N,P)), assert(current_prog(N1,P1)).


/*---------------
add_clauses(Numbers,Clauses,Aux): the clauses in the list Clauses are added to the current
program. They are appropriately numbered.
-----------------*/
%add_clauses([],[],_).
%add_clauses([M|Ms],[C|Cs],Aux) :-
%     max_clause(N), M is N+1, assert(clause3(M,C,Aux)),
%     retract(max_clause(N)), assert(max_clause(M)), add_clauses(Ms,Cs,Aux).

% Changed for efficiency improvement 22/10/97

add_clauses(LN,LC,Aux) :- max_clause(Max), add_clauses(Max,LN,LC,Aux).
add_clauses(Max,[],[],_) :- retract(max_clause(_)), assert(max_clause(Max)).
add_clauses(Max,[M|Ms],[C|Cs],Aux) :-
     M is Max+1, assert(clause3(M,C,Aux)), add_clauses(M,Ms,Cs,Aux).

/*---------------
update_programs(LA,PN): it updates the tree of programs.
-----------------*/
%below is the case where the transformation concerns a list of clauses in the current program,
%like, for instance, the fold, delete_clauses, subsumption, restriction rules...

update_programs(([E|LN],R,Ns),PN) :-
    current_prog(M,P),
    delete_insert([E|LN], P, Ns, P1),
    expand(M,([E|LN],R,Ns),P1,PN),
    update_current_prog((M,P),(PN,P1)).

% below is the case where the transformation concerns one clause only in the current program,
% like, for instance, the ts_fold, rearrange_and_fold, unfold (with its various options),
% add_goal, delete_goal, move_lit, case_split, delete_clause, replacement_clause,
%  replacement_goal... A special case is done for addclause, because an additional parameter
% is needed  for the position where to add the new clause.

update_programs(('at last position',addclause,Ns),PN) :- !,
    current_prog(M,P),
    append(P,Ns,P1),
    expand(M,(last,addclause,Ns),P1,PN),
    update_current_prog((M,P),(PN,P1)).
update_programs((N,addclause,Ns),PN) :- !,
    current_prog(M,P),
    pos1(N,P,Pos),
    split(P,Pos,BeforeNodes,N,AfterNodes),
    append(BeforeNodes,Ns,[N|AfterNodes],P1), 
    expand(M,(N,addclause,Ns),P1,PN),
    update_current_prog((M,P),(PN,P1)).
update_programs(('at last position',moveclause,N),PN) :- !,
    current_prog(M,P),
    append(P,[N],P1),
    expand(M,(last,moveclause,N),P1,PN),
    update_current_prog((M,P),(PN,P1)).
update_programs((Pos,moveclause,N),PN) :- !,
    current_prog(M,P),
    pos1(Pos,P,Position),
    split(P,Position,BeforeNodes2,Pos,AfterNodes2),
    (del(N,BeforeNodes2,BeforeNodes) -> AfterNodes=AfterNodes2 ;
    (del(N,AfterNodes2,AfterNodes), BeforeNodes=BeforeNodes2 )),
     append(BeforeNodes,[N],[Position|AfterNodes],P1),
     expand(M,(Pos,moveclause,N),P1,PN),
     update_current_prog((M,P),(PN,P1)).

update_programs((N,R,Ns),PN) :- 
    current_prog(M,P),
    pos1(N,P,Pos),
    split(P,Pos,BeforeNodes,N,AfterNodes),
    append(BeforeNodes,Ns,AfterNodes,P1),
    expand(M,(N,R,Ns),P1,PN),
    update_current_prog((M,P),(PN,P1)).

% below is the case where the transformation does not use any input clauses. It is the case
% for ts_define, gen_define, add_clause.
update_programs((R,N),PN) :-
      (R=def; R =gen_def ; R=addclause ; R=after_lloyd_topor ; R=synth),
      current_prog(M,P),
      append(P,N,P1),
      expand(M,(R,N),P1,PN),
      update_current_prog((M,P),(PN,P1)).

/*----------
expand(Node,LA,LN,SNode): SNode is added to the tree as a successor node of the node Node.
SNode is labeled by LN and the arc (Node,SNode) is labeled by LA.
-----------*/
expand(Node,LA,LN,SN) :-
   maxnode(M), update(Node,M,SN), new_node(M,Node,LA,LN), update_maxnode.

/*----------
update(Node,M,SN): update the sons of the node Node after introduction of SN, a new successor.
-----------*/
update(Node,M,SN) :- 
   node(Node,F,OldSons,LA,LN), retract(node(Node,F,OldSons,LA,LN)), SN is M+1,
   append(OldSons,[SN],NewSons), asserta(node(Node,F,NewSons,LA,LN)).

/*----------
new_node(M,Node,LA,LN): a new successor node of Node is created, labeled by LN and with an
ingoing arc labeled by LA.
-----------*/
new_node(M,Node,LA,LN) :- SN is M+1, asserta(node(SN,Node,[],LA,LN)).

/*----------
update_maxnode: update the numbers of the nodes in the tree.
-----------*/
update_maxnode:- maxnode(M), New is M+1, retract(maxnode(M)), asserta(maxnode(New)).

/*----------
concat_law(File_Theory,Law): add a new law at the end of a file of a theory.
-----------*/
concat_law(File_Theory,(L->R)) :-
   open(File_Theory,append,S), numbervars((L->R),0,_),
   write(S,L), write(S,' -> '), write(S,R), write(S, '.'), nl(S), close(S).
/*----------
assert_comment_history(P,Comment): assert comments associated with program P.
-----------*/
assert_comment_history(P,Comment) :- assert(comment_history(P,Comment)).
%\end{verbatim}
%\end{small}
%\end{document}




