% folding_rules.pl
%
% This file contains the predicates which specify the effects of applying the
% transformation rules associated with the predicates ts_fold, fold and
% rearrange_and_fold which are defined in top_level.pl, after selection and
% checking of the input parameters.

:- use_module(library(lists), [ append/3,
                                nth/4,
                                min_list/2,
                                same_length/2 ]).
:- use_module(library(terms), [ variant/2,
                                subsumes/2 ]). %,unify_with_occurs_check/2

/*---------
 apply_ts_fold(C,[A1,...,An],D): Fold clause A wrt
 atoms at positions A1,...,An
 using clause D.
----------*/
apply_ts_fold(Clause_Number,Atom_Numbers,Def_Number) :- 
    current_prog(PN,CNs), 
    (label_arc(PN,(Clause_Number,
        (fold,Atom_Numbers,Def_Number), NewCN),PNSucc) ->
    (label_node(PNSucc,CNsSucc),
       update_current_prog((PN,CNs),(PNSucc,CNsSucc)))
    ; (apply_ts_fold(Clause_Number,Atom_Numbers,Def_Number,Clause),
       add_clauses(NewCN,[Clause],nil),
       update_programs((Clause_Number,
         (fold,Atom_Numbers, Def_Number),NewCN),PNSucc))),
    (derivation_trace_on ->
     new_clauses(fold,Clause_Number,Atom_Numbers,Def_Number,NewCN);true).


% Pos is the position where the insertion in the body of the folded clause must operate.

apply_ts_fold(Clause_Number,Atom_Numbers,Def_Number,Clause) :-
    (label_arc(A,(gen_def,[Def_Number])) ; label_arc(A,(def,[Def_Number]))),
    current_program(PN), (check_ancestor(A,PN) ->
    (clause_number_to_clause(Clause_Number,(H,B)),
    clause_number_to_clause(Def_Number,(H_def,B_def)),
    positions_to_elts(B,Atom_Numbers,Atoms), delete_2(Atoms,B,B1),
    ts_conditions(H_def,B_def,H,Atoms,B1), min_list(Atom_Numbers,Pos),
    nth(Pos,NewBody,H_def,B1) , Clause=(H,NewBody))
    ; msg_not_in_ancestor([Def_Number]), fail).

% ts_conditions(H_def,B_def,H,Atoms,B1): let C and D be respectively two clauses such that
% head(C)=H, body(C) is made out of the atoms of the two lists Atoms B1,
% head(D)=H_def and body(D) is made out of the atoms in list B_def.
% We check that (C,Atoms,D) satisfies the Tamaki-Sato conditions.

ts_conditions(H_def,B_Def,H,Atoms,B1) :-
    int_var(H_def,B_Def,V), subsumes(B_Def,Atoms),
    free_variables(H_def,[],[],V1), intersection_vars_empty(V,V1),
    free_variables([H|B1],[],[],V2), intersection_vars_empty(V,V2).

/*---------
 apply_fold(LC,LC2,GenH,A): Fold clauses in LC using clauses in LC2. GenH is the
 head of the folded clause, and A, the atom in its body which unifies with all
 heads of clauses in LC.
----------*/
apply_fold(Clauses_Numbers,Clauses_Numbers2,GenHead,A,NewCN) :- 
   (fold_cond2(Clauses_Numbers,Clauses_Numbers2,GenHead,A,NewClause) ->
   apply_fold(Clauses_Numbers,Clauses_Numbers2,NewClause,NewCN); 
   msg_impossible_folding, fail).

%apply_fold(Clauses_Numbers,Clauses_Numbers2,NewClause) :-
% Changed 16/10/97
apply_fold(Clauses_Numbers,Clauses_Numbers2,NewClause,NewCN) :-
   add_clauses(NewCN,[NewClause],nil), 
   update_programs((Clauses_Numbers,(fold, Clauses_Numbers2), NewCN), _), 
   (derivation_trace_on ->
        new_clauses(fold,Clauses_Numbers,Clauses_Numbers2,NewCN);true).
/*---------
 apply_general_fold(LC,LC2,GenH,A): Fold clauses in LC using clauses in LC2.
 GenH is the head of the folded clause, and A, the atom in its body which unifies
 with all heads of clauses in LC.
 This predicate is called by the rule called rearrange_and_fold in top_level.pl.
----------*/
apply_general_fold(CsN,CsN2,GenHead,A,NewCN) :- 
   (general_fold_cond2(CsN,CsN2,GenHead,A,NewClause) ->
   apply_fold(CsN,CsN2,NewClause,NewCN); msg_impossible_folding, fail).

% Below are defined the predicates which check the conditions for folding a list
% C1...Cn of clauses belonging to the current program wrt another list D1...Dn
% of clauses.
%
% The conditions of application are the following:
% 
% 1) D1...Dn belong to a program Pj which is an ancestor of current program,
%
% 2) C1...Cn are of the form:                    % 3) D1...Dn are of the form:  
%                                                %
%                                                %
% C1: (GenHead :- L, Bd(C1), R) Theta_1          %    D1:   Hd(D1) :- Bd(D1)
%               .                                %               .                 
%               .                                %               .                 
%               .                                %               .                 
%                                                %
% Cn: (GenHead :- L, Bd(Cn), R) Theta_n          %    Dn:   Hd(Dn) :- Bd(Dn)       
%                                                %
%
% 4) D1...Dn are unifiable to A via mgu Theta_1...Theta_n
%
% 5) For any clause D: Hd(D) :- Bd(D) in Pj different from D1...Dn, Hd(D) is not
%     unifiable to A
%
% If the above conditions are satisfied, the folding of C1...Cn using
% D1...Dn is possible and the folded clause is Hd :- L, A, R.

/*------------
 check_input_clauses_for_folding(LC,LD,GenHead,A) is true if the conditions for
 folding described above are verified with LC and LD being two lists of clauses
 numbers corresponding to C1...Cn and D1...Dn respectively.
-------------*/
% failure with appropriate message if the arguments are not lists of same length.
check_input_clauses_for_folding(Clauses_Numbers,Clauses_Numbers2,_,_) :-
    \+ same_length(Clauses_Numbers,Clauses_Numbers2), !,
    msg_impossible_folding_1, fail.

% failure with appropriate message if the list of clauses wrt which the folding
% must be done do not belong to an ancestor of current program. For efficiency,
% it is first checked among the ancestors which have been derived by an intro-
% duction definition step (we expect that the clauses which are used for folding
% are more likely to be clauses that have been introduced togehter by an intro-
% duction definition step). Then it is checked systematically in all ancestors.
% If there exist an ancestor which contains D1...Dn exists, it is checked
% whether C1...Cn and D1...Dn are of the form described in the file header
% (points 2) and 3)). If it is the case, GenHead and A are returned as output
% and will be used to form the resulting folded clause. Otherwise, the folding
% step fails with an appropriate error message.
check_input_clauses_for_folding(Clauses_Numbers,Clauses_Numbers2,GenHead,A) :-
   label_arc(Anc,(gen_def,Defs)),
   seteq(Defs,Clauses_Numbers2), !,
   current_program(PN),
   (check_ancestor(Anc,PN) -> (heads_unifiable(Clauses_Numbers,_,GenHead),
   heads_unifiable(Clauses_Numbers2,_,A)) 
   ; (msg_not_in_ancestor(Clauses_Numbers2), fail)).
check_input_clauses_for_folding(Clauses_Numbers,Clauses_Numbers2,GenHead,A) :-
    current_program(PN), select_ancestor(Pj,PN), label_node(Pj,Clauses_Pj), 
    subset(Clauses_Numbers2,Clauses_Pj), heads_unifiable(Clauses_Numbers2,_,A),
    heads_unifiable(Clauses_Numbers,_,GenHead),
    no_other_head_unifiable(Clauses_Numbers2,Clauses_Pj,A), !.

% if C1...Cn and D1...Dn are not of the form specified in points 2) and 3), the
% folding step is not possible. There is a failure with an appropriate message.
check_input_clauses_for_folding(_,C_N2,_,_) :- msg_impossible_folding_2(C_N2).
/*-------------
 fold_cond2(Clauses, Defs, Gen_Head, A, Folded) is true if Folded is the clause
 which corresponds to the folding of the list Clauses of clauses wrt the list
 Defs of clauses. Depending on the use of fold_cond2 (within an interactive
 folding step with parameters Clauses and Defs selected by the user or in the
 context of the strategy determ, the checking specified by predicate
 check_input_clauses_for_folding in order to fold Clauses wrt Defs has been
 performed or not. If it has been performed (interactive folding step),
 Gen_Head and A are known (non variable terms). Otherwise, they are inferred by
 using predicate "head_unifiable" (the other conditions specified in the file
 header are always satisfied when a folding step is attempted in the "determ"
 strategy. Since the presence of unit clauses create some problems, they are
 first put aside (omit_first_unit_clauses), then the folding is attempted on
 the non unit clauses (fold_cond2_), and if it succeeds, it is verified that it
 is correct also for the unit clauses (verify_recover_unit_clauses).
--------------*/
fold_cond2(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   fold_cond2_(Clauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

omit_first_unit_clauses([],[],[],[],[],[]).
omit_first_unit_clauses([C|Cls],[D|Defs],[C|Units],[D|DefUnits],Cls2,Defs2) :-
    is_a_unit_clause(D),!,
    omit_first_unit_clauses(Cls,Defs,Units,DefUnits,Cls2,Defs2).
omit_first_unit_clauses([C|Cls],[D|Defs],Units,DefUnits,[C|Cls2],[D|Defs2]) :-
    omit_first_unit_clauses(Cls,Defs,Units,DefUnits,Cls2,Defs2).

verify_recover_unit_clauses(_,[],[]).
verify_recover_unit_clauses(Folded,[U|Units],[D|DefUnits]) :-
    clause_number_to_clause(D, (H,[])), clause_number_to_clause(U, CU),
    copy_term(Folded,Folded2), copy_term(H,H2), unfold_body(Folded2,H2, CU),
    verify_recover_unit_clauses(Folded,Units,DefUnits).

unfold_body((H,Bs),A,U) :- member_rest(B,Bs,R), A = B, variant((H,R),U).

fold_cond2_(Cls, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable(Cls,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable(Defs,_,A) ; true),
    minimal_hd_generalization_clauses(Cls,Gen_Head,_,_,Gen_Cl),
    minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
    recover_folded_clause(Gen_Cl,Gen_Def, Folded).
/*---------
 minimal_hd_generalization_clauses(NumCls,Gen_Head,Cls,Gen_Cls,Dis_Cl): NumCls is
 a list of clause numbers corresponding to clauses Cls, Gen_Head is an atom which
 generalizes the head of each clause in Cls, Gen_Cls is the list Cls of clauses
 after generalization so that their heads is Gen_Head (they might contain
 equalities in their bodies) Dis_Cl is a disjunctive clause with head Gen_Head and
 whose body is the disjunction of the bodies in Gen_Cls
-----------*/
minimal_hd_generalization_clauses(NumCls,Gen_Head,Cls,Gen_Cls,Dis_Cl) :-
    Gen_Head=..[_|Args_Gen_Head],
    minimal_hd_generalization_clauses(NumCls,Args_Gen_Head,Cls,Gen_Cls),
    disjunctive_clause(Gen_Cls,Dis_Cl).
minimal_hd_generalization_clauses([],_,[],[]).
minimal_hd_generalization_clauses([N|P],Gen_Head,[(Hd,Bd)|L],[C|Cs]) :-
    clause_number_to_clause(N, (Hd,Bd)), Hd=..[F|Args],
    minimal_move_arguments_in_body(Args,Gen_Head,GenArgs,Eqs),
    GenHd=..[F|GenArgs], 
    insere_equalities_after_inequalities(Eqs,Bd,ExtBody),
    C=(GenHd,ExtBody),
    minimal_hd_generalization_clauses(P,Gen_Head,L,Cs).

disjunctive_clause([(H,Bd)], (H, [Bd]) ) :-!.
disjunctive_clause([(H,Bd)|Gen_Clauses], (H, [Bd | Bds]) ) :-
    disjunctive_clause(Gen_Clauses, (H, Bds)).

minimal_move_arguments_in_body([],_,[],[]).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[A|GenArgs],Eqs) :-
     var(A), occurs_at_same_pos(A,Args,B,GenHead), !, A = B,
     minimal_move_arguments_in_body(Args,GenHead,GenArgs,Eqs).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[X|GenArgs],[X=A|Eqs]) :-
    var(B), !, minimal_move_arguments_in_body(Args,GenHead,GenArgs,Eqs).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[X|GenArgs],Eqs) :-
    A =.. [F|ArgsA], B =.. [F|ArgsB], copy_term(B,X),
    X =.. [F|ArgsX], append(ArgsA,Args,ArgsA2),
    append(ArgsB,GenHead,GenHead2), append(ArgsX,GenArgs,GenArgs2),
    minimal_move_arguments_in_body(ArgsA2,GenHead2,GenArgs2,Eqs).

% Particular case where Args2 is a term more general than Args1
occurs_at_same_pos(V1,Args1,_,_) :- free_of_var(V1,Args1), !.
occurs_at_same_pos(_,[],_,[]).
occurs_at_same_pos(A,[T1|T1s],B,[T2|T2s]) :-
    occurs_at_same_pos_in_term(A,T1,B,T2), occurs_at_same_pos(A,T1s,B,T2s).
occurs_at_same_pos_in_term(A,T1,_,_) :- free_of_var(A,T1), !.
occurs_at_same_pos_in_term(A,T1,B,T2) :-  A == T1, B == T2.
occurs_at_same_pos_in_term(A,T1,B,T2) :-
    nonvar(T1), nonvar(T2), T1=..[_|Args1], T2=..[_|Args2],
    occurs_at_same_pos(A,Args1,B,Args2).

insere_equalities_after_inequalities(Body,[X|B2],[X|ExtBody]):-
    (var(X) ; nonvar(X), X = \+(_=_)), !,
    insere_equalities_after_inequalities(Body,B2,ExtBody).
insere_equalities_after_inequalities([],B,B) :- !.
insere_equalities_after_inequalities(Body,B2,B) :- append(Body,B2,B).
    
recover_folded_clause( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
    subseq_body_match_def( H1, Bds, A, Bds2, New_Bd_C).
/* Corrected 6 Nov 97
subseq_body_match_def(_,[],_,[],_).
subseq_body_match_def( H1, [Bd_C|Bds], A, [Bd_D|Bds2], New_Bd_C ) :-
    free_variables(H1,[],[],VarsH1),
    folded_part(F,Folded_Part,G, Bd_C),
    \+ \+ unify_with_occurs_check(Folded_Part,Bd_D),
    free_variables(Bd_D,VarsH1,[],VarsInt1),
    free_variables((F,G),VarsH1,[],VarsInt2),
    free_variables(VarsInt1,VarsInt2,[],VarsInt1),
    unify_with_occurs_check(Folded_Part,Bd_D),
%    free_variables(VarsInt1,[],[],Vars), VarsInt1==Vars,
    free_variables(VarsInt2,[],[],Vars), VarsInt2==Vars,
    append([F,[A],G], New_Bd_C),
    subseq_body_match_def( H1, Bds, A, Bds2, New_Bd_C ).
*/

subseq_body_match_def(_,[],_,[],_).
subseq_body_match_def( H1, [Bd_C|Bds], A, [Bd_D|Bds2], New_Bd_C ) :-
    same_length(Folded_Part,Bd_D),	
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part(F,Folded_Part,G, Bd_C),
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,F,G,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
%    free_variables(VarsIntD,[],[],VarsIntD),
% Corrected the 17/02/98    
    free_variables(Bd_D,A,[],VarsIntD), % internal vars of Def unchanged
    free_variables(VarsFoldedPart,[],[],VarsFoldedPart), % Folded part is an instance of Bd_Def
%    free_variables(A,[],LV,VA),
    free_variables(A,[],[],VA2), in_var(LV,VA2),
    append([F,[A],G], New_Bd_C),
    subseq_body_match_def( H1, Bds, A, Bds2, New_Bd_C ).

in_var([],_).
in_var([X|L],L2) :- belongs(X,L2), in_var(L,L2).

folded_part(F, Folded_Part, G, Bd_C) :-
    body_without_equalities(Bd_C, Eqs, Real_Bd_C),
%    append([F, [B1|B1s], G], Real_Bd_C), append(Eqs, [B1|B1s], Folded_Part).
        append([F, B1s, G], Real_Bd_C), append(Eqs, B1s, Folded_Part).
folded_part_whole_bd(Ineqs,Folded_Part,[],Bd_C) :-
    body_without_equalities(Bd_C, Eqs, Bd_C1),
    body_without_inequalities(Bd_C1,Ineqs,Bs),
    append(Eqs, Bs, Folded_Part).


linking_vars_folded_part(H,F,G,VarsFolded_Part,LV) :-
    free_variables((H,F,G),[],[],VarsHFG),
    common_vars(VarsHFG,VarsFolded_Part,LV).
% Added the 16/2/98 for the case F=G=[] (folding of the whole body)
linking_vars_folded_part(H,VarsFolded_Part,LV) :-
    free_variables(H,[],[],VarsHFG),
    common_vars(VarsHFG,VarsFolded_Part,LV).

common_vars([],_,[]).
common_vars([V|VarsHFG],VarsFolded_Part,[V|LV]):-   
    belongs(V,VarsFolded_Part), !,
    common_vars(VarsHFG,VarsFolded_Part,LV).   
common_vars([_|VarsHFG],VarsFolded_Part,LV):-   
    common_vars(VarsHFG,VarsFolded_Part,LV).   

/*------------
 general_fold_cond2 is a generalization of fold_cond2 where Clauses might be
 reordered to perform the folding step.
-------------*/

general_fold_cond2(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   permutation(Clauses2,PClauses2),
   general_fold_cond2_(PClauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

general_fold_cond2_(Clauses, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable(Clauses,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable(Defs,_,A) ; true),
    minimal_hd_generalization_clauses(Clauses,Gen_Head,_,_,Gen_Clause),
    minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
    general_recover_folded_clause(Gen_Clause,Gen_Def, Folded).

general_recover_folded_clause( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
    subseq_body_weak_match_def( H1, Bds, A, Bds2, New_Bd_C).

% sorting is performed for efficiency reason.
/*subseq_body_weak_match_def(_,[],_,[],_).
subseq_body_weak_match_def( H1,[Bd_C|Bds],A,Bds2,New_Bd_C ) :-
    del_und(Bd_D,_,Bds2,Bds3),same_length(B1s,Bd_D),
    sort_list(Bd_D,Bd_D_Sorted), sort_list(Bd_C,Bd_C_Sorted),
    
    free_variables(H1,[],[],VarsH1),
    subseq(Bd_C_Sorted,B1s,FG),
    \+ \+ unify_with_occurs_check(B1s,Bd_D_Sorted),
    free_variables(Bd_D_Sorted,VarsH1,[],VarsInt1),
    free_variables(FG,VarsH1,[],VarsInt2),
    free_variables(VarsInt1,VarsInt2,[],VarsInt1),
    unify_with_occurs_check(B1s,Bd_D_Sorted),
    free_variables(VarsInt1,[],[],Vars), VarsInt1==Vars,
    [A|FG]= New_Bd_C,subseq_body_weak_match_def( H1,Bds,A,Bds3,New_Bd_C ).
*/
subseq_body_weak_match_def(_,[],_,[],_).
subseq_body_weak_match_def( H1,[Bd_C_unsorted|Bds],A,[Bd_D_unsorted|Bds2],New_Bd_C ) :-
    same_length(Folded_Part,Bd_D_unsorted),
    sort_list(Bd_D_unsorted,Bd_D), sort_list(Bd_C_unsorted,Bd_C),
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part(F,Folded_Part,G, Bd_C),
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,F,G,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
%    free_variables(VarsIntD,[],[],VarsIntD),
% Corrected the 17/02/98    
    free_variables(Bd_D,A,[],VarsIntD),
    free_variables(VarsFoldedPart,[],[],VarsFoldedPart),
%    free_variables(A,[],LV,VA),
%    free_variables(A,[],[],VA),
    free_variables(A,[],[],VA2), in_var(LV,VA2),
    append([F,[A],G], New_Bd_C),
%   sort_list(New_Bd_C,New_Bd_C_sorted),
    subseq_body_weak_match_def( H1,Bds,A,Bds2,New_Bd_C ).

fold_cond2_whole_bd(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   fold_cond2_whole_bd_(Clauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

fold_cond2_whole_bd_(Cls, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable(Cls,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable(Defs,_,A) ; true),
    minimal_hd_generalization_clauses(Cls,Gen_Head,_,_,Gen_Cl),
    minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
    recover_folded_clause_whole_bd(Gen_Cl,Gen_Def, Folded).

recover_folded_clause_whole_bd( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
    subseq_body_match_def_whole_bd( H1, Bds, A, Bds2, New_Bd_C).

subseq_body_match_def_whole_bd(_,[],_,[],_).
subseq_body_match_def_whole_bd( H1, [Bd_C|Bds], A, [Bd_D|Bds2], New_Bd_C ) :-
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part_whole_bd(F,Folded_Part,G,Bd_C),
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
%    free_variables(VarsIntD,[],[],VarsIntD),
% Corrected the 17/02/98    
    free_variables(Bd_D,A,[],VarsIntD),
    free_variables(VarsFoldedPart,[],[],VarsFoldedPart),
%    free_variables(A,[],LV,VA),
    free_variables(A,[],[],VA2), in_var(LV,VA2), 
    append([F,[A],G], New_Bd_C),
%New_Bd_C=[A],
    subseq_body_match_def_whole_bd( H1, Bds, A, Bds2, New_Bd_C ).




general_fold_cond2_whole_bd(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   general_fold_cond2_whole_bd_(Clauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

general_fold_cond2_whole_bd_(Cls, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable(Cls,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable(Defs,_,A) ; true),
    minimal_hd_generalization_clauses(Cls,Gen_Head,_,_,Gen_Cl),
    minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
    general_recover_folded_clause_whole_bd(Gen_Cl,Gen_Def, Folded).

general_recover_folded_clause_whole_bd( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
    subseq_body_weak_match_def_whole_bd( H1, Bds, A, Bds2, New_Bd_C).

subseq_body_weak_match_def_whole_bd(_,[],_,[],_).
subseq_body_weak_match_def_whole_bd( H1,[Bd_C_unsorted|Bds],A,[Bd_D_unsorted|Bds2],New_Bd_C ) :-
    same_length(Folded_Part,Bd_D_unsorted),
    sort_list(Bd_D_unsorted,Bd_D), sort_list(Bd_C_unsorted,Bd_C),
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part_whole_bd(F,Folded_Part,G,Bd_C),
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
%    free_variables(VarsIntD,[],[],VarsIntD),
% Corrected the 17/02/98    
    free_variables(Bd_D,A,[],VarsIntD),
    free_variables(VarsFoldedPart,[],[],VarsFoldedPart),
%    free_variables(A,[],LV,VA),
    free_variables(A,[],[],VA2), in_var(LV,VA2), 
    append([F,[A],G], New_Bd_C),
%New_Bd_C=[A],
    subseq_body_weak_match_def_whole_bd( H1,Bds,A,Bds2,New_Bd_C ).






