% check_statement(+Newp,+F):-  Newp :- F is a statement of the form
% where Newp and F have the same set of free variables.

check_statement(Newp,F) :-
        ((fv(Newp,Vars1), fv(F,Vars2), sort(Vars1,V1), sort(Vars2,V1)) ->
                true ;
                (error_msg(not_statement,[Newp,F]),
                 discard_tmp_program,
                 fail)).

% new_statement(Newp, F) :-  Newp :- F is a statement of the form
% Pred(V1,...,Vn) :- F, where:
% (i) Pred is a new predicate name, and
% (ii) V1,...,Vn  are the free variables of F.

new_statement(Newp, F) :-
        fv(F,Vars),
        gen_new_name_functor(Pred),
        Newp=..[Pred|Vars].

error_msg(not_statement,[Newp,F]) :-
        write('ERROR: '), write(Newp), write(' and '), write(F),
        write(' have different free variables').
error_msg(not_atom,[Newp]) :-
        write('ERROR: '), write(Newp), write(' is not an atom').
error_msg(not_new_pred,[P]) :-
        write('ERROR: '), write(P), write(' is not a new predicate').
error_msg(no_type,[Pred]) :-
        write('ERROR: no type definition for '), write(Pred).
error_msg(not_lit,[L]) :-
        write('ERROR: '), write(L), write(' is not a literal').
error_msg(neg_lit,[L]) :-
        write('ERROR: negative literal'), write(L),
        write(' cannot be unfolded').


check_closed_fo_formula(Phi) :-
        fo_formula(Phi),
        fv(Phi,L),
        L=[], !.
check_closed_fo_formula(_) :-
        write('not a closed first order formula'), nl, fail.

% fo_formula(F) :-
% F is a first order formula constructed from
% true, false, atomic formulas,
% and, or, ->, <->, e (exists), a (forall).

fo_formula(true).
fo_formula(false).
fo_formula(F) :- atomic_formula(F), !.
fo_formula(\+F) :- fo_formula(F).
fo_formula((F1 and F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 or F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 -> F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 <-> F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula(e(X,F)) :- nonvar(X), list_of_vars(X), fo_formula(F).
fo_formula(a(X,F)) :- nonvar(X), list_of_vars(X), fo_formula(F).

atomic_formula(F) :-
        F\=(\+ _),
        F\=(_ and _),
        F\=(_ or _),
        F\=(_ -> _),
        F\=(_ <-> _),
        F\=e(_,_),
        F\=a(_,_).

% list_of_vars(Vs) :-  Vs is a list of variable names

list_of_vars([]) .
list_of_vars([V|Vs]) :- var(V), list_of_vars(Vs).


%fv(+F,?Vs): F -> Vs
%fv(F,Vs) :- Vs is the set of free variables in F

fv(true,[]).
fv(false,[]).
fv(A,V) :-
        atomic_formula(A), !,
        term_variables(A,V).
fv((\+ F),V) :- fv(F,V).
fv((F1 and F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 or F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 -> F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 <-> F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv(e(X,F),V) :-
        fv(F,V1),
        delete_all(X,V1,V).
fv(a(X,F),V) :-
        fv(F,V1),
        delete_all(X,V1,V).


% delete_all(+Xs,+Ys,?Zs): Xs,Ys -> Zs
% Zs is obtained by deleting from Ys all
% occurrences of the terms in Xs.

delete_all([],L,L).
delete_all([X|Xs],Ys,Zs) :-
        delete_one(X,Ys,Zs1),
        delete_all(Xs,Zs1,Zs).

delete_one(_,[],[]).
delete_one(X,[Y|Ys],Zs) :- X==Y, !, delete_one(X,Ys,Zs).
delete_one(X,[Y|Ys],Zs) :- Zs=[Y|Zs1], delete_one(X,Ys,Zs1).

% MODE: check_property(+Phi,+Prog)
% TYPE: check_property(predicate,list(def))
% SEMANTICS: Phi is a nullary predicate defined by a single clause in
% the current program Prog

check_property(Phi,Prog,Property) :-
	atom(Phi),
	def((Phi/0),Prog,(Phi/0,D)),
	D=[(_,bd([],[Property]))].

% =====================================================
% LLOYD-TOPOR TRANSFORMATION (ARRAYS)

% MODE: lloyd_topor_transf(+Newp,+Phi,?Defs)
% TYPE:
% SEMANTICS: lloyd_topor_transf(+Newp,+Phi,?Defs) :-
% apply Lloyd-Topor Transformation to the statement Newp :- Phi
% and derive a list Defs of single-clause definitions.
% In the list Defs = [D1,...,Dn] a predicate occurring in the
% body of Dj is defined in Di (i<j) or in the initial program P.
% We assume that Phi is typable by using the type definitions
% declared in the program by the type/3 predicate and stored as
% tmp_type_definition/3 facts.
% Definitions are asserted as facts of
% the tmp_definition_clause/2 % predicate.

lloyd_topor_transf_arrays(Pred,Phi,Defs,NDefs) :-
        check_statement(Pred,Phi),
        rewrite_formula_arrays(Phi,Phi1),
        statement_into_clauses_arrays((Pred,Phi1),Defs),
	length(Defs,NDefs).


% rewrite_formula(Phi,Phi1) :- Phi1 is logically equivalent to Phi and uses
% only true, \+ (negation), and (conjunction), e (exists).
rewrite_formula_arrays(true,(\+ false)).
rewrite_formula_arrays(false,false).
rewrite_formula_arrays(F1,F2) :- atomic_formula(F1), !, F1=F2.
rewrite_formula_arrays((\+ F1),(\+ F2)) :- rewrite_formula_arrays(F1,F2).
rewrite_formula_arrays((F1 and F2), ((R1) and (R2))) :-
        rewrite_formula_arrays(F1,R1),
        rewrite_formula_arrays(F2,R2).
rewrite_formula_arrays((F1 or F2), (\+ ((\+ R1) and (\+ R2)))) :-
        rewrite_formula_arrays(F1,R1),
        rewrite_formula_arrays(F2,R2).
rewrite_formula_arrays((F1 -> F2), (\+ ((R1) and (\+ R2)))) :-
        rewrite_formula_arrays(F1,R1),
        rewrite_formula_arrays(F2,R2).
rewrite_formula_arrays((F1 <-> F2),
                ((\+ ((R1) and (\+ R2))) and (\+ ((\+ R1) and (R2))))) :-
        rewrite_formula_arrays(F1,R1),
        rewrite_formula_arrays(F2,R2).
rewrite_formula_arrays(e(X,F),e(X,R)) :- rewrite_formula_arrays(F,R).
rewrite_formula_arrays(a(X,F),(\+ e(X,(\+ R)))) :- rewrite_formula_arrays(F,R).

% statement_into_clauses(Stat,Defs) :- Defs is the list of clauses derived by
% applying the typed LT transformation starting from the statement Stat.
% Conjunctions are eliminated in favour of flat lists.
% Clauses with false body are duplicate literals are deleted.
% Type atoms are added.

statement_into_clauses_arrays(Stat,Defs) :-
        ltt_arrays(Stat,[],Cls,_),
        and_flatten_cls(Cls,Cls1),
	normalize_constraints_arrays(Cls1,Cls2),
	separe_all_constraints(Cls2,Cls3),
	t_sort_arrays(Cls3,Cls4), %%% Temporarily here becouse not compatible with Defs
	cls_to_defs(Cls4,Cls5),
        % disabled for efficiency reasons: simplify(Cls4,Cls5),
	Cls6 = Cls5,
        %%% t_sort_arrays(Cls5,Cls6),
        types_of_prog(Cls6),
        add_types_to_prog(Cls6,Defs).

% MODE:
% TYPE:
% SEMANTICS: transform the literal \+ C (where C is a constraint
%            into the corresponding complementar constraint
normalize_constraints_arrays([],[]).
normalize_constraints_arrays([(H,B)|Cls],[(H,NewB)|NewCls]) :-
	normalize_one_clause(B,NewB),
	normalize_constraints_arrays(Cls,NewCls).

normalize_one_clause([],[]).
normalize_one_clause([L|Ls],[NewAtom|NewLs]) :-
	functor(L,\+,_), L =.. [\+,Atom],
	is_a_constraint(Atom),
	complement_c([Atom],[NewAtom]),
	normalize_one_clause(Ls,NewLs).
normalize_one_clause([L|Ls],[L|NewLs]) :-
	normalize_one_clause(Ls,NewLs).


% MODE:
% TYPE:
% SEMANTICS: FUNCTIONAL ! -> must return only one result !!
separe_all_constraints([],[]).
separe_all_constraints([Cl|Cls],[Cl2|NewCls]) :-
	separe_constraints_arrays(Cl,NewCl),
	copy_term(NewCl,Cl2),	
	separe_all_constraints(Cls,NewCls).

separe_constraints_arrays((H,B),(H,bd(C,Body))) :- separe_constr_arrays((H,B),(H,bd(C,Body))), !.

separe_constr_arrays((H,[]),(H,bd([],[]))).
separe_constr_arrays((H,[C|Bs]),(H,bd([C|Cs],NewBs))) :-
	is_a_constraint(C),
	separe_constr_arrays((H,Bs),(H,bd(Cs,NewBs))).
separe_constr_arrays((H,[L|Bs]),(H,bd(C,[L|Ls]))) :-
		separe_constr_arrays((H,Bs),(H,bd(C,Ls))).

% --------------------------



% and_flatten_cls(Cs,Ds) :-  transform the bodies of the
% statements in Cs from a conjunction (which uses the connective and) to a
% flat list.

and_flatten_cls([],[]).
and_flatten_cls([(H,B)|Cs],[(H,FB)|Ds]) :-
        and_flatten(B,FB),
        and_flatten_cls(Cs,Ds).

and_flatten(true,[]).
and_flatten((C1 and C2), F) :-
        !,
        and_flatten(C1,F1),
        and_flatten(C2,F2),
        append(F1,F2,F).
and_flatten(L,F) :- F=[L].


