% MODULE  : solver_clpr
% provides : solve/3,satisfiable/1,entails/2,solve_constraints/2,has_satisfiable_constraint/1
% requires : module(),...,library(),...,

% :- module(solver_clpr).%,[solve/3,satisfiable/1,entails/2,solve_constraints/2,has_satisfiable_constraint/1]).

%% :- use_module(library(clpr)).
:- use_module(utilities).%,[is_member/2]).
:- use_module(library(lists), [append/3,member/2]).

% exports:
% solve(Cns1,Vars,Cns2)
% satisfiable(C)
% entails(Cns1,Cns2)
% solve_constraints(Clause1,Clause2)
% has_satisfiable_constraint(Clause)

:- dynamic projection_temporary/1.

% ====================================================================
% SOLVE FORMULA : elimination of quantifiers from first order formulas
%		  on the closed real field domain

% a([A,B,C],((A>B) -> (A>C))) : false
% a([A,B],(A<B -> e([C],(A<C and C<B)))) : true
%
% esempio 1 per il folding :
% a([A,C],e([A1,B1],a([X,Y,Z],(e([D],(A-1 = X and C = Y and D=Z)) <-> e([C1],(A1=X and B1=Y and C1 = Z)))))) : true
% a([X,Y,Z],(e([D],(A-1 = X and C = Y and D=Z)) <-> e([C1],(A1=X and B1=Y and C1 = Z)))) : A=:=1.0+A1 and C=:=B1 or A=:=1.0+A1 and C=:=B1
%
% esempio 2 per il folding :
% a([A,B,C],(e([A1,B1,C1],a([X,Y,Z,W],(e([D],(D=X and (A-3)=Y and B=Z and C=W)) <-> e([D1,E1],(A1=(D1+2) and E1=X and D1=Y and B1=Z and C1=W))))))) : true
% a([X,Y,Z,W],(e([D],(D=X and (A-3)=Y and B=Z and C=W)) <-> e([D1,E1],(A1=(D1+2) and E1=X and D1=Y and B1=Z and C1=W)))) : B=:=B1 and A=:=1.0+A1 and C=:=C1
%
% esempio 3 per il folding :
% a([A,C],e([A1,B1],a([E,F],((A-1 = E and C = F) <-> (A1=E and B1=F))))) : true
% a([A,B,C],e([A1,B1,C1],a([X,Y,Z,W,H,K],(e([D],(D=X and (A-3)=Y and D=Z and B=W and X=H and C=K)) <-> e([D1,E1],(A1=(D1+2) and E1=X and D1=Y and E1=Z and B1=W and E1=H and C1=K)))))) : true
%
% esempio 4 per il folding :
% C1: new4(A,[B|C]) :- {B=<A}, {D>=A}, list(C), \+new3(D,C).
% C2: new4(A,[B|C]) :- {A<B}, {D>=B}, list(C), \+new3(D,C).
% D:  new4(A,B) :- {C-A>=0.0}, list(B), \+new3(C,B).
% Test: a([A1,C1],e([A,B],a([X,Y,Z],(e([D1],(D1>=A1 and C1=X and D1=Y and C1=Z)) <-> e([C],((C-A)>=0 and B=X and C=Y and B=Z)))))) : true
% Compute: a([X,Y,Z],(e([D1],(D1>=A1 and C1=X and D1=Y and C1=Z)) <-> e([C],((C-A)>=0 and B=X and C=Y and B=Z))))
% Substitution: [A1=:=A and C1=:=B]
% Test: a([B2,C2],e([A,B],a([X,Y,Z],(e([D2],(D2>=B2 and C2=X and D2=Y and C2=Z)) <-> e([C],((C-A)>=0 and B=X and C=Y and B=Z)))))) : true
% Compute: a([X,Y,Z],(e([D2],(D2>=B2 and C2=X and D2=Y and C2=Z)) <-> e([C],((C-A)>=0 and B=X and C=Y and B=Z))))
% Substitution: [B2=:=A and C2=:=B]
%
% a([A1,C1],e([A,B],a([X,Y,Z],(e([D1],(B1=<A1 and D1>=A1 and C1=X and D1=Y and C1=Z)) <-> e([C],((C-A)>=0 and B=X and C=Y and B=Z))))))
%

solve_f(A,D) :-
	solve_formula(A,B),
	project_list_of_constraints(B,[],C), !,
	eliminate_repeated(C,D).

eliminate_repeated(A,D) :-
	flatten_or(A,B),
	remove_duplicates(B,D).

% CONSTANTS
solve_formula(\+ true, false).
solve_formula(\+ false, true).
solve_formula((true and C),D) :- solve_formula(C,D).
solve_formula((false and _),false).
solve_formula((true or _),true).
solve_formula((false or C),D) :- solve_formula(C,D).

% NOT
solve_formula(\+(A and B),(C or D)) :-
	solve_formula((\+ A),C),
	solve_formula((\+ B),D).
solve_formula(\+(A or B),C) :-
	solve_formula(((\+ A) and (\+B)),C).
solve_formula((\+(\+(A))),B) :-
	solve_formula(A,B).
solve_formula((\+(e(Vars,A))),C) :-
	solve_formula(e(Vars,A),B),
	solve_formula((\+(B)),C).
solve_formula((\+ (a(Vars,A))),C) :-
	solve_formula((e(Vars,(\+ A))),C).
solve_formula((\+(A -> B)),C) :-
	solve_formula((A and (\+ B)),C).
solve_formula((\+(A <-> B)),(C or D)) :-
	solve_formula(((\+ A) and B),C),
	solve_formula((A and (\+ B)),D).

% OR
solve_formula((A or B),(C or D)) :-
	solve_formula(A,C),
	solve_formula(B,D).

% AND
solve_formula(((A or B) and C),(D or E)) :-
	solve_formula((A and C),D),
	solve_formula((B and D),E).
solve_formula((A and (B or C)),(D or E)) :-
	solve_formula((A and B),D),
	solve_formula((A and C),E).
solve_formula((A and B),(A and B)) :-
	constraint((A and B)).
solve_formula((A and B),E) :-
	\+ constraint((A and B)),
	\+ constraint(A),
	solve_formula(A,D),
	solve_formula((D and B),E).
solve_formula((A and B),E) :-
	\+ constraint((A and B)),
	constraint(A),
	solve_formula(B,D),
	solve_formula((A and D),E).

% IMPLIES
solve_formula((A -> B),(C or D)) :-
	solve_formula((\+ A),C),
	solve_formula(B,D).

% IFF
solve_formula((A <-> B),(C or D)) :-
	solve_formula((A and B),C),
	solve_formula(((\+ A) and (\+ B)),D).

% EXISTS
solve_formula(e(Vars,A),C) :-
	solve_formula(A,B),
	project_list_of_constraints(B,Vars,C).

% FORALL
solve_formula(a(Vars,A),C) :-
	solve_formula((\+ e(Vars,(\+ A))),C).

% CONSTRAINT
solve_formula(\+ A, B) :-
	complement_c([A],[B]).
solve_formula(\+ A, (LHS>RHS or LHS<RHS)) :-
	(functor(A,=,_);functor(A,=:=,_)),
	A=..[_,LHS,RHS].
solve_formula(A, A) :-
	constraint(A).

constraint(A) :-
	\+ functor(A,and,_),
	(atomic_constraint(A) ; A==true ; A==false).
constraint((A and B)) :-
	atomic_constraint(A),
	constraint(B).
constraint((A and B)) :-
	constraint(A),
	constraint(B).


atomic_constraint(C) :- is_a_constraint(C).
atomic_constraint(false).
atomic_constraint(true).

project_list_of_constraints(A,ExVars,Aproj2) :-
	\+ functor(A,or,_),
	(constraint(A) -> true; (write('ERROR in "project_list_of_constraints/2"'), fail)),
	flatten_and(A,Anew),
	(member(false,Anew) ->
		Aproj2 = false ;
		(term_variables(Anew,Vars),
		 subtract2(Vars,ExVars,OutVars),
		 remove_trues(Anew,Anew2),
		 solve(clpr,Anew2,OutVars,Aproj),
		 introduce_and(Aproj,Aproj2)
		)
	).
project_list_of_constraints((A or B),ExVars,(Aproj or Bproj)) :-
	project_list_of_constraints(A,ExVars,Aproj),
	project_list_of_constraints(B,ExVars,Bproj).


remove_trues([],[]).
remove_trues([true|A],B) :- remove_trues(A,B).
remove_trues([X|A],[X|B]) :- X\==true, remove_trues(A,B).


flatten_and(B,[B]) :-
	atomic_constraint(B).
flatten_and((A and B),E) :-
	flatten_and(A,C),
	flatten_and(B,D),
	append(C,D,E).


flatten_or(B,[B]) :-
	constraint(B).
flatten_or((A or B),E) :-
	flatten_or(A,C),
	flatten_or(B,D),
	append(C,D,E).


introduce_and([B],B).
introduce_and([A|D],(A and B)) :-
	introduce_and(D,B).
introduce_and([true],true).
introduce_and([false],false).
introduce_and([],true).

% ---
% prova(A and B and C).
% >>> 'and' is RIGHT ASSOCIATIVE !!!
% A and B and C ---becomes---> A and (B and C)

prova(A and B) :- write(A), write(' - '), write(B).

% ---

% ====================================================================

solve(Cns,Vars,Cns1):-
	solve(clpr,Cns,Vars,Cns1).

solve_constraints((Hd,bd(Cns,Lits)),(Hd,bd(Cns1,Lits))) :-
	commonvars(Cns,(Hd,Lits),CVars),
	solve(clpr,Cns,CVars,Cns1).

has_satisfiable_constraint((_,bd(Cns,_))) :- satisfiable(clpr,Cns).

satisfiable(C) :-
	satisfiable(clpr,C).

satisfiable(clpr,C) :-                    %from F.  solve_constraints.pl
	solve(clpr,C,[],C1),
	(is_member(false,C1) ->
		fail;
		true).

commonvars(E1,E2,CVars) :-
        setof(X,(
        term_variables(E1,V1),
        term_variables(E2,V2),
        member(X,V1),
        is_member(X,V2)), LV) -> CVars=LV ; CVars=[].

%%% From F. solvers.pl

%%% solve(Solver,C,IntVars,PC):
%%% PC is the projection of C over the variables in the list IntVars
%%% using solver Solver
%%%  - C,PC are constraints
%%%  - T is a term
%%% IMPORTANT NOTICE: vars(IntVars) must be included in vars(C)

solve(clpr,ConsLst, IntVars, PConsLst) :-
	!,
	(ConsLst==[] ->
		 PConsLst=[]
	;
	    ( is_member(false,ConsLst) ->
			  PConsLst=[false]
	    ;
		list_to_constraint(clpr,ConsLst,C),
		copy_term((C,IntVars),(C1,IntVars1)),
		retractall(projection_temporary(_)),
		assert((projection_temporary(IntVars1) :- C1)),
		( call_residue(projection_temporary(IntVars1), _R1) ->
		    get_singletons(IntVars,IntVars1,Eqs,DumpVars,DumpVars1),
		    C1,dump(DumpVars1,DumpVars,D),
		    fix_equalities(D,DF), 
		    append(Eqs,DF,PConsLst_tmp),
		    eliminate_useless_equalities(PConsLst_tmp,IntVars,PConsLst),!
		;
		    PConsLst=[false]
		)
	    )
	).

% MODE: eliminate_useless_equalities(+PConsLst_tmp,+IntVars,-PConsList)
% TYPE: eliminate_useless_equalities(list(constraint),list(var),list(constraint))
eliminate_useless_equalities([],_,[]).
eliminate_useless_equalities([C|Cs],IntVars,NewCs) :-
	C = (X=:=Y),
	((var(X), is_not_member(X,IntVars),Y=X);(var(Y), is_not_member(Y,IntVars),X=Y)),
	eliminate_useless_equalities(Cs,IntVars,NewCs).
eliminate_useless_equalities([C|Cs],IntVars,[C|NewCs]) :-
	eliminate_useless_equalities(Cs,IntVars,NewCs).

% convert a list of constraints into a clpr constraint
list_to_constraint(clpr,L,C) :-
	( L=[] -> C={1=1}
	;
	    list_to_clpr_constraint(L,C)
	).

list_to_clpr_constraint([X],{X}).
list_to_clpr_constraint([X,Y|Xs],{X,CXs}):-
	list_to_clpr_constraint([Y|Xs],{CXs}).

get_singletons([],[],[],[],[]).
get_singletons([VC|VCs],[VC1|VC1s],P,DV,DV1) :-
	(ground(VC1) ->
	    % build equality
	    functor(PC,=:=,2),
	    arg(1,PC,VC),
	    arg(2,PC,VC1),
	    P=[PC|PCs],
	    get_singletons(VCs,VC1s,PCs,DV,DV1)
	;
	    % collect vars to dump on
	    DV=[VC|DVs],DV1=[VC1|DV1s],
	    get_singletons(VCs,VC1s,P,DVs,DV1s)
	).

fix_equalities([C|Cs],[C1|C1s]) :-
	fix_equality(C,C1),
	fix_equalities(Cs,C1s).
fix_equalities([],[]).
fix_equality(LHS=RHS,LHS=:=RHS) :- !.
fix_equality(C,C).


entails(CL,DL) :-
	!,
	((CL==[]) -> (DL=[];DL=[1=1])
	;
	    ( (DL==[]) ->
		( true )
	    ;
		( (\+ satisfiable(clpr,DL)) ->
		    (CL=[false])
		;

		    (satisfiable(clpr,CL) ->
			( list_to_constraint(clpr,CL,C),
			 list_to_constraint(clpr,DL,D),
			 entails_clpr(C,D) )
		    ;
			( true )
		    )
		)
	    )
	).



% entails_clpr(C,D): C entails D 
%  - C,D are constraints of the form {c1,...,cn}
%
% Note for clpr: (does this apply also to  other domains?)
% copy_term needed to not bind variables to rationals
% this would cause trouble with   numbervars/3

entails_clpr(C,D) :-
	(copy_term(C,C1),\+C1 -> true
	;
	% entailed does not like curly brackets
	    (copy_term((C,D),(C2,{D2})),
		C2,entailed(D2) -> true
	    ;
		fail)
	).


equivalent_constraints(Cns1,Cns2) :-
	entails(Cns1,Cns2),
	entails(Cns2,Cns1).

%%% primitives(C1,C2)
%%% C2 is the result of transforming equalities in C1 to a conjunction of inequalities
primitives([],[]).
primitives([C|Cs],PC) :-
	(functor(C,'=:=',2) ->
	    arg(1,C,LHS), arg(2,C,RHS),
	    PC1 =.. ['=<',LHS,RHS], PC2 =.. ['=<',RHS,LHS],
	    PC=[PC1,PC2|PCs]
	;
	    PC=[C|PCs]

	),
	primitives(Cs,PCs).
