%:- module(solver_clpr,[solve/3,satisfiable/1,entails/2,solve_constraints/2,has_satisfiable_constraint/1]).

:- use_module(library(clpr)).

% exports:
% solve(Cns1,Vars,Cns2)
% satisfiable(C)
% entails(Cns1,Cns2)
% solve_constraints(Clause1,Clause2)
% has_satisfiable_constraint(Clause)

:- dynamic projection_temporary/1.

solve(Cns,Vars,Cns1):-
	solve(clpr,Cns,Vars,Cns1).

solve_constraints((Hd,bd(Cns,Lits)),(Hd,bd(Cns1,Lits))) :-
	commonvars(Cns,(Hd,Lits),CVars),
	solve(clpr,Cns,CVars,Cns1).

has_satisfiable_constraint((_,bd(Cns,_))) :- satisfiable(clpr,Cns).

satisfiable(C) :-
	satisfiable(clpr,C).

satisfiable(clpr,C) :-                    %from F.  solve_constraints.pl
	solve(clpr,C,[],C1),
	(member_chk(false,C1) ->
		fail;
		true).

commonvars(E1,E2,CVars) :-
        setof(X,(
        term_variables(E1,V1),
        term_variables(E2,V2),
        member(X,V1),
        member_chk(X,V2)), LV) -> CVars=LV ; CVars=[].

% member_chk(+X,+Ys) :- there is a member Y of the list of terms Ys
% (possibly with variables) such that X==Y. No variable is bound.

member_chk(X,[Y|_]) :- X==Y, !.
member_chk(X,[_|Ys]) :- member_chk(X,Ys).


%%% From F. solvers.pl

%%% solve(Solver,C,IntVars,PC):
%%% PC is the projection of C over the variables in the list IntVars
%%% using solver Solver
%%%  - C,PC are constraints
%%%  - T is a term
%%% IMPORTANT NOTICE: vars(IntVars) must be included in vars(C)

solve(clpr,ConsLst, IntVars, PConsLst) :-
	!,
	(ConsLst==[] ->
		 PConsLst=[]
	;
	    ( member_chk(false,ConsLst) ->
			  PConsLst=[false]
	    ;
		list_to_constraint(clpr,ConsLst,C),
		copy_term((C,IntVars),(C1,IntVars1)),
		retractall(projection_temporary(_)),
		assert((projection_temporary(IntVars1) :- C1)),
		( call_residue(projection_temporary(IntVars1), _R1) ->
		    get_singletons(IntVars,IntVars1,Eqs,DumpVars,DumpVars1),
		    C1,dump(DumpVars1,DumpVars,D),
		    fix_equalities(D,DF), 
		    append(Eqs,DF,PConsLst)
		;
		    PConsLst=[false]
		)
	    )
	).

list_to_constraint(clpr,L,C) :-
	( L=[] -> C={1=1}
	;
	    list_to_clpr_constraint(L,C)
	).

list_to_clpr_constraint([X],{X}).
list_to_clpr_constraint([X,Y|Xs],{X,CXs}):-
	list_to_clpr_constraint([Y|Xs],{CXs}).

get_singletons([],[],[],[],[]).
get_singletons([VC|VCs],[VC1|VC1s],P,DV,DV1) :-
	(ground(VC1) ->
	    % build equality
	    functor(PC,=:=,2),
	    arg(1,PC,VC),
	    arg(2,PC,VC1),
	    P=[PC|PCs],
	    get_singletons(VCs,VC1s,PCs,DV,DV1)
	;
	    % collect vars to dump on
	    DV=[VC|DVs],DV1=[VC1|DV1s],
	    get_singletons(VCs,VC1s,P,DVs,DV1s)
	).

fix_equalities([C|Cs],[C1|C1s]) :-
	fix_equality(C,C1),
	fix_equalities(Cs,C1s).
fix_equalities([],[]).
fix_equality(LHS=RHS,LHS=:=RHS) :- !.
fix_equality(C,C).



entails(CL,DL) :-
	!,
	(CL==[] -> DL=[]
	;
	    (DL==[] ->
		true
	    ;
		(\+ satisfiable(clpr,DL) ->
		    CL=[false]
		;

		    (satisfiable(clpr,CL) ->
			list_to_constraint(clpr,CL,C),
			list_to_constraint(clpr,DL,D),
			entails_clpr(C,D)
		    ;
			true
		    )
		)
	    )
	).


% entails_clpr(C,D): C entails D 
%  - C,D are constraints of the form {c1,...,cn}
%
% Note for clpr: (does this apply also to  other domains?)
% copy_term needed to not bind variables to rationals
% this would cause trouble with   numbervars/3

entails_clpr(C,D) :-
	(copy_term(C,C1),\+C1 -> true
	;
	% entailed does not like curly brackets
	    (copy_term((C,D),(C2,{D2})),
		C2,entailed(D2) -> true
	    ;
		fail)
	).


equivalent_constraints(Cns1,Cns2) :-
	entails(Cns1,Cns2),
	entails(Cns2,Cns1).

%%% primitives(C1,C2)
%%% C2 is the result of transforming equalities in C1 to a conjunction of inequalities
primitives([],[]).
primitives([C|Cs],PC) :-
	(functor(C,'=:=',2) ->
	    arg(1,C,LHS), arg(2,C,RHS),
	    PC1 =.. ['=<',LHS,RHS], PC2 =.. ['=<',RHS,LHS],
	    PC=[PC1,PC2|PCs]
	;
	    PC=[C|PCs]

	),
	primitives(Cs,PCs).
