% MODULE: specialize_ctlf
% provides: specialize_ctl/5
% requires: library(lists) [append],
%           library(terms),
%           library(ordsets),
%           basic_operations [hd_pred,atom_pred],
%           solver_clpr,
%	    remove_unfold [sel_elem,cfact_subsumes_clause_chk]

%:- module(specialize_ctlf, [specialize_ctl/5]).

% :- use_module(library(lists)). %[append]
:- use_module(library(terms)).
:- use_module(library(ordsets)).
:- use_module(basic_operations).


% TYPES
%
% Base types: term, predicate, nat, rat, any.
% Type constructors: list/1.
% Type definitions:
%    atom == predicate(term1,...,termN).
%    literal == atom; \+ atom.
%    clause == (atom, bd(list(constraint),list(literal))).
%    clause_tree = list((atom,list(list(clause)))).
%    arity == nat.
%    def == (predicate/arity,list(clause)).
%    constraint == rat=<rat; rat>=rat; rat<rat; rat>rat; rat=:=rat; rat=\=rat.
%    bit == 0; 1.
%    uselection == leftmost; ctl.
%    fselection == ctl.


% MODE: specialize_ctl(+Cls1,+Prog1,+Defs1, -Prog2,-Defs2)
% TYPE: specialize_ctl(list(clause),list(def),clause_tree,list(def),clause_tree)
% SEMANTICS: Program Prog2 and definitions Defs2 are obtained by applying the
% specialization strategy starting from the list Cls1 of clauses, the program Prog1,
% and the list Defs1 of definitions.

specialize_ctl(Cls1,Prog1,Defs1,Prog2,Defs2) :-
	specialize_ctl_cls(Cls1,Prog1,Defs1,Cls2,Defs2,1),
	preds_in_cls(Cls2,[],Preds),
	cls_to_defs(Preds,Cls2,Prog2).


% MODE: specialize_ctl_cls(+Cls1,+Prog1,+Defs1, -Cls2,-Defs2,+N)
% TYPE: specialize_ctl_cls(list(clause),list(def),clause_tree, list(def),clause_tree,nat)
% SEMANTICS: The list Cls2 of clauses and the definitions Defs2 are obtained by applying the
% specialization strategy starting from the list Cls1 of clauses, the program Prog1,
% and the list Defs1 of definitions.
% N counts the number of recursive calls of the strategies

specialize_ctl_cls([],_Prog,Defs,[],Defs,_).
specialize_ctl_cls([Cl|Cls],Prog_in,Defs_in,Cls_out,Defs_out,N) :-
	write(N), nl, N1 is N+1,
	ctl_unfold(Cl,Prog_in,UnfoldedCls),
	remove_subsumed(UnfoldedCls,SCls),
         (derivation_trace_on ->
          (write(' -unfold- '), nl, show_cls(SCls), nl);
          true),
 	ctl_define_fold(SCls,Defs_in,FoldedCls,Defs1,NewDefs),
         (derivation_trace_on ->
          (write(' -define- '), nl, show_cls(NewDefs),  nl,
	   write(' -fold- '), nl, show_cls(FoldedCls), nl);
          true),
	append(NewDefs,Cls,Cls1),
	append(FoldedCls,Cls_out1,Cls_out),
	specialize_ctl_cls(Cls1,Prog_in,Defs1,Cls_out1,Defs_out,N1).


% UNFOLD SUBSIDIARY STRATEGY

% MODE: ctl_unfold(+Cl,+Prog, -UCls)
% TYPE: ctl_unfold(clause,list(def), list(clause))
% SEMANTICS: UCls is a list of clauses obtained by unfolding clause Cl
% using program Prog.

ctl_unfold(Cl,Prog,UCls) :-
	unfold_clause(leftmost,Cl,Prog,Cls), !,
	repeat_unfold(ctl,Cls,Prog,UCls).
ctl_unfold(Cl,_Prog,UCls) :-
	UCls=[Cl].


% MODE: repeat_unfold(+Uselection,+Cls1,+Prog, -Cls2)
% TYPE: repeat_unfold(uselection,list(clause),list(def),list(clause))
% SEMANTICS: Cls2 is a list of clauses obtained by repeatedly applying
% the unfolding rule according to the selection rule Uselection
% starting from the list Cls1 of clauses and using program Prog.

repeat_unfold(_Uselection,[],_Prog,[]) :- !.
repeat_unfold(Uselection,[Cl|Cls_in],Prog,Cls_out) :-
	unfold_clause(Uselection,Cl,Prog,UCls),!,
	append(UCls,Cls_in,Cls_in1),
	repeat_unfold(Uselection,Cls_in1,Prog,Cls_out).
repeat_unfold(Uselection,[Cl|Cls_in],Prog,Cls_out) :-
	Cls_out=[Cl|Cls_out1],
	repeat_unfold(Uselection,Cls_in,Prog,Cls_out1).


% MODE: unfold_clause(+Uselection,+Cl,+Prog, -Cls)
% TYPE: unfold_clause(uselection,clause,list(def),list(clause))
% SEMANTICS: Cls is a list of clauses obtained by unfolding once
% clause Cl using program Prog according to the selection rule Uselection.

unfold_clause(Uselection,Cl,Prog,Cls) :-
	unfolding_selection(Uselection,Cl,Cl1),
	pos_unfold(Cl1,Prog,Cls).


% MODE: unfolding_selection(+Uselection,+Cl, -ClSel)
% TYPE: unfolding_selection(uselection,clause,clause_sel)
% SEMANTICS: ClSel is obtained from clause Cl by splitting
% the list of body literals of Cl into two sublists LLits and [S|RLits],
% according to the rule Uselection. The first literal S of the second
% sublist is the (positive) literal selected for unfolding.

unfolding_selection(leftmost,(H,bd(Cns,[A|Lits])),(H,bd(Cns,[],[A|Lits]))) :- !,
	\+ A=(\+ _).
unfolding_selection(ctl,(H,bd(Cns,[A|Lits])),(H,bd(Cns,LLits,[S|RLits]))) :-
	 unfoldable(ctl,A), !,
	(LLits=[], S=A, RLits=Lits).
unfolding_selection(ctl,(H,bd(Cns,[A|Lits])),(H,bd(Cns,LLits,[S|RLits]))) :-
	LLits=[A|LLits1],
 	unfolding_selection(ctl,(H,bd(Cns,Lits)),(H,bd(Cns,LLits1,[S|RLits]))).


% MODE: unfoldable(+Uselection,+L)
% TYPE: unfoldable(uselection,literal)
% SEMANTICS: Every clause is unfoldable wrt literal L selected according to
% the selection rule Uselection

unfoldable(ctl,t(_,_)).
unfoldable(ctl,ts(_,_)).
unfoldable(ctl,elem(_,_)).
unfoldable(ctl,sat(_,F)) :- \+ F=ef(_), \+ F=af(_), \+ F=eu(_,_), \+ F=au(_,_).
unfoldable(ctl,sat_all(_,_)).
unfoldable(ctl,mem(_,_)).
unfoldable(ctl,exists_state(_)).
unfoldable(ctl,prop(_)).


% MODE: pos_unfold(+ClSel,+Prog, -Cls)
% TYPE: pos_unfold(clause_sel,list(def), list(clause))
% SEMANTICS: Cls is a list of clauses obtained by unfolding once
% clause ClSel wrt the selected atom using program Prog.

pos_unfold((H,bd(C,G,[A|Ls])),Prog,Cls) :-
	literal_pred(A,P),
	def(P,Prog,DefP),
	resolvents((H,bd(C,G,[A|Ls])),DefP,Cls).


% MODE: resolvents(+Cl,+Cls, -Res)
% TYPE: resolvents(clause_sel,list(clause), list(clause))
% SEMANTICS: Res is s list of clauses obtained by resolving Cl
% with each clause in Cls wrt a selected literal in the body of Cl

resolvents(_Cl,[],[]).
resolvents(Cl1,[Cl2|Cls],Res) :-
	copy_term(Cl1,(H1,bd(C1,G1,[A1|Ls1]))),
	copy_term(Cl2,(H2,bd(C2,G2))),
	A1=H2,
	append(C1,C2,C3),
	term_variables((H1,G1,G2,Ls1),V),
	solve(C3,V,C4),
	\+ C4=[false], !,
	append(G1,G2,G1G2),
	append(G1G2,Ls1,G1G2Ls1),
	Res=[(H1,bd(C4,G1G2Ls1))|Res1],
	resolvents(Cl1,Cls,Res1).
resolvents(Cl1,[_Cl2|Cls],Res) :-
	resolvents(Cl1,Cls,Res).


% REMOVE_SUBSUMED SUBSIDIARY STRATEGY

% MODE: remove_subsumed(+Def1,-Def2)
% TYPE: remove_subsumed(list(clause),list(clause))
% SEMANTICS: Def2 is obtained by removing from Def1 all clauses
% subsumed by a constrained fact in Def1.

remove_subsumed([],[]).
remove_subsumed([Cl1|Cls],Ds) :-
        sel_elem(Cl2,Cls,Cls1),
        cfact_subsumes_clause_chk(Cl1,Cl2), !,
        remove_subsumed([Cl1|Cls1],Ds).
remove_subsumed([Cl1|Cls],Ds) :-
        member(Cl2,Cls),
        cfact_subsumes_clause_chk(Cl2,Cl1), !,
        remove_subsumed(Cls,Ds).
remove_subsumed([Cl|Cls],[Cl|Ds]) :-
        remove_subsumed(Cls,Ds).


% MODE: sel_elem(-X,+Xs,-Ys)
% TYPE: sel_elem(any,list(any),list(any))
% SEMANTICS: Ys is obtained by deleting the element X from Xs
% IMPORTED FROM: remove_unfold

% MODE: cfact_subsumes_clause_chk(+Cl1,+Cl2)
% TYPE: cfact_subsumes_clause_chk(clause,clause)
% SEMANTICS: Cl1 is a constrained fact and Cl1 subsumes Cl2
% IMPORTED FROM: remove_unfold


% DEFINE_FOLD SUBSIDIARY STRATEGY

% MODE: ctl_define_fold(+Cls,+Defs_in, -FoldedCls,-Defs_out,-NewDefs_out)
% TYPE: ctl_define_fold(list(clause),clause_tree,list(clause),clause_tree,list(clause))
% SEMANTICS:
% - FoldedCls is the list of clauses obtained by folding all clauses in Cls wrt
%   foldable literals by using definition clauses in Defs_in or
%   by introducing the new definition clauses NewDefs_out
% - Defs_out = Defs_in U NewDefs_out

ctl_define_fold(Cls,Defs_in,FoldedCls,Defs_out,NewDefs_out) :-
	repeat_define_fold(ctl,Cls,Defs_in,[],FoldedCls,Defs_out,NewDefs_out).


% MODE: repeat_define_fold(+Fselection,+Cls,+Defs_in,+NewDefs_in, -FoldedCls,-Defs_out,-NewDefs_out)
% TYPE: repeat_define_fold(fselection,list(clause), clause_tree,list(clause),clause_tree,list(clause))
% SEMANTICS:
% - FoldedCls is the list of clauses obtained by folding all clauses in Cls wrt literals selected
%   via Fselection by using definition clauses in Defs_in or
%   by introducing a list, say NewDefs, of new definition clauses
% - NewDefs_out = NewDefs U NewDefs_in
% - Defs_out = Defs_in U NewDefs_out

repeat_define_fold(_Fselection,[],Defs,NewDefs,[],Defs,NewDefs).
repeat_define_fold(Fselection,[Cl|Cls],Defs_in,NewDefs_in,FoldedCls,Defs_out,NewDefs_out) :-
	define_fold_clause(Fselection,Cl,Defs_in,NewDefs_in,FCl,Defs1,NewDefs1),!,
	Cls1=[FCl|Cls],
	repeat_define_fold(Fselection,Cls1,Defs1,NewDefs1,FoldedCls,Defs_out,NewDefs_out).
repeat_define_fold(Fselection,[Cl|Cls],Defs_in,NewDefs_in,FoldedCls,Defs_out,NewDefs_out) :-
	FoldedCls=[Cl|FoldedCls1],
	repeat_define_fold(Fselection,Cls,Defs_in,NewDefs_in,FoldedCls1,Defs_out,NewDefs_out).


% MODE: define_fold_clause(+Fselection,+Cl,+Defs_in,+NewDefs_in, -FoldedCl,-Defs_out,-NewDefs_out)
% TYPE: define_fold_clause(fselection,clause,clause_tree,list(clause),clause,clause_tree,list(clause))
% SEMANTICS:
% - FoldedCl is the clause obtained by folding clause Cl wrt all literals selected
%   via Fselection by using definition clauses in Defs_in or
%   by introducing a list, say NewDefs, of new definition clauses
% - NewDefs_out = NewDefs U NewDefs_in
% - Defs_out = Defs_in U NewDefs_out

define_fold_clause(Fselection,Cl,Defs_in,NewDefs_in,FoldedCl,Defs_out,NewDefs_out) :-
	folding_selection(Fselection,Cl,Cl1),
	define_clause(Cl1,Defs_in,NewDefs_in,Def,Defs_out,NewDefs_out),
	fold_clause(Cl1,Def,FoldedCl).


% MODE: folding_selection(+Fselection,+Cl, -Cl_sel)
% TYPE: folding_selection(fselection,clause, clause_sel)
% SEMANTICS: Cl_sel is a clause obtained by splitting the list G of body literals
% of Cl into two parts: G=G1@[L|G2], where L is the selected literal via
% the folding selection rule Fselection

folding_selection(ctl,(H,bd(Cns,[L|Lits])),(H,bd(Cns,LLits,[SL|RLits]))) :-
	foldable(ctl,L),!,
	LLits=[], SL=L, RLits=Lits.
folding_selection(ctl,(H,bd(Cns,[L|Lits])),(H,bd(Cns,LLits,[SL|RLits]))) :-
	LLits=[L|LLits1],
	folding_selection(ctl,(H,bd(Cns,Lits)),(H,bd(Cns,LLits1,[SL|RLits]))).


% MODE: foldable(+Fselection,+L)
% TYPE: foldable(fselection,literal)
% SEMANTICS: L can be folded according to the folding selection rule Fselection.

foldable(ctl,sat(_,_)).
foldable(ctl,exists_state(_)).
foldable(ctl,L) :- L=(\+ A), foldable(ctl,A).


% MODE: define_clause(+Cl,+Defs_in,+NewDefs_in, -Def,-Defs_out,-NewDefs_out)
% TYPE: define_clause(clause_sel,clause_tree,list(clause), clause,clause_tree,list(clause))
% SEMANTICS: Given a clause Cl with a selected literal L in its body,
% a tree Defs_in of definition clauses, and a list NewDefs_in of new definition clauses,
% - Def is either a clause in Defs_in or a new definition clause that can be used
%   to fold Cl wrt L
% - Defs_out = if Def in Defs_in then Defs_in else [Def] U Defs_in
% - NewDefs_out = if Def in Defs_in then NewDefs_in else [Def] U NewDefs_in

define_clause(Cl,Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out) :-
	Cl=(H,bd(Cns,_,[L|_])),
	candidate_def(Cns,L,CD),
	atom_pred(H,Pred),
	ancestor_defs(Pred,Defs_in,As),
	introduce_def(CD,As,Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out).


% MODE: candidate_def(+Cns,+L, -CD)
% TYPE: candidate_def(list(constraint),literal,clause)
% SEMANTICS: CD is a clause of the form: newp(X1,...,Xn) :- Constraints, Atom
% where: newp is a new predicate symbol,
% Atom is the atom occurring in the literal L,
% X1,...,Xn are the variables occurring in Atom, and
% Constraints is the projection of Cns wrt X1,...,Xn

candidate_def(Cns,L,CD) :-
	copy_term((Cns,L),(Cns1,L1)),
	gen_new_name_functor(Pred),
	term_variables(L1,V),
	solve(Cns1,V,Cns2),
	NewH=..[Pred|V],
	atom(L1,A),
	CD=(NewH,bd(Cns2,[A])).


% MODE: introduce_def(+CD,+As,+Defs_in,+NewDefs_in, -Def,-Defs_out,-NewDefs_out)
% TYPE: introduce_def(clause,list(clause),clause_tree,list(clause),clause,clause_tree,list(clause))
% SEMANTICS:
% if     there exists a definition clause D in Defs_in which is more general than CD
% then   Def=D, Defs_out=Defs_in, NewDefs_out=NewDefs_in
%        Defs_out is the tree of definition clauses obtained by adding Def as a leaf of Defs_in
%        with ancestors As=As1@[D|As2], and
%        NewDefs_out=[Def|NewDefs_in]
% else   Def=CD, Defs_out is obtained by adding Def as a leaf of Defs_in, and NewDefs_out=[Def|NewDefs_in]

introduce_def(CD,_As,Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out):-
	CD=(_H,bd(_Cns,[A])),
	copy_term(A,A1),
	select_cls(A1,Defs_in,ClsA1),
	member(D,ClsA1),
	more_general_def(D,CD), !,
	Def=D,
	Defs_out=Defs_in,
	NewDefs_out=NewDefs_in.
%introduce_def(CD,As,Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out) :-
%	similar_anc(CD,As,Ancs),
%	Ancs=(As1,[D|As2]),
%	embedded_clause(D,CD), !,
%	write('generalize'), nl,
%	generalize(CD,(As1,[D|As2]),Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out).
introduce_def(CD,As,Defs_in,NewDefs_in, Def,Defs_out,NewDefs_out):-
	Def=CD,
	CD=(_,bd(_,[A])),
	copy_term(A,A1),
	add_def((A1,[Def|As]),Defs_in,Defs_out),
	NewDefs_out=[Def|NewDefs_in].


% MODE: select_cls(+A,+Defs, -ClsA)
% TYPE: select_cls(atom,clause_tree,list(clause))
% SEMANTICS: ClsA is the list of clauses indexed by A in Defs

select_cls(_,[],[]).
select_cls(A,[(A1,Def1)|_Defs],ClsA) :- A=A1, !, select_heads(Def1,ClsA).
select_cls(A,[_|Defs],ClsA) :- select_cls(A,Defs,ClsA).


% MODE: select_heads(+LL, -L)
% TYPE: select_heads(list(list(any)),list(any))
% SEMANTICS: L is the list obtained from the list LL of lists
% by taking all first elements of the elements of LL

select_heads([],[]).
select_heads([[H|_]|Ls],[H|T]) :- select_heads(Ls,T).


% MODE: more_general_def(+Cl1, +Cl2)
% TYPE: more_general_def(clause,clause)
% SEMANTICS: bd(Cl1) is a variant of bd(Cl2) via a substitution \rho,
% the constraint in Cl2 entails the constraint in Cl1,
% and the variables in hd(Cl2)\rho are a suset of those in hd(Cl1)

more_general_def((H1,bd(Cns1,A1)),(H2,bd(Cns2,A2))) :-
	variant(A1,A2),
	A1=A2,
	entails(Cns2,Cns1),
	term_variables(H1,Vars1),
	term_variables(H2,Vars2),
	subset_chk(Vars2,Vars1).


% MODE: subset_chk(?Ts1,?Ts2)
% TYPE: subset_chk(list(term),list(term))
% SEMANTICS: each term in Ts1 occurs in Ts2

subset_chk([],_).
subset_chk([T1|Ts1],Ts2) :-
	member_chk(T1,Ts2),
	subset_chk(Ts1,Ts2).


% MODE: ancestor_defs(+Pred,+Defs_in, -As)
% TYPE: ancestor_defs(predicate/arity,clause_tree,list(clause))
% SEMANTICS: As is the list of ancestor clauses in the tree of clauses
% Defs_in of the definition clause which has head predicate Pred

ancestor_defs(Pred,[(_,Ls)|_],As) :-
	a_defs(Pred,Ls,As), !.
ancestor_defs(Pred,[_|Defs],D) :-
	ancestor_defs(Pred,Defs,D).


% MODE: a_defs(+Pred,+Clss, -As)
% TYPE: a_defs(predicate/arity,list(list(clause)),list(clause))
% SEMANTICS: As is a list of clauses in the list Clss of lists of
% clauses, such that the head predicate of the first clause of Clss
% is Pred

a_defs(Pred,[[Cl|Cls]|_],As) :-
	hd_pred(Cl,Pred), !,
	As=[Cl|Cls].
a_defs(Pred,[_|Ls],As) :-
	a_defs(Pred,Ls,As).


% MODE: fold_clause(+Cl,+Def, -FoldedCl)
% TYPE: fold_clause(clause_sel,clause,clause)
% SEMANTICS: FoldedCl is the result of folding Cl wrt a selected literal
% using definition Def

fold_clause(Cl,Def,FoldedCl) :-
	copy_term(Cl,(H,bd(Cns,LLits,[L|RLits]))),
	copy_term(Def,(H1,bd(_,[A1]))),
	replace_literal(L,A1,H1,L1),
	append(LLits,[L1|RLits],NewLits),
	FoldedCl=(H,bd(Cns,NewLits)).


% MODE: replace_literal(+L,+A1,+H1, -L1)
% TYPE: replace_literal(literal,atom,atom,literal)
% SEMANTICS: if L is unifiable with A1 then unify H1 and L1
% else if L is unifiable with \+ A1 then unify \+ H1 and L1

replace_literal(L,A1,H1,L1) :-
	L=A1, !, L1=H1.
replace_literal(L,A1,H1,L1) :-
	L= (\+ A1), L1= (\+ H1).


% MODE: add_def(+Def,+Defs1, -Defs2)
% TYPE: add_def(def,clause_tree,clause_tree)
% SEMANTICS: Def of the form (A,Cls)
% and Defs2=[...,(A,[Cls|Cls1s]),...] if Defs1=[...,(A,Cls1s),...]
% Defs2=[...,(A,[Cls])] if no (A,Cls1s) occurs in Defs1

add_def((A,Cls),[],[(A,[Cls])]).
add_def((A,Cls),[(A1,Cls1s)|Defs1],Defs2) :-
	A=A1, !,
	Defs2=[(A1,[Cls|Cls1s])|Defs1].
add_def((A,Cls),[(A1,Cls1s)|Defs1],[(A1,Cls1s)|Defs2]) :-
	add_def((A,Cls),Defs1,Defs2).
