% struct.pl
%
% This file contains predicates which manipulate the data structures.
%
% A transformation tree is represented by terms of the form node(N,M,S,R,L),
% where N is a node number (1 for the root), M is the node number of the
% father of N (0 for the father's root), S is the list of the node numbers
% which are sons of M, R is a term identifying the rule applied on the program
% at node M to produce the program at node N (0 for the root), and L is the
% list of the clauses' numbers in the program at node N. A clause is
% represented by a term of the form clause3(N,C,Aux), where C is a term
% identifying a clause, N is a clause number, and Aux a parameter set to `prog'
% for all clauses in the intitial program. C is of the form (Hd,Bd) where Hd
% and Bd are an atom and a list of atoms respectively.

:- use_module(library(lists), [ append/3, member/2 ]).
:- use_module(library(terms), [ variant/2]). %, unify_with_occurs_check/2 ]).

label_arc(Node,LA,SN) :- node(SN,Node,_,LA,_).

label_arc(Node,LA) :- node(Node,_,_,LA,_).

label_node(Node,LN) :- node(Node,_,_,_,LN).

type_pnat(N) :- integer(N), N>0.

type_clause(H) :- \+(H=(_:-_)), \+(number(H)), (type_list(H) -> H=[];true).
type_clause(C) :-
   C=(Head:-Body), \+(number(Head)), (type_list(Head) -> Head=[];true), type_goal((Body)).

type_goal(G) :- \+(G=(_,_)), \+(number(G)), (type_list(G) -> G=[];true).
type_goal(G) :- G=(A,B), \+(number(A)), nonvar(A), (type_list(A)->A=[];true), type_goal(B).

type_list(X) :- X ==[].
type_list(X) :- nonvar(X), X=[_|T], type_list(T).

type_law(L) :- L=(LL->RL), type_goal(LL), type_goal(RL).

clause_number_to_clause(N,C) :- clause3(N,C,_).

predname_and_arity(Head,Pred_Head,Arity) :- functor(Head,Pred_Head,Arity).

head_of_clause(C,Head) :-    clause3(C, (Head,_), _).

body_of_clause(C,Body) :-    clause3(C, (_,Body), _).

head_and_body(C,Hd,Bd) :- clause3(C, (Hd,Bd), _).

is_a_unit_clause(C) :- body_of_clause(C,[]).

continuation_body(C) :- body_of_clause(C,B), is_a_list_of_continuations(B).
is_a_list_of_continuations([]).
is_a_list_of_continuations([C|Cs]) :-
	is_a_continuation(C),
	is_a_list_of_continuations(Cs).
is_a_continuation('@'(_)).

is_a_tautology(N) :-
   clause_number_to_clause(N,(H1,Bd)),
   Bd==[H1], !.

%is_an_equality(Atom) :- nonvar(Atom), Atom = (_=_).

is_an_equality(Atom,LHS,RHS) :- Atom = (LHS=RHS).

is_a_conjunction((_,_)).

is_a_negative_goal(C,A) :- is_a_negative_goal(C,A,_).
is_a_negative_goal(C,A,Atom) :-
   clause_number_to_clause(C,(_,Bd)),
   pos1(Lit,Bd,A),
   is_a_negative_literal(Lit,Atom).

is_a_negative_literal(Lit,Atom) :- nonvar(Lit), Lit = (not(Atom)).
%is_a_negative_literal(Lit,Atom) :- nonvar(Lit), Lit = (\+ Atom).
is_a_negative_literal(Lit,Atom) :- nonvar(Lit), Lit = (\+(Atom)).

is_a_negation(not/1).
is_a_negation((\+)/1).
current_clauses(Cls) :- current_prog(_,Cls).

current_program(P) :- current_prog(P,_).
initial_program(Cls) :- node(1,_,_,_,Cls).
number_clause(Num) :- number(Num).

length_body_clause(Cl,L) :- body_of_clause(Cl,B), length(B,L).

check_exists_prog_in_transf_tree(P) :- label_node(P,_), !.
check_exists_prog_in_transf_tree(P) :- msg_unknown_program(P).

check_exists_transf_tree :- clause(clause3(_,_,_),_), !.
check_exists_transf_tree :- msg_failure_no_selection.

check_exists_def :-
   label_arc(Prog,(def,_)), current_program(Current),
   check_ancestor(Prog,Current),!.
check_exists_def :-
   label_arc(Prog,(gen_def,_)), current_program(Current),
   check_ancestor(Prog,Current),!.

check_exists_law :- law(_,_), !.

check_exists_predicate(P/A) :-
     (var(P);var(A)), !, msg_unknown_predicate(P/A), msg_no_action, fail.
check_exists_predicate(P/A) :-
     current_program(PN),
     functor(Pred,P,A), (undefined_predicate(PN,Pred) ->
     (msg_unknown_predicate(P/A), msg_no_action, fail) ; true).

check_ancestor(N,PN) :- ancestors(PN,Ancestors), member(N,[PN|Ancestors]), !.

check_goal(G) :- type_goal(G) -> true; msg_not_a_goal, fail.

check_clause(C) :- type_clause(C) -> true; msg_not_a_clause, fail.

check_law(L) :- type_law(L) -> true; msg_not_a_law, fail.

check_clause_with_non_empty_body(Clause_Number,Length) :- 
    body_of_clause(Clause_Number,B), length(B,Length), (Length > 0 -> true ;
    write('Clause '),  write(Clause_Number), write(' has an empty body !'), fail).

check_head_variable(Head,Body) :-
     (vars_subset(Head,Body) -> true ;
     write('Warning : variables occurring in the head and not in the body.'), nl).

check_input_clause(C) :- 
     check_pnat(C), (clause3(C,(_,_),_) -> true ;
     write(C), write(' is not a clause number.'), fail),
     current_clauses(CNs), (member(C,CNs) -> true ; 
     msg_not_a_clause_in_current_prog(C), fail).

check_input_var_numbers_and_types(CN,Vs,Ts) :-
    same_length(Vs,Ts), !,
    clause_number_to_clause(CN,C),
    free_variables(C,[],[],L),
    length(L,N),
    check_input_var_numbers_and_types2(N,Vs,Ts).

check_input_var_numbers_and_types(_,Vs,Ts) :-
    msg_must_have_same_length(Vs,Ts).

check_input_var_numbers_and_types2(_,[],[]).
check_input_var_numbers_and_types2(N,[V|Vs],[T|Ts]) :-
     ( (integer(V), V > 0, V =< N) -> true ;
       msg_wrong_var_position(V,N)),
     check_input_type(T),
     check_input_var_numbers_and_types2(N,Vs,Ts).

check_input_var_number(CN,V) :- 
     clause_number_to_clause(CN,C),
     free_variables(C,[],[],L),
     length(L,N),
     ( (integer(V), V > 0, V =< N) -> true ;
       msg_wrong_var_position(V,N)).


check_input_type(T) :-
   Atom =.. [T,_],
   current_program(PN),
   (undefined_predicate(PN,Atom) -> 
         msg_undefined_predicate(T/1,PN) ; true).

check_pnat(N) :- type_pnat(N) -> true; msg_not_a_pnat, fail.

check_input_clauses([]) :- nltab, write('Enter a non empty list'), nl.
check_input_clauses([C]) :- !,	check_input_clause(C).
check_input_clauses([C|L]) :- !, check_input_clause(C), check_input_clauses(L).
check_input_clauses(_) :- nltab, write('A list of clause numbers is expected !'), nl, fail.

check_input_atom(C,A) :- 
   check_pnat(A), body_of_clause(C,B), length(B,L),
   ((A =< L, A >=1) -> true ; 
   (L =0 -> (write('Clause '), write(C),
   write(' has an empty body ! No atom can be selected in it for unfolding'), fail) 
   ; (write(A), write(' does not denote a position in the body of '), write(C), nl, 
   write('It must be in [1..'), write(L), write(']'), nl, fail))).

check_input_program(Pj) :-
    current_program(PN), (PN=Pj -> true ; (check_ancestor(Pj,PN) -> true 
    ; nl, write(Pj), write(' is not an ancestor of current program.'), nl, fail)).

check_input_atoms(L,[A]) :- integer(A), A =< L, !.
check_input_atoms(L,[A|As]) :- integer(A), A =< L, !, check_input_atoms(L,As).
check_input_atoms(_,[_|_]) :-!, msg_bad_positions_in_body, fail.
check_input_atoms(_,A) :- msg_must_be_a_list_of_atoms(A), fail.

check_input_clause_to_unfold(C) :- 
   check_pnat(C), (body_of_clause(C,B) -> true ;
   write(C), write(' is not a clause number.'), fail),
   current_clauses(CNs), (member(C,CNs) -> true ;
   write('Clause '), write(C), write(' is not in the current program.'), fail),
   (B=[] -> (write('Clause '), write(C), write(' has an empty body !'), fail) ; true).

check_input_clause_to_unfold(C, Length) :- 
   check_pnat(C), (body_of_clause(C,B) -> true ;  
   write(C), write(' is not a clause number.'), fail),
   current_clauses(CNs), (member(C,CNs) -> true ; 
   write('Clause '),  write(C), write(' is not in the current program.'), fail),
  (B=[] -> (write('Clause '), write(C), write(' has an empty body !'), fail)
  ; length(B,Length)).

check_input_clause_to_fold(C,L) :- check_input_clause_to_unfold(C,L).

check_input_clause_for_goal_rep(C,L) :- check_input_clause_to_unfold(C,L).
check_input_clause_for_move_lit(C,L) :- check_input_clause_to_unfold(C,L), L >1.
check_input_clause_for_goal_deletion(C,L) :- check_input_clause_to_unfold(C,L).

check_input_clause_to_case_split(C) :- 
    check_pnat(C), (clause3(C,(_,_),_) -> true 
    ; write(C), write(' is not a clause number.'), fail),
    current_clauses(CNs), (member(C,CNs) -> true 
    ; write('Clause '),  write(C), write(' is not in the current program.'), fail).

check_new_pred(Head,_,_) :- (check_new_pred(Head) -> true ; fail).
check_new_pred(Head_def) :-
   predname_and_arity(Head_def,F,Arity),
   current_clauses(Cls),
   setof((Head,Body),N^X^(member(N,Cls), clause3(N,(Head,Body),X)), Clauses),
   (setof(Head_Def, N_Def^Body_Def^definition(N_Def, (Head_Def,Body_Def)),
   Heads) -> true  ; Heads=[]),
   not_in_prog(F,Arity,Clauses), not_in_def(F,Arity,Heads).

definition(N,(Head,Body)) :-
    clause3(N,(Head,Body),[def|L]), label_arc(Prog,(def,L)),
    current_program(Current), check_ancestor(Prog,Current).
definition(N,(Head,Body)) :-
    clause3(N,(Head,Body),[gen_def|L]), label_arc(Prog,(gen_def,L)),
    current_program(Current), check_ancestor(Prog,Current).

not_in_prog(F,A,Clauses) :- not_functor_ex(F,A,Clauses), !.
not_in_prog(F,A,_) :- msg_already_used_in_prog(F/A).

not_in_def(F,A,Heads) :- not_functor(F,A,Heads), !.
not_in_def(F,A,_) :- msg_already_used_in_def(F/A).

not_functor_ex(_,_,[]).
not_functor_ex(F,Arity,[C|Cs]) :- 
   C=(Head,Body), functor(Head,F1,Arity1), (\+ F = F1 ; \+ Arity=Arity1), !,
   not_functor(F,Arity,Body), not_functor_ex(F,Arity,Cs).

not_functor(_,_,[]).
not_functor(F,Arity,[B|Bs]) :- 
   functor(B,F1,Arity1), (\+ F = F1 ; \+ Arity=Arity1), !,
   not_functor(F,Arity,Bs).

split_body_continuation(['@'(Cont)],[],Cont) :-!.
split_body_continuation([A|B_Cont],[A|B],Cont) :-
   split_body_continuation(B_Cont,B,Cont).



body_without_inequalities([\+(A=B)|IneqsBd_C],[\+(A=B)|Ineqs],Bd_C):-!,
    body_without_inequalities(IneqsBd_C,Ineqs,Bd_C).
body_without_inequalities(Bd_C,[],Bd_C).

body_without_equalities([A=B|IneqsBd_C],[A=B|Ineqs],Bd_C):-!,
    body_without_equalities(IneqsBd_C,Ineqs,Bd_C).
body_without_equalities(Bd_C,[],Bd_C).

check_instantiation_head_before_unfold_neg(_,_,_,[],[]) :- !.
check_instantiation_head_before_unfold_neg(C,Atom,Pos,[(Hd,Bd)|L],[Bd|L2]) :-
   subsumes_chk(Hd,Atom), !,
   unify_with_occurs_check(Atom,Hd),
   check_instantiation_head_before_unfold_neg(C,Atom,Pos,L,L2).
check_instantiation_head_before_unfold_neg(C,_,Atom_Number,_,_) :-
   msg_neg_lit_not_enough_instantiated(C,Atom_Number).


exists_clause_with_head_unifiable_with(Pi,Atom) :-
   clause_with_head_unifiable_with(Pi,Atom,_,_), !.
clause_with_head_unifiable_with(Pi,Atom,Hd,Bd) :-
    label_node(Pi,Clauses_Pi), 
    member(D,Clauses_Pi), clause_number_to_clause(D,(Hd,Bd)),
    unifiable(Hd,Atom).

clause_with_head_unifiable_with(Atom,Hd,Bd) :-
    clause3(_,(Hd,Bd),prog),
    unifiable(Hd,Atom).

non_failed_body(L) :- current_clauses(L2), non_failed_body(L,L2).

non_failed_body([],_).
non_failed_body([E|R],L) :-
    is_a_continuation(E),
    non_failed_body(R,L).
non_failed_body([E|R],L) :-
    member(X,L), head_of_clause(X,H), \+ \+ E = H, !, non_failed_body(R,L).
non_failed_body([E|R],L) :-
    current_program(PN),
    undefined_predicate2(E,L), predname_and_arity(E,F,A),
    msg_undefined_predicate(F/A,PN), non_failed_body(R,L).

undefined_predicate(Pi,Clause_Number,Atom_Number,F/A) :-
    body_of_clause(Clause_Number,B), pos1(Atom,B,Atom_Number),
    functor(Atom,F,A), undefined_predicate(Pi,Atom).
undefined_predicate(Pi,Atom) :-
    label_node(Pi,L), undefined_predicate2(Atom,L).
%undefined_predicate(Pi,Atom) :-
%    (check_new_pred(Atom) -> fail ; true).

undefined_predicate2('@'(_),_) :- !, fail.
undefined_predicate2(\+(Atom),L) :- !, undefined_predicate2(Atom,L).
undefined_predicate2(not(Atom),L) :- !, undefined_predicate2(Atom,L).
undefined_predicate2(_,[]).
undefined_predicate2(Pred,[X|_]) :-
    functor(Pred,F,A), functor(GenAtom,F,A), head_of_clause(X,H),
    \+ \+ GenAtom = H, !, fail.
undefined_predicate2(Pred,[_|L]) :- undefined_predicate2(Pred,L).

rearrange_definitions([],[]).
rearrange_definitions([D|L],[(Head_def,Body_list)|L2]) :-
   (D=(Head_def :- Body_def) -> conj_to_list((Body_def),Body_list);
                                D=(Head_def,Body_list)), rearrange_definitions(L,L2).
rearrange_definition(D2, (Head,Body_list)) :-
    (D2=(Head:- Body) -> conj_to_list((Body),Body_list); Head=D2, Body_list=[]).

check_one_clause_definition(Def) :- one_clause_definition(Def), !.
check_one_clause_definition(Def) :- 
	msg_not_one_clause_def_for_ts_fold(Def), fail.
one_clause_definition(Def) :- one_clause_definition(_,Def,_).
one_clause_definition(Prog,Def,(Head,Body)) :- 
       current_program(P), definition_clauses(Prog,[Def]),
       check_ancestor(Prog,P), clause_number_to_clause(Def,(Head,Body)).
definition_clauses(Prog,Def) :- label_arc(Prog,(def,Def)).
definition_clauses(Prog,Def) :- label_arc(Prog,(gen_def,Def)).
available_definitions_for_ts_fold(L) :- setof(N, one_clause_definition(N), L), !.
available_definitions_for_ts_fold([]).

heads_unifiable([],H,H).
heads_unifiable([C|L],H,H3) :-
   head_of_clause(C,Hd_C),
   (var(H) -> H2 = Hd_C ; anti_unify(Hd_C,H,H2)), heads_unifiable(L,H2,H3).

available_laws_for_rep(L) :- setof(N, C^law(N,C), L), !.
available_laws_for_rep([]).

is_available_law(Law) :- law(Law,_).

law_line(Law,Law2,Line) :-
   append(_,[45,62|_],Line), last(Line,Line2,46), !, name(Law,Line), name(Law2,Line2).

numb_fetch_clauses([],[]).
numb_fetch_clauses([NC|NCs],[(NC,C)|Cs]) :- fetch_clause(NC,C), numb_fetch_clauses(NCs,Cs).

fetch_clauses([],[]).
fetch_clauses([NC|NCs],[C|Cs]) :- fetch_clause(NC,C), fetch_clauses(NCs,Cs).
fetch_clause(NC,C) :- clause_number_to_clause(NC,C).

select_ancestor(N,PN) :- ancestors_inv(PN,Ancestors), member(N,[PN|Ancestors]).

ancestors(1,[]).
ancestors(Node,C) :- node(Node,F,_,_,_), ancestors(F,N), append(N,[F],C).
ancestors_inv(1,[1]).
ancestors_inv(Node,[Node|C]) :- node(Node,F,_,_,_), ancestors_inv(F,C).


get_predicates_body(L,C,L2) :- body_of_clause(C,Body), get_pred_arity(L,Body,L2).

get_pred_arity(L,[],L).
get_pred_arity(L,[B|Body],L2) :-
   predname_and_arity(B,P,A),
   (member(P/A,L) -> get_pred_arity(L,Body,L2) ; get_pred_arity([P/A|L],Body,L2) ).

no_other_head_unifiable(_,[],_).
no_other_head_unifiable(Clauses_Numbers2,[C|L],A) :- member(C,Clauses_Numbers2),!,
no_other_head_unifiable(Clauses_Numbers2,L,A).
no_other_head_unifiable(Clauses_Numbers2,[C|L],A) :-
   head_of_clause(C,H), (H=A -> fail ; no_other_head_unifiable(Clauses_Numbers2,L,A)).

check_equalities([],[]).
check_equalities([(Hd,Bd)|L1],[(Hd,Bd2)|L]) :-
    true_equalities(Bd,Bd2), !, check_equalities(L1,L).
check_equalities([_|L1],L2) :- check_equalities(L1,L2).

true_equalities([],[]).
true_equalities([\+ (A=B)|L],L2):- \+ unify_with_occurs_check(A,B), true_equalities(L,L2).
true_equalities([\+ (A=B)|_],_):- A==B, !, fail.
true_equalities([E|L],[E|L2]):-	true_equalities(L,L2).

verify_instantiation_body(B1, Bd_D,Hd_D) :-
   anti_unify(B1,Bd_D,S1,S2,T), variant(Bd_D,T), free_variables(Hd_D,[],[],Vars_Hd_D),
   verify_instantiation_var_body(S1, S2, Vars_Hd_D).

verify_instantiation_var_body([],[],_).
verify_instantiation_var_body([X=Y|S1], [X=_|S2], Vars_Hd_D):-
    var(Y), !, verify_instantiation_var_body(S1, S2, Vars_Hd_D).
verify_instantiation_var_body([X=_|S1], [X=Z|S2], Vars_Hd_D):-
    belongs(Z,Vars_Hd_D), !, verify_instantiation_var_body(S1, S2, Vars_Hd_D).

check_list_of_positions(X,XX) :- get_positions(X,Positions), append([91|Positions], [93], XX).
    
get_positions([],[]).
get_positions([X,X2|L], [X|L2]) :-
    integer(X), X < 57, X > 48,
    integer(X2), X2 < 57, X2 > 48, !, get_positions([X2|L],L2).
get_positions([X|L], L3) :-
    integer(X), X < 57, X > 48, !,  get_positions(L, L2),
    (L2=[] -> L3=[X] ; append([X,44],L2,L3)).
get_positions([_|L], L2) :- get_positions(L, L2).

append_conjunction(A,B,A) :- B==true, !.
append_conjunction(B,A,A) :- B==true, !.
append_conjunction(X,C,(A,B,C)) :- nonvar(X), X=(A,B), !.
append_conjunction(A,B,(A,B)).

%make_disjunction(_,A,A) :- A==true!.
%make_disjunction(true,_,true) :- !.
make_disjunction(A,B,(A;B)).


completely_folded((_,B)) :- 
	member(Atom,B), same_functor(Atom,H), clause3(_,(H,_), prog), !, fail.
completely_folded((_,_)).
