% top_level.pl
%
% This file contains the top level predicates of the transformer, i.e. the
% predicates which correspond to commands that can be issued by the user.
% 
% "rules", "commands", and "strategies" are top level commands which show
% to the user the menu of all other top level commands, respectively concerned
% with transformation rules, general facilities commands and strategies.
%
% "rules" shows commands related to the application of transformation rules.
% Issuing such a command has one of the following effect:
%    - initializing a transformation tree to an initial program selected by the
%      user, or restoring a transformation tree which has been saved in the past
%    - creating a new branch in the current transformation tree labelled by the
%      chosen transformation rule. The new node is labelled by the program derived
%      thereby and which becomes the new current program
%    - no effect if the chosen transformation rule does not apply to the current
%      program (for instance the conditions of application are not satisfied or
%      wrong parameters are given)
%    - setting the value of current program to a program in the transformation tree
%     (the transformation tree is unchanged, but the pointer which points to the
%     current program is moved to another program in the tree)
%    - adding or removing a replacement law. The transformation tree is unchanged
%      but the set of available laws which can be selected for applying the
%      replacement goal transformation rule is modified.
%
% "commands" shows commands which do not affect the transformation tree. They are
%  not necessary to perform any transformation step. They provide facilities for
%  consulting or saving information related to transformation states.
%
% "strategies" shows commands associated with strategies, i.e. commands which apply
%  automatically some predefined sequences of transformation rules.
%  There are three available strategies:
%  * lap: Loop Absorption Strategy, which is described in:
%   Proietti, M. and Pettorossi, A.:
%   Construction of Efficient Logic Programs by Loop Absorption and Generalization.
%   Proceedings Meta'90 (M. Bruynooghe, ed.), Leuven, Belgium (1990), p. 57-81.
%  * uve:  Elimination of Unnecessary Variables, which is described in:
%  Proietti, M. and Pettorossi, A.:
%  Unfolding-Definition-Folding, in this order, for Avoiding Unnecessary Variables
%  in Logic Programs. Theoretical Computer Science 142 (1) 1995, p. 89-124. 
%     * determ: Derivation of `semi-deterministic' programs via specialization,
%     as described in:
%     Pettorossi, A., Proietti, M., Renault, S.:
%     Reducing Nondeterminism while Specializing Logic Programs,
%     24th ACM Symposium on Principles of Programming Languages (POPL'97),
%     La Sorbonne, Paris, France, January 15-17, 1997, p. 414-427.
%   
%     "lap" and "uve" are top-level predicates that must be used once the
%     initial program has been selected. They use a menu in order to get the
%     necessary parameters fromn the user. Unlike "lap and "uve", "determ"
%     does not get the necessary parameters by means of a menu. All necessary
%     parameters are supposed to have been set preliminary by the user in
%     the file determ_prog_parameters.pl which is at the root of the
%     MAP directory. Parameters for this strategy include predicate
%     modes, optionnal use of the case split rule, various option for the
%     unfolding substrategy. The name of the initial program (which must be
%     defined in directory Progs/) and the specialized clause with which
%     the transformation process must start are considered as parameters. As a
%     result, "determ" does not require the preliminary selection of an
%     initial program neither a definition introduction transformation step.
% 
%
% Top level commands have an on-line manual entry stored in file man.pl.

:- use_module(library(system), [ file_exists/1 ]).
:- use_module(library(lists), [ append/3,
                                member/2 ]).

%%%%%%%%%%% Available rules and commands

ts_d :- ts_define.
d :- gen_define.

unf :- ud_p.              %Maurizio 16/07/08 - not in manual
unf(C) :- ud_p(C).	  %Maurizio 16/07/08 - not in manual
unf(C,A) :- ud_p(C,A).    %Maurizio 16/07/08 - not in manual

fld(L1,L2) :- f(L1,L2).   %Maurizio 16/07/08 - not in manual

u :- unfold.
u(C) :- unfold(C).
u(C,A) :- unfold(C,A).
uu :- unfold.
uu(C) :- unfold_d(C).
uu(C,A) :- unfold_d(C,A).
uu_p :- unfold_d_p.
uu_p(C) :- unfold_d_p(C).
uu_p(C,A) :- unfold_d_p(C,A).
uu_i :- unfold_d_i.
uu_i(P,C) :- unfold_d_i(P,C).
uu_i(P,C,A) :- unfold_d_i(P,C,A).
ud :- unfold_d.
ud(C) :- unfold_d(C).
ud(C,A) :- unfold_d(C,A).
ud_p :- unfold_d_p.
ud_p(C) :- unfold_d_p(C).
ud_p(C,A) :- unfold_d_p(C,A).
ud_i :- unfold_d_i.
ud_i(P,C) :- unfold_d_i(P,C).
ud_i(P,C,A) :- unfold_d_i(P,C,A).
ud_facts(C) :- unfold_d_facts(C) .
ud_facts :- unfold_d_facts.
ud_eqs(C) :- unfold_d_eqs(C).
ud_eqs :- unfold_d_eqs.
ts_f :- ts_fold.
ts_f(C,As,B) :- ts_fold(C,As,B).
f(L,L2) :- fold(L,L2).
g_f(L,L2) :- rearrange_and_fold(L,L2).
r_f(L,L2) :- rearrange_and_fold(L,L2).
rep :- replace_goal.
rep(A,B,C) :- replace_goal(A,B,C).
mv :- move_lit.
mv(X,Y,Z) :- move_lit(X,Y,Z).
add_g :- add_goal.
add_g(C) :- add_goal(C).
del_g :- delete_goal.
del_g(C,As) :- delete_goal(C,As).
restr(P/A) :- restriction(P/A).
del_cl :- delete_clauses.
del_cl([E|L]) :- !, delete_clauses([E|L]).
del_cl(C) :- delete_clause(C).
del_f_cl :- delete_failing_body_clauses.
del_f_cl(Cs) :- delete_failing_body_clauses(Cs).
add_cl :-add_clause.
rep_cl(C) :- replace_clause(C).
mv_cl(C) :- move_clause(C).
subs :- subsumption.
subs(Cs) :- subsumption(Cs).
r_subs :- r_subsumption.
r_subs(Cs) :- r_subsumption(Cs).
split(C,A) :- split_clause(C,A).
i(C,V,T) :- instantiate_clause(C,V,T).
st :- start.
st(P) :- start(P).
load_s :- load_session.
set(P) :- set_current_program(P).
add_l :- add_law_in_theory.
add_law :- add_law_in_theory.
add_l(T) :- add_law_in_theory(T).
del_l :- delete_law_from_theory.
del_law :- delete_law_from_theory.
del_l(T) :- delete_law_from_theory(T).
sel_l :-  select_law_in_theory.
sel_th :-  select_theory.
sel_t :-  select_theory.
p1 :- show_initial_program.
p0 :- show_initial_program.
p :- show_current_program.
p([P|Ps]) :- !, show_programs([P|Ps]).
p(P) :- show_program(P).
pp([P|Ps]) :- !, show_short_programs([P|Ps]).
pp(P) :- show_short_program(P).
pp :- show_current_short_program.
h :- show_history.
h(P) :- show_history(P).
cls :- show_clauses.
cls(L) :- show_clauses(L).
cc(L) :- show_clauses(L).
defs :- show_definitions.
l :- show_laws.
laws :- show_laws.
directory_progs :- dir_p.
directory_theories :- dir_t.
directory_sessions :- dir_s.
directory_final_progs :- dir_f.
%sv_p :- save_current_program.
s_p :- save_current_program.
s_p(P) :- save_prog(P).
%sv_h :- save_history.
s_h :- save_history.
%sv_h(P) :- save_history(P).
s_h(P) :- save_history(P).
%sv_th :- save_current_theory.
s_th :- save_current_theory.
s_l :- save_current_theory.
%sv_s :- save_session.
s_s :- save_session.
s_der :- save_session.
s_sess :- save_session.
com_h :- insert_comment_history.
comment_history :- insert_comment_history.
r :- rules.
c :- commands.
s :- strategies.
determ :- determ_strat.

% the following commands are not listed in the manual as top level commands
cs(C1,C2) :- case_split(C1,C2). 
g_cs(Cs) :- global_case_split(Cs). 
load_th(T) :- select_theory(T). 
load_th :- select_theory. 
load_l(T) :- select_law_in_theory(T). 
load_l :- select_law_in_theory.
dup(L) :- dup_clause(L).
dup_cl(L) :- dup_clause(L).

/*---------
 ts_define: Introduce a new definition. The user is prompted to enter the new
 definition made of one clause only. It corresponds to the Tamaki and  Sato
 definition. The predicate in the head of the definition must be a  new predicate.
 There is a warning message if the head of the definition uses a variable which
 does not appear in its body.
----------*/
ts_define :-
    check_exists_transf_tree, ask_enter_definition(Head,Body),
    check_head_variable(Head,Body), predname_and_arity(Head,Pred_Head,Arity),
    check_new_pred(Head,Pred_Head,Arity), apply_ts_define(Head,Body).
/*---------
 gen_define: Introduce a new definition. The user is prompted to enter the new
 definition made of possibly several clauses. The predicate in the head of each
 clause must be the same for each clause and must be a new predicate.
---------*/
gen_define :-
   check_exists_transf_tree,
   ask_enter_first_definition_clause(D,Pred,Arity),
   ask_enter_next_definition_clause([D],Pred,Arity,Defs),
   apply_gen_define(Defs).
/*----------
 unfold: Unfold a clause wrt an atom of its body using clauses in initial program.
 The clause number and the atom number are selected by the user via a menu. This
 unfolding rule fails when there is no head of clause of the initial program
 which unifies with the selected atom.
unfold(C): Unfold clause C wrt atom 1 (the leftmost in its body).
unfold(C,A): Unfold clause C wrt atom A.
-----------*/
unfold :- 
   check_exists_transf_tree,
   current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_position_to_unfold(Length,Pos),
   apply_unfold(Clause_Number,Pos).
unfold(Clause_Number) :-  
    check_exists_transf_tree,
    check_input_clause_to_unfold(Clause_Number),
    apply_unfold(Clause_Number,1).   % Default atom is at position 1
unfold(Clause_Number,Atom_Number) :- 
  check_exists_transf_tree,
  check_input_clause_to_unfold(Clause_Number),
  check_input_atom(Clause_Number,Atom_Number),
  apply_unfold(Clause_Number,Atom_Number).
/*----------
 unfold_d: Unfold a clause wrt an atom of its body using clauses in initial
 program. The clause number and the atom number are selected by the user via a
  menu. This unfolding deletes the clause to be unfolded when the selected atom
 is a failing goal. A failing goal is an atom which unifies with no head of
 clause in the current program, and which is defined in the initial program.
 This last condition is to prevent from deleting clauses containing bip's or
 predicates whose definitions have been omitted on purpose by the user because
 they are not necessary for the derivation.
 unfold_d(C): Unfold clause C wrt atom 1 (the leftmost in its body).
 unfold_d(C,A): Unfold clause C wrt atom A.
-----------*/
unfold_d :-
   check_exists_transf_tree, current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_position_to_unfold(Length,Pos), apply_unfold_d(Clause_Number,Pos).
unfold_d(Clause_Number) :- 
   check_exists_transf_tree, check_input_clause_to_unfold(Clause_Number),
   apply_unfold_d(Clause_Number,1). %Default atom is number 1
unfold_d(Clause_Number,Atom_Number) :-
   check_exists_transf_tree, check_input_clause_to_unfold(Clause_Number),
   check_input_atom(Clause_Number,Atom_Number),
   apply_unfold_d(Clause_Number,Atom_Number).
/*-----------
 unfold_d_p/[0,1,2]: same as unfold_d/[0,1,2] except that the clauses of the
 current program are used for unfolding instead of those in the initial program.
-------------*/
unfold_d_p :- 
   check_exists_transf_tree, current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_position_to_unfold(Length,Pos),
   apply_unfold_d_p(Clause_Number,Pos).
unfold_d_p(Clause_Number) :- check_exists_transf_tree,
   check_input_clause_to_unfold(Clause_Number),
   apply_unfold_d_p(Clause_Number,1). %Default atom is number 1
unfold_d_p(Clause_Number,Atom_Number) :- 
  check_exists_transf_tree,
  check_input_clause_to_unfold(Clause_Number),
  check_input_atom(Clause_Number,Atom_Number),
  apply_unfold_d_p(Clause_Number,Atom_Number).
/*-----------
 unfold_d_i/[0,2,3]: same as unfold_d/[0,1,2] except that the clauses of any
 intermediary program may be used for unfolding.
-------------*/
unfold_d_i :-
   check_exists_transf_tree,
   current_prog(PN,CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_position_to_unfold(Length,Pos), 
   ancestors(PN,Ancestors),
   ask_program_for_unfolding(PN,Ancestors,Pi),
   apply_unfold_d_i(Pi,Clause_Number,Pos).
unfold_d_i(Program_Number,Clause_Number) :-
   check_exists_transf_tree,
   check_input_program(Program_Number),
   check_input_clause_to_unfold(Clause_Number),
   apply_unfold_d_i(Program_Number,Clause_Number,1).
unfold_d_i(Program_Number,Clause_Number,Atom_Number) :-
   check_exists_transf_tree,
   check_input_program(Program_Number),
   check_input_clause_to_unfold(Clause_Number),
   check_input_atom(Clause_Number,Atom_Number),
   apply_unfold_d_i(Program_Number,Clause_Number,Atom_Number).
/*--------
 unfold_d_eqs: Unfold equalities in the body of all clauses.
 unfold_d_eqs(C): Unfold equalities in the body of clause C.
----------*/ 
unfold_d_eqs :-
   check_exists_transf_tree,
   current_clauses(CNs),
   unfold_d_eqs_ok(CNs).
unfold_d_eqs([]) :- !.
unfold_d_eqs([C|L]) :- !,
   check_exists_transf_tree,
   unfold_d_eq(C),
   unfold_d_eqs(L).
unfold_d_eqs(C) :- unfold_d_eq(C).
unfold_d_eqs_ok([]).
unfold_d_eqs_ok([C|L]) :- apply_unfold_d_eq(C), unfold_d_eqs_ok(L).
unfold_d_eq(C) :- check_input_clause(C), apply_unfold_d_eq(C).
unfold_d_eq(C) :- msg_not_a_clause_in_current_prog(C).
/*---------------
 unfold_d_facts: Simplify the body of all clauses by unfolding all instances
 of facts, deleting clauses with a failing body, and simplifying equalities
 and inequalities when possible. 
 unfold_d_facts(C): Simplify clauses body of clause C.
----------------*/
unfold_d_facts:-
   check_exists_transf_tree, current_clauses(CNS), unfold_d_facts_ok(CNS).
unfold_d_facts([]) :- !.
unfold_d_facts([C|L]) :-!,
   check_exists_transf_tree, unfold_d_facts(C), unfold_d_facts(L).
unfold_d_facts(C) :-!, unfold_d_facts([C]).
unfold_d_facts_ok([]).
unfold_d_facts_ok([C|L]) :- apply_simpl_body(C), unfold_d_facts_ok(L).
/*-----------
 ts_fold: Fold a clause using a definition (Tamaki and Sato's folding rule).
 The clause used to fold must have been introduced by the definition introduc-
 tion rule introducing one clause only, and in an ancestor of current program.
-------------*/
ts_fold :-    
    check_exists_transf_tree, available_definitions_for_ts_fold(L),
    (L=[] -> msg_definition_clauses_empty
    ; current_clauses(CNs), ask_clause_number(CNs,C_N), 
      check_clause_with_non_empty_body(C_N,Length),
      ts_fold_(L,Length,C_N)).
ts_fold(C,Atoms,Def) :-
   check_exists_transf_tree, check_input_clause_to_fold(C,Length),
   check_input_atoms(Length,Atoms), check_one_clause_definition(Def),
   apply_ts_fold(C,Atoms,Def).
ts_fold_(L,Length,C_N) :-
    ask_positions_in_body(Length,Atom_Numbers), ask_definition_number(L,Ans),
    apply_ts_fold(C_N,Atom_Numbers,Ans).
/*---------
 fold(C_Ns,C_Ns2): Fold clauses in L1 using clauses in L2. No permutations in
 clauses bodies. L2 must be clauses in an ancestor of current program.
----------*/
fold_with_no_check(C_Ns,C_Ns2,NewClause) :- 
    heads_unifiable(C_Ns2,_,A),
    heads_unifiable(C_Ns,_,GenHd),
    apply_fold(C_Ns,C_Ns2,GenHd,A,NewClause).

fold(C_Ns,C_Ns2) :- check_exists_transf_tree,
   check_input_clauses_for_folding(C_Ns,C_Ns2,GenH,A),
   apply_fold(C_Ns,C_Ns2,GenH,A,_).
/*---------
 rearrange_and_fold(L1,L2): Fold clauses in L1 using clauses in L2.
 Same as fold but permutations of atoms in clauses bodies are done if necessary.
----------*/
rearrange_and_fold([C1|CsN],[C2|CsN2]) :-  !,
   check_exists_transf_tree,
   check_input_clauses_for_folding([C1|CsN],[C2|CsN2],GenH,A),
   apply_general_fold([C1|CsN],[C2|CsN2],GenH,A,_).
rearrange_and_fold(C1,C2) :- rearrange_and_fold([C1],[C2]).
/*---------
 replace_goal: Replace a goal in a clause by another goal via a replacement law.
-----------*/
replace_goal :- 
   check_exists_transf_tree, available_laws_for_rep(L),
   (L=[] -> msg_laws_empty ; current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   replace_goal_(L,Length,Clause_Number)).
replace_goal_(L,Length,Clause_Number) :-
   ask_positions_in_body(Length,Atom_Numbers), ask_law_number(L,Ans),
   apply_replace_goal(Clause_Number,Atom_Numbers,Ans).
replace_goal(C,Atoms,Law) :- 
   check_exists_transf_tree, check_input_clause_for_goal_rep(C,Length),
   check_input_atoms(Length,Atoms), is_available_law(Law),
   apply_replace_goal(C,Atoms,Law).
/*-----------
 move_lit: Move an atom within the body of a clause. The initial and final
 positions are selected by the user via a menu.
 move_lit(C,P1,P2): Move atom from position P1 to position P2 in the body of C.
-------------*/
move_lit :-  
   check_exists_transf_tree, current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_initial_position(Length,Pos1), ask_final_position(Length,Pos2),
   apply_move_lit(Clause_Number,Pos1,Pos2).
move_lit(Clause_Number,Pos1,Pos2) :-
   check_exists_transf_tree,
   check_input_clause_for_move_lit(Clause_Number,Length),
   check_input_atoms(Length,[Pos1,Pos2]),
   apply_move_lit(Clause_Number,Pos1,Pos2).
/*-----------
 add_goal: Add a goal in the body of a clause. The clause and the position of
 the new goal are selected by the user.
------------*/
add_goal :-
   check_exists_transf_tree, current_prog(P,CNs), show_program(P),
   ask_clause_number(CNs,Clause_Number),ask_enter_goal_to_add(L), 
   length_body_clause(Clause_Number,Length_Body),
   ask_position_of_new_goal(Length_Body,Pos),
   apply_add_goal(Clause_Number,Pos,L).
add_goal(Clause_Number) :-
   check_exists_transf_tree, check_input_clause(Clause_Number),
   ask_enter_goal_to_add(L), length_body_clause(Clause_Number,Length_Body),
   ask_position_of_new_goal(Length_Body,Pos),
   apply_add_goal(Clause_Number,Pos,L).
/*-------------
 delete_goal: Delete a goal in the body of a clause. The clause and the
 position of the goal to delete are selected by the user.
-------------*/
delete_goal :-
   check_exists_transf_tree, current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number),
   check_clause_with_non_empty_body(Clause_Number,Length),
   ask_positions_in_body(Length,Atom_Numbers),
   apply_delete_goal(Clause_Number,Atom_Numbers).
delete_goal(Clause_Number,Atom_Numbers) :-
   check_exists_transf_tree,
   check_input_clause_for_goal_deletion(Clause_Number,Length),
   check_input_atoms(Length,Atom_Numbers),
   apply_delete_goal(Clause_Number,Atom_Numbers).
/*------------
 restriction(P/A): Eliminates transient definitions, that is, clauses that
 define a predicate that is never use in the body of a clause in the program
 which defines predicate P of arity A.
--------------*/
restriction(P/A) :-
   check_exists_transf_tree, check_exists_predicate(P/A),
   current_clauses(Cls),
   apply_restriction([P/A], Cls, Cls2),
   (seteq(Cls,Cls2) -> msg_unchanged_after_elimination(P/A) ; true).
/*---------
 delete_clauses: Delete clauses whose number are given by the user via a menu.
 delete_clauses(L): Delete clauses in L.
-----------*/
delete_clauses :- check_exists_transf_tree, current_clauses(CNs),
   ask_clause_number(CNs,Clause_Number), apply_delete_clause(Clause_Number),
   (yesno('Another clause to delete', n) -> delete_clauses ; true).
delete_clauses([]).
delete_clauses([C|Cs]) :- delete_clause(C), delete_clauses(Cs).
delete_clause(Clause_Number) :- 
   check_exists_transf_tree, check_input_clause(Clause_Number),
   apply_delete_clause(Clause_Number).
delete_clause(Clause_Number) :- 
   msg_not_a_clause_in_current_prog(Clause_Number).
/*----------- 
 delete_failing_body_clauses: Delete all clauses with a failing body.
 delete_failing_body_clauses(L): Delete clauses in L with a failing body.
 A failing body is a body which contains a failing goal. 
-------------*/
delete_failing_body_clauses :-
   check_exists_transf_tree, current_clauses(CNS),
   delete_failing_body_clauses(CNS).
delete_failing_body_clauses(CNS) :-
   check_exists_transf_tree, delete_failing_body_clauses(CNS,CNS,Applied),
   (Applied==applied -> true ;  msg_unchanged_after_failing_bodies).
delete_failing_body_clauses([],_,_) :- !.
delete_failing_body_clauses([C|L],CNs,Applied) :-!,
   member(C,CNs), !, body_of_clause(C,R),
   (non_failed_body(R,CNs) -> true 
   ; (Applied=applied, apply_del_cl_failing_bd(C))),
     delete_failing_body_clauses(L,CNs,Applied).
delete_failing_body_clauses([C|L],CNS,Applied) :-
   msg_not_a_clause_in_current_prog(C),
   delete_failing_body_clauses(L,CNS,Applied).
delete_failing_body_clauses(C,A,B) :- delete_failing_body_clauses([C],A,B).
/*---------
 add_clause: Add a clause to the current program at a position which is given
 by the user. 
-----------*/
add_clause :-
    check_exists_transf_tree, current_clauses(CNs),ask_enter_clause_to_add(C),
    ask_position_introduction(CNs,Pos), apply_add_clause(C,Pos).
/*---------
 replace_clause(C): Replace clause C by another clause given by the user.
-----------*/
replace_clause(CN) :- check_exists_transf_tree, check_input_clause(CN),
    ask_enter_replacing_clause(C), apply_replace_clause(CN,C).
/*---------
 move_clause(C): Replace clause C by another one. The new clause is given by
 the user.
-----------*/
move_clause(CN) :-
   check_exists_transf_tree, check_input_clause(CN),
   current_clauses(CNs), ask_position_introduction(CNs,Pos),
   apply_move_clause(CN,Pos).
/*-----------
 subsumption: Apply subsumption to current program.
 subsumption(Cs): Apply subsumption to clauses in Cs.
-------------*/
subsumption :-
   check_exists_transf_tree, current_clauses(CNs),
   apply_subsumption_clauses(CNs,CNs2),
   (CNs=CNs2 -> msg_unchanged_after_subsumption ; true).
subsumption(Cs) :-
   check_exists_transf_tree, (check_input_clauses(Cs)-> 
           (apply_subsumption_clauses(Cs,Cs2),
            (Cs==Cs2 -> msg_unchanged_after_subsumption ; true)) ; true).
/*-----------
 split_clause/2: Case split without predicates modes
-------------*/
split_clause(Clause_Number,Atom) :-
   check_exists_transf_tree, check_input_clause(Clause_Number),
   apply_split_clause(Clause_Number,Atom,Clauses, Applied),
   (Applied== applied -> ( add_clauses(NewCNs,Clauses,nil), 
   update_programs((Clause_Number, case_split, NewCNs), _), 
   new_clauses(case_split,Clause_Number,NewCNs)) ; true).
/*-----------
 global_case_split/1: Case split by global analysis
-------------*/
global_case_split([C|Cls]) :- apply_case_split_global([C|Cls],_,_).
/*-----------
 binary_case_split/2: Case split of a clause wrt another clause
-------------*/
case_split(Clause_Number,Clause_Number2):-
   check_input_clause(Clause_Number),
   check_input_clause(Clause_Number2), !,
   apply_case_split(Clause_Number,Clause_Number2,Clauses),
   add_clauses(NewCNs,Clauses,nil), 
   update_programs((Clause_Number,(case_split,Clause_Number2), NewCNs), _), 
   new_clauses(case_split,Clause_Number,Clause_Number2,NewCNs).
/*-------------
instantiate_clause(C,V,T): Instantiate clause C wrt its Vth variable which has
type T. Variables are numbered according to the position of their first
occurrence, when reading the clause from left to right, starting from 1.
T must be a predicate name defined in the current program.
--------------*/
instantiate_clause(C,[V|Vs],[T|Ts]) :- !,
   check_exists_transf_tree,
   check_input_clause(C),
   check_input_var_numbers_and_types(C,[V|Vs],[T|Ts]),
   apply_instantiate_clause(C,[V|Vs],[T|Ts]).
instantiate_clause(C,V,T) :-
   check_exists_transf_tree,
   check_input_clause(C),
   check_input_var_number(C,V),
   check_input_type(T),
   apply_instantiate_clause_one_var(C,V,T).

/*---------------
 start: Initialize a new transformation process with an initial program to be
 selected by the user in the menu made out of the files stored in subdirectory
 Progs/
 start(P): Initialize a new transformation process with P for initial program.
 P must be the name of a file stored in subdirectory Progs/
-----------------*/
start :-
   erase_environment, (input -> init_environment ; msg_failure_no_selection).
start(P) :- file_exists(P), !, apply_start(P).
start(P) :- my_concat_atom(['Progs/',P],F), file_exists(F), !, apply_start(F).
start(P) :- msg_unknown_file(P), fail.
/*---------
 load_session: Restore a transformation tree corresponding to a derivation
 which has been saved by the user in a previous session with the command s_sess.
 The user selects the desired transformation tree via a menu made out of all
 files in subdirectory Sessions/ .
-----------*/
load_session :-
   erase_environment, (input_session -> true ; msg_no_selection).
/*-----------------
 set_current_program(N): Set current program to program N.
-------------------*/
set_current_program(NP) :-
   check_exists_prog_in_transf_tree(NP), apply_set_current_program(NP).
/*-----------------
 add_law_in_theory(T): Add a law to a file T to be selected by the user in the
 menu made out of the files stored in subsirectory Theories/. The law to add
 is given by the user. It is a term of the form A1,...,An -> B1,...,Bn.
 add_law_in_theory: add a law to the current theory.
--------------------*/
add_law_in_theory(T) :-
   my_concat_atom(['Theories/',T],F), file_exists(F), !,
   apply_add_law_in_theory(F).
add_law_in_theory(T) :- msg_unknown_file(T), fail.
add_law_in_theory :- check_exists_transf_tree, msg_show_existing_laws,
    ask_enter_law(Law), check_law(Law), apply_add_law(Law).
/*-----------------
 delete_law_from_theory(T): Delete a law from a file T to be selected by the
 user in the menu made out of the files stored in subsirectory Theories/.
delete_law_from_theory: Delete a law from current theory.
-------------------*/
delete_law_from_theory(T) :-
   my_concat_atom(['Theories/',T],F), file_exists(F), !,
   apply_delete_law_from_theory(T,F).
delete_law_from_theory(T) :- msg_unknown_file(T), fail.
delete_law_from_theory :- check_exists_transf_tree, available_laws_for_rep(L),
    (L=[] -> msg_laws_empty ; ask_law_number(L,Ans), apply_delete_law(Ans)).
/*-------------
 select_law_in_theory: Add to the current theory a law from a theory to be
 selected by the user in the menu made out of the files stored in directory
 Theories/
--------------*/
select_law_in_theory :-
   check_exists_transf_tree, my_directory_files('Theories',LL),
   non_empty_directory(LL,L), !,
   (select_law_in_theory(1,L,L) -> true ; msg_no_selection).
select_law_in_theory :- msg_no_available_theories.
/*-------------
 select_law_in_theory: Add to the current theory a law from a theory to be
 selected by the user in the menu made out of the files stored in directory
 Theories/
--------------*/
select_theory :-
   check_exists_transf_tree, my_directory_files('Theories',LL),
   non_empty_directory(LL,L), !,
   (select_theory(1,L,L) -> true ; msg_no_selection).
select_theory :- msg_no_available_theories.

/*--------
 show_program(N): Show program N.
 show_short_program(N): Show program N omitting clauses occurring in the
 initial one.
-----------*/
show_program(N) :-
   check_exists_transf_tree, label_node(N,Clause_Numbers), !,
   apply_show_program(Clause_Numbers).
show_program(N) :- msg_unknown_program(N).
show_short_program(1) :- show_program(1).
show_short_program(N) :-
   check_exists_transf_tree, label_node(N,Clause_Numbers), !, 
   apply_show_short_program(Clause_Numbers).
show_short_program(N) :- msg_unknown_program(N).
/*--------
 show_programs(Ps): Show programs in list Ps.
 show_short_programs(Ps): Show programs in Ps omitting clauses occurring
 in the initial one.
-----------*/
show_programs(L) :- check_exists_transf_tree, show_programs_ok(L).
show_programs_ok([]).
show_programs_ok([N|Ns]) :-
   write('program '), write(N), write(':'), nl,
   (label_node(N,Clause_Numbers) -> apply_show_program(Clause_Numbers)
   ; write(' does not exist.')), nl, show_programs_ok(Ns).
show_short_programs(L) :- check_exists_transf_tree, show_short_programs_ok(L).
show_short_programs_ok([]).
show_short_programs_ok([N|Ns]) :- 
   write('program '), write(N), write(':'), nl,
   (label_node(N,Clause_Numbers) ->
     (N = 1 ->  apply_show_program(Clause_Numbers)
   ; apply_show_short_program(Clause_Numbers))
   ; write(' does not exist.')), nl, show_short_programs_ok(Ns).
/*-----------------
 show_initial_program: Shows initial program.
-------------------*/
show_initial_program :- check_exists_transf_tree, apply_show_initial_program, !.
/*-----------------
 show_current_program: Show current program.
 show_current_short_program: Show current program, omitting clauses in the
 initial one.
-------------------*/
show_current_program :-
   check_exists_transf_tree, current_clauses(NCs),
   apply_show_current_program(NCs).
show_current_short_program :- 
   check_exists_transf_tree, current_clauses(Clause_Numbers),
   apply_show_short_program(Clause_Numbers).
/*---------------
 show_history: Show history of the current program.
 show_history(N): Shows history of program N
-----------------*/
show_history :- check_exists_transf_tree, current_program(N), show_history(N).
show_history(N) :- check_exists_prog_in_transf_tree(N), apply_show_history(N).
/*----------
 show_clauses: Show clauses
------------*/
show_clauses :-	check_exists_transf_tree, apply_show_clauses.
show_clauses(L) :- check_exists_transf_tree, show_clauses_ok(L).
show_clauses_ok([]) :- !.
show_clauses_ok([N|L]) :- !, show_clauses_ok(N), nl, show_clauses_ok(L).
show_clauses_ok(N) :- clause_number_to_clause(N,C), !, show_numb_clause((N,C)).
show_clauses_ok(N) :- msg_unknown_clause(N).
/*---------
 show_definitions: Show existing definitions
----------*/
show_definitions:-
   check_exists_transf_tree, check_exists_def, !, apply_show_definition.
show_definitions:- msg_definition_clauses_empty.
/*---------
show_laws: Show current theory
----------*/
 show_laws :- check_exists_law, apply_show_law.
 show_laws :- msg_laws_empty.
/*----------
 dir: Show the content of subdirectory Progs/
------------*/
dir_p :-
   my_directory_files('Progs',LL), non_empty_directory(LL,L), !,
   show_available_programs(1,L,L,_),!.
dir_p :-  msg_no_available_progs.
/*----------
 dir_t: Show the content of subdirectory Theories/
------------*/
dir_t :-
   my_directory_files('Theories',LL), non_empty_directory(LL,L), !,
   show_available_theories(1,L,L,_),!.
dir_t :-  msg_no_available_theories.
/*----------
 dir_s: Show the content of subdirectory Sessions/
------------*/
dir_s :-
   my_directory_files('Sessions',LL), non_empty_directory(LL,L), !,
   show_available_sessions(1,L,_),!.
dir_s :-  msg_no_available_sessions.
/*----------
 dir_f: Show the content of subdirectory Final_Progs/
------------*/
dir_f :-
   my_directory_files('Final_Progs',LL), non_empty_directory(LL,L), !,
   show_available_final_programs(1,L,L,_),!.
dir_f :-  msg_no_available_final_progs.

/*-------------
 save_prog(P): Save program P. By default, this file is stored in Final_Progs/,
 unless the user specifies another name. P is the current program if omitted.
 save_program: save current program.
---------*/
save_prog(P) :-
   check_exists_transf_tree, current_program(N),
   check_exists_prog_in_transf_tree(P), set(P),
   save_current_program, set(N).
save_program :- save_current_program.
save_current_program :-
   check_exists_transf_tree, current_clauses(NCs),
   msg_enter_name_program(File2), name(Name_File2,File2),
   msg_save_current_program(Name_File2), tell(Name_File2),
   apply_show_only_clauses(NCs), told.
/*---------------------
 save_history(PN): Produce a file in which all programs of the transformation
 sequence from P1 to PN together with information about the rules used are
 stored. By default, this file is stored in  Histories/history_PN, unless the
 user specifies another name.
 save_current_history: save current history.
-----------------------*/
save_history(N) :- check_exists_prog_in_transf_tree(N), apply_save_history(N).
save_history :-
   check_exists_transf_tree, current_program(N), apply_save_history(N).
/*-------------
 save_current_theory: Save the current theory. By default, this file is stored
 in Theories/, unless the user specifies another name.
---------*/
save_current_theory :- check_exists_law, apply_save_current_theory.
save_current_theory :- msg_laws_empty.
/*--------
 save_session : Save current session state
----------*/
save_session :-
   check_exists_transf_tree, file_input(File), name(File,L),
   append("Progs/",Prog,L), append("Sessions/Session_",Prog,History_F),
   msg_ask_save_session(History_F,File2), name(Name_File2,File2),
   msg_save_current_session(Name_File2),
   tell(Name_File2), listing_transf_tree, told.
/*---------
 insert_comment_history: Allow the user to enter some comments which will be
 associated with the current program in the history of the derivation.
----------*/
insert_comment_history :-
   check_exists_transf_tree, current_program(P),
   apply_insert_comment_history(P,[]).
/*---------
 helps: Show a general help menu
----------*/
helps :- msg_helps.
/*---------
 rules: Show the list of top-level commands related to transformation rules. 
----------*/
rules :- msg_help_rules.
/*---------
 commands: Show the list of top-level commands which provide facilities for
 consulting or saving informations related to transformation states.
----------*/
commands :- msg_help_commands.
/*---------
 strategies: Show the list of top-level commands related to strategies
----------*/
strategies :- msg_help_strategies.









%%%%%%%% NEW TOP LEVEL PREDICATES %%%%%%%%%%%

/*-----------
 r_subsumption: Apply subsumption to current program with possible reordering
  in the bodies.
 r_subsumption(Cs): Apply r_subsumption to clauses in Cs.
-------------*/
r_subsumption :-
   check_exists_transf_tree, current_clauses(CNs),
   apply_r_subsumption_clauses(CNs,CNs2),
   (CNs=CNs2 -> msg_unchanged_after_subsumption ; true).
r_subsumption(Cs) :-
   check_exists_transf_tree, (check_input_clauses(Cs)-> 
           (apply_r_subsumption_clauses(Cs,Cs2),
            (Cs==Cs2 -> msg_unchanged_after_subsumption ; true)) ; true).





go(X,X,_) :-!.
go(X,Y,Pos) :-
	ud(X,Pos), Z is X+1, go(Z,Y,Pos).
	
find_f(H) :-
	current_prog(Pi,Cls),
	bagof(C, has_same_head(Cls,H,C),L),
	((clause3(_,_,[gen_def|L2]),
	same_length(L,L2),
	(strat(folding_with_clause_permutation) ->
             (permutation(L,PermClauses),
             try_to_fold_with_defs(Pi,PermClauses,L2)) ;
	try_to_fold_with_defs(Pi,L,L2))) ; cls(L)).
has_same_head(Cls,H,C) :-
	member(C,Cls),
	clause3(C,(H2,_),_), variant(H,H2).