:- use_module(library(terms)).
:- use_module(library(ordsets)).
:- compile('basic_operations.pl'),
   compile('solver_clpr.pl'),
   compile('specialize_ctlstar_strategy.pl'),
   compile('ru_cfacts_strategy.pl').


% Input: a locally stratified CLP program P that specifies a
% CTL property Phi of a transtition system.
% P is represented as a set of facts program_clauses(H,B)

verify :-
	check_exists_transf_tree,	 % from struct.pl
       	initial_program_and_property(Prog,Phi),
       	verify(Prog,Phi,A),
       	nl, write(A), nl.

% initial_program_and_property(Prog,Phi) :-
% (i) get the clauses of the current program as
% a list Cls of clauses of the form (H,bd(Cns,Lits));
% (ii) transform P into a list of definitions Pin.

initial_program_and_property(Prog,Phi) :-
        findall((H,bd(Cns,Lits)),
        	(current_prog(_,Ns),
        	 member(N,Ns),
        	 clause3(N,(H,B),_),
		 split_bd(B,bd(Cns,Lits))),Cls),
	hd_preds(Cls,Preds),
	cls_to_defs(Preds,Cls,Prog),
	read_property(Phi,Prog).

hd_preds([],[]).
hd_preds([C|Cls],Preds) :-
	hd_pred(C,P),
	hd_preds(Cls,Preds1),
	(member(P,Preds1) -> Preds=Preds1 ; Preds=[P|Preds1]).

cls_to_defs([],_,[]).
cls_to_defs([(P/N)|Preds],Cls,[(P/N,Def)|Defs]) :-
	filter_hd_pred((P/N),Cls,Def),
	cls_to_defs(Preds,Cls,Defs).

filter_hd_pred(_,[],[]).
filter_hd_pred((P/N),[C|Cls],Def) :-
	hd_pred(C,P/N) ->
		(Def=[C|Def1], filter_hd_pred((P/N),Cls,Def1));
		filter_hd_pred((P/N),Cls,Def).


read_property(Phi,Prog) :-
        write('property to be verified:'), nl,
        write('enter a nullary predicate defined by a single clause in the current program'), nl,
        read(Phi),
        (check_property(Phi,Prog) ->
		 nl;
		(nl, write('error: not a nullary predicate defined by a single clause in
		the current program'), nl)).

check_property(Phi,Prog) :-
	atom(Phi),
	def((Phi/0),Prog,[_]).

def(_,[],[]).
def(P,[(P1,Def1)|Defs],Def2) :-
	P=P1 -> Def1=Def2 ; def(P,Defs,Def2).

% verify(Prog,Phi,Answer) :-
% First i) apply the specialization strategy, then
% ii) apply the remove-unfold strategy, and finally
% iii) inspect the final program:
% If Phi. belongs to the final program then Answer is true else Answer
% is false.

verify(Prog,Phi,Answer) :-
	%specialization
       	statistics(runtime,[_,T1]),
	initial_def_and_prog(Phi,Prog,(Phi,Bd),Pin),
	write('---specialization strategy--- '), nl, nl,
        specialize_ctlstar_strategy([(Phi,Bd)],Pin,[(Phi,Bd)],Pspec,Defs_out),
         (derivation_trace_on ->
          (write(' -definitions- '), nl, show_cls(Defs_out), nl,
	   write(' -specialized program- '), nl, show_prog(Pspec), nl);
	   (length(Defs_out,N), nl, write('#definitions: '), write(N), nl)),
%	statistics(runtime,[_,T2]),
	nl,

	%post removal and unfolding
	write('---remove-unfold strategy--- '), nl, nl,
        ru_cfacts_strategy(Phi,Pspec,Pout),
         (derivation_trace_on ->
          (write(' -final program- '), nl, show_prog(Pout), nl);
          true),
	statistics(runtime,[_,T3]),

        inspect(Phi,Pout,Answer),
%      	Tsp is T2-T1, write('Specialization time '), write(Tsp), write(ms), nl,
 %     	Tru is T3-T2, write('Remove-Unfold time '), write(Tru), write(ms), nl,
      	T is T3-T1, write('Total time '), write(T), write(ms), nl.

% initial_def_and_prog(Phi,Prog,(Phi,Bd),Pin) :-  the set of definitions, represented
% as ++++++, is initialized to {Phi :- Bd}

initial_def_and_prog(Phi,Prog,Cl,Pin) :-
	del_def((Phi/0),Prog,[Cl],Pin),
         (derivation_trace_on ->
          (nl, write(' -define- '), nl, show_cls([Cl]),  nl);
          true).

del_def(P,[(P1,D1)|Defs],D,DDefs) :-
	P=P1 -> (D=D1, DDefs=Defs) ;
	(DDefs=[(P1,D1)|DDefs1], del_def(P,Defs,D,DDefs1)).

inspect(Phi,Prog,Answer) :-
	(def(Phi/0,Prog,Def),member((Phi,Bd),Def)) ->
		(Bd=bd([],[]) ->  Answer=true ; Answer='verification failed') ;
		Answer=false.

