% MODULE: ufmcfr
% provides: verify/0
% requires: library(ordsets)
%           basic_operations
%           basic_op,determ_prog_util,environment,input_output,man,msg_user,struct,top_level
%           ctl_operators,
%           specialize_ctlf,
%           remove_unfold,
%           solver_clpr,

% Input: a locally stratified CLP program P that specifies a
% CTL property Phi of a transition system.
% P is read from a file, consulted via the command 'st.'

%:- module(ufmc, [verify/0]).

:- use_module(library(ordsets)).
:- compile('basic_operations.pl'),
   compile('ctl_operators.pl'),
   compile('solver_clpr.pl'),
   compile('specialize_ctlf.pl'),
   compile('remove_unfold.pl').


% TYPES
%
% Base types: term, predicate, nat, real.
% Type constructors: list/1.
% Type definitions:
% atom == predicate(term1,...,termN).
% literal == atom; \+ atom.
% clause == (atom, bd(list(constraint),list(literal))).
% clause_tree = list((atom,list(list(clause)))).
% arity == nat.
% def == (predicate/arity,list(clause)).
% answer == true; false; unknown.
% constraint == real=<real; real>=real; real<real; real>real; real=:=real; real=\=real.


% MODE: verify
% TYPE: verify
% SEMANTICS:    reads a transition system and a temporal property
%               writes 'true' if the property is proved
%		       'false' if the property is disproved
%                      'unknown' if the property is neither proved nor disproved

verify :-
	check_exists_transf_tree,
       	initial_program_and_property(Prog,Phi),
       	statistics(runtime,[_,T1]),
       	verify(Prog,Phi,A),
       	statistics(runtime,[_,T2]),
	nl, write(A), nl,
      	nl, T is T2-T1, write('Elapsed time '), write(T), write(ms), nl.


% MODE: initial_program_and_property(-Prog,-Phi)
% TYPE: initial_program_and_property(list(def),predicate)
% SEMANTICS: Prog is the list of clauses of the current program
% Phi is a nullary predicate symbol defined by a single clause in the program

initial_program_and_property(Prog,Phi) :-
        findall((H1,bd(Cns1,Lits1)),
        	(current_prog(_,Ns),
        	 member(N,Ns),
        	 clause3(N,(H1,B1),_),
		 split_bd(B1,bd(Cns1,Lits1))),Cls1),
        findall((H2,bd(Cns2,Lits2)),
                (ctl_clause(H2,B2),
        	(split_bd(B2,bd(Cns2,Lits2)))),Cls2),
	append(Cls1,Cls2,Cls),
	preds_in_cls(Cls,[],Preds),
	cls_to_defs(Preds,Cls,Prog),
	read_property(Phi,Prog).


% MODE: preds_in_cls(+Cls,+InPreds, -OutPreds)
% TYPE: preds_in_cls(list(clause),list(predicate/arity),list(predicate/arity))
% SEMANTICS: OutPreds is the list of predicates occurring in Cls and in InPreds

preds_in_cls([],Preds,Preds).
preds_in_cls([Cl|Cls],InPreds,OutPreds) :-
        preds_in_cl(Cl,InPreds,Ps),
        preds_in_cls(Cls,Ps,OutPreds).


% MODE: preds_in_cl(+Cl,+InPreds, -OutPreds)
% TYPE: preds_in_cl(clause,list(predicate/arity),list(predicate/arity))
% SEMANTICS: OutPreds is the list of predicates occurring in Cl and in InPreds

preds_in_cl((H,bd(_C,G)),InPreds,Preds) :-
	preds_in_goal([H|G],InPreds,Preds).


% MODE: preds_in_goal(+Goal,+InPreds, -OutPreds)
% TYPE: preds_in_goal(list(literal),list(predicate/arity),list(predicate/arity))
% SEMANTICS: OutPreds is the list of predicates occurring in Goal and in InPreds

preds_in_goal([],InPreds,InPreds).
preds_in_goal([A|As],InPreds,Preds) :-
	literal_pred(A,P/N),
	(member_chk(P/N,InPreds)  ->
		preds_in_goal(As,InPreds,Preds) ;
		preds_in_goal(As,[P/N|InPreds],Preds)).


% MODE: cls_to_defs(+Preds,+Cls, -Defs)
% TYPE: cls_to_defs(list(predicate/arity),list(clause),list(def))
% SEMANTICS: the list Cls of clauses is transformed into the list Defs of
% predicate definitions consisting of a definition for each predicate in Preds

cls_to_defs([],_,[]).
cls_to_defs([(P/N)|Preds],Cls,[(P/N,Def)|Defs]) :-
	filter_hd_pred((P/N),Cls,Def),
	cls_to_defs(Preds,Cls,Defs).


% MODE: filter_hd_pred(+Pred,+Cls, -Def)
% TYPE: filter_hd_pred(predicate/arity,list(clause),def)
% SEMANTICS: Def is (Pred,D) where D is the list of clauses
% occurring in Cls whose head predicate is Pred

filter_hd_pred(_,[],[]).
filter_hd_pred((P/N),[C|Cls],Def) :-
	hd_pred(C,P/N) ->
		(Def=[C|Def1], filter_hd_pred((P/N),Cls,Def1));
		filter_hd_pred((P/N),Cls,Def).


% MODE: read_property(-Phi, +Prog)
% TYPE: read_property(predicate,list(def))
% SEMANTICS: Phi is read from the user.
% Phi is a nullary predicate defined by a single clause in the current program Prog.

read_property(Phi,Prog) :-
        write('property to be verified:'), nl,
        write('enter a nullary predicate defined by a single clause in the current program'), nl,
        read(Phi),
        (check_property(Phi,Prog) ->
		 nl;
		(nl, write('error: not a nullary predicate defined by a single clause in
		the current program'), nl)).


% MODE: check_property(+Phi,+Prog)
% TYPE: check_property(predicate,list(def))
% SEMANTICS: Phi is a nullary predicate defined by a single clause in
% the current program Prog

check_property(Phi,Prog) :-
	atom(Phi),
	def((Phi/0),Prog,D),
	D=[_].


% MODE: def(+Pred,+Defs, -Def)
% TYPE: def(predicate/arity,list(def),def)
% SEMANTICS: in Defs there exists a definition (Pred,Def) for Pred

def(P,[(P1,Def1)|_Defs],Def2) :-
	P=P1, !,
	Def1=Def2.
def(P,[_|Defs],Def2) :-
	def(P,Defs,Def2).


% MODE: verify(+Prog,+Phi, -Answer)
% TYPE: verify(list(def),predicate, answer)
% SEMANTICS: 
% i) apply the specialization strategy to Prog starting from the single clause for Phi, then
% ii) apply the remove-unfold strategy, and finally
% iii) inspect the final program.
% If Phi. belongs to the final program then Answer is 'true';
% If no clause for Phi occurs in the final program then Answer is 'false'.
% Otherwise Answer is 'unknown'.

verify(Prog,Phi,Answer) :-
	%specialization
	initial_def_and_prog(Phi,Prog,C,Pin),
	bd_lit(C,H),  %function used to index the defs
	write('---specialization strategy--- '), nl, nl,
        specialize_ctl([C],Pin,[(H,[[C]])],Pspec,Defs_out),
         (derivation_trace_on ->
          (nl, nl, write(' -definitions- '), nl, show_deftree(Defs_out), nl,
	   write(' -specialized program- '), nl, show_prog(Pspec), nl);
	   (deftree_length(Defs_out,N), nl, nl, write('#definitions: '), write(N), nl)),
	nl,
	%post removal and unfolding
	write('---post remove-unfold strategy--- '), nl, nl,
        remove_unfold(Pspec,Pout),
         (derivation_trace_on ->
          (nl, nl, write(' -final program- '), nl, show_prog(Pout), nl);
          true),
        inspect(Phi,Pout,Answer).


% MODE: initial_def_and_prog(+Phi,+Prog, -Cl,-Pin)
% TYPE: initial_def_and_prog(predicate,list(def),clause,list(def))
% SEMANTICS: Pin is obtained by deleting the definition (Phi/0,Cl) from Prog

initial_def_and_prog(Phi,Prog,Cl,Pin) :-
	del_def((Phi/0),Prog,[Cl],Pin),
         (derivation_trace_on ->
          (nl, write(' -define- '), nl, show_cls([Cl]),  nl);
          true).


% MODE: del_def(+P,+Defs1, -Def,-Defs2)
% TYPE: del_def(predicate/arity,list(def),def,list(def))
% SEMANTICS:
% (P,Def) is the definition of P in Defs1 and Defs2 is obtained by deleting
% (P,Def) from Defs1

del_def(_P,[],[],[]).
del_def(P,[(P1,D1)|Defs],D,DDefs) :-
	P=P1 -> (D=D1, DDefs=Defs) ;
	(DDefs=[(P1,D1)|DDefs1], del_def(P,Defs,D,DDefs1)).


% MODE: bd_lit(+Cl, -L)
% TYPE: bd_lit(clause,literal)
% SEMANTICS: L is the unique literal in the body of Cl

bd_lit((_,Bd),H) :- Bd=bd(_,[H]).


% MODE: specialize_ctl(+Cls1,+Prog1,+Defs1, -Prog2,-Defs2)
% TYPE: specialize_ctl(list(clause),list(def),clause_tree,list(def),clause_tree)
% SEMANTICS: Program Prog2 and definitions Defs2 are obtained by applying the
% specialization strategy starting from the list Cls1 of clauses, the program Prog1,
% and the list Defs1 of definitions.
% IMPORTED FROM: specialize_ctl


% MODE: remove_unfold(+Pin, -Pout)
% TYPE: remove_unfold(list(def), list(def))
% SEMANTICS: Pin is transformed into Pout by
%    ((removing useless clauses;
%      removing subsumed clauses) ;
%    (positively and negatively unfolding wrt atoms defined by constrained facts)*)*
% IMPORTED FROM: remove_unfold


% MODE: show_prog(+Defs)
% TYPE: show_prog(list(def))
% SEMANTICS: displays the list Defs of definitions

show_prog([]).
show_prog([(_,Def)|Defs]) :-
	show_cls(Def),
	show_prog(Defs).


% MODE: show_cls(+Cls)
% TYPE: show_cls(list(clause))
% SEMANTICS: displays the list Cls of clauses

show_cls([]).
show_cls([(H,bd(Cns,Lits))|Cls]) :-
	append(Cns,Lits,Bd),
	show_clause(0,(H,Bd)),
	nl,
	show_cls(Cls).


% MODE: show_deftree(+DefTree)
% TYPE: show_deftree(clause_tree)
% SEMANTICS: displays the tree DefTree of clauses (grouped by body atoms)

show_deftree([]).
show_deftree([(_,Def)|Defs]) :-
	show_clstree(Def),
	show_deftree(Defs).


% MODE: show_clstree(+ClsTree)
% TYPE: show_clstree(list(list(clause)))
% SEMANTICS: displays the tree ClsTree of clauses (represented as
% a list of lists of lists of clauses)

show_clstree([]).
show_clstree([[Cl|Cls]|Clss]) :-
	write('definition:'), nl,
	show_cls([Cl]),
	(Cls=[] -> (write('no ancestor'), nl); (write('ancestors:'), nl)),
	show_cls(Cls), nl,
	show_clstree(Clss).


% MODE: deftree_length(+DefTree, -N)
% TYPE: deftree_length(clause_tree,nat)
% SEMANTICS: N is the number of leaves of the tree DefTree of clauses
% (grouped by body atoms)

deftree_length([],0).
deftree_length([(_,Def)|Defs],N) :-
	length(Def,N1),
	deftree_length(Defs,N2),
	N is N1+N2.


% MODE: inspect(+Phi,+Prog, -Answer)
% TYPE: inspect(predicate,list(def),answer)
% SEMANTICS:
% If Phi. belongs to Prog then Answer is 'true';
% If no clause for Phi occurs in Prog then Answer is 'false'.
% Otherwise Answer is 'unknown'.

inspect(Phi,Prog,Answer) :-
	(def(Phi/0,Prog,Def),member((Phi,Bd),Def)) ->
		(Bd=bd([],[]) ->  Answer=true ; Answer=unknown) ;
		Answer=false.
