:- use_module(library(terms)).
:- use_module(library(ordsets)).

:- op(960,xfy, [<->]).
:- op(950,xfy, [or]).
:- op(900,xfy, [and]).
:- dynamic tmp_program_clause/2.
:- dynamic tmp_type_definition/3.
:- dynamic tmp_definition_clause/2.
:- dynamic is_type/1.
:- dynamic tmp_clause/2.

% Given: (1) a locally stratified program P which is the current program
% in the transformation sequence represented as a set of facts
% tmp_program_clause(H,B), and
% (2) a typed closed first order formula Phi,
% (A) prove(Phi,A) holds iff the truth value of the typed first order
% formula Phi in the least Herbrand model M(P) of program P is A, and
% (B) synth3(Newp,Phi,SynthesizedP) holds only if Phi is a typed open first
% order formula and SynthesizedP is a program defining a new
% atom Newp, s.t.
% M(P) |= Phi  iff  M(SynthesizedP) |= Newp
% Types are defined as predicates in P.
% Types are declared by the predicate type/3 in P.


prove :-
      read_closed_formula(Phi),
      initialize_tmp_program,
      statistics(runtime,[_,T1]),
      prove(Phi,A),
      nl, write(A), nl,
      statistics(runtime,[_,T2]), nl,
      T is T2-T1, write('Elapsed time '), write(T), write(ms), nl,
      discard_tmp_program, !,
      A==true.

prove(Phi) :-
      check_closed_fo_formula(Phi),
      initialize_tmp_program,
      statistics(runtime,[_,T1]),
      prove(Phi,A),
      nl, write(A), nl,
      statistics(runtime,[_,T2]), nl,
      T is T2-T1, write('Elapsed time '), write(T), write(ms), nl,
      discard_tmp_program, !,
      A==true.

synth :-
      read_new_predicate(Newp),
      read_formula(Phi),
      initialize_tmp_program,
      statistics(runtime,[_,T1]),
      synth3(Newp,Phi,SynthesizedP),
      statistics(runtime,[_,T2]), nl,
      T is T2-T1,
      discard_tmp_program,
      update_transf_tree(Ns,SynthesizedP),
      nl, write('Synthesized Program: '), nl,
      show_clauses(Ns), nl,
      write('Elapsed time '), write(T), write(ms), nl.

synth(Newp,Phi) :-
      check_new_predicate(Newp),
      check_fo_formula(Phi),
      initialize_tmp_program,
      statistics(runtime,[_,T1]),
      synth3(Newp,Phi,SynthesizedP),
      statistics(runtime,[_,T2]), nl,
      T is T2-T1,
      discard_tmp_program,
      update_transf_tree(Ns,SynthesizedP),
      nl, write('Synthesized Program: '), nl,
      show_clauses(Ns), nl,
      write('Elapsed time '), write(T), write(ms), nl.

% read_new_predicate(?Newp) :-  read Newp from the user and
% check that:  Newp is an atom with a new predicate name.

read_new_predicate(Newp) :-
        write('enter a new predicate to be synthesized'), nl,
        read(Newp),
        check_new_predicate(Newp).

%check_new_predicate(+Newp)

check_new_predicate(Newp) :-
        (\+ atomic_formula(Newp) ->
                (error_msg(not_atom,[Newp]), fail) ;
                ((functor(Newp,P,A),
                  cl_number_in_current_prog(P/A,_)) ->
                        ask_user(redefine,[P/A]) ; true)).


% read_closed_formula(Phi) :-  read Phi from the user and
% check that: (i) Phi is a first order formula and
% (ii) the set of free variables of Phi is empty

read_closed_formula(Phi) :-
        write('enter a closed first order formula'), nl,
        read(Phi),
        check_closed_fo_formula(Phi).

check_closed_fo_formula(Phi) :-
        fo_formula(Phi),
        fv(Phi,L),
        L=[], !.
check_closed_fo_formula(_) :-
        write('not a closed first order formula'), nl, fail.

% fo_formula(F) :-
% F is a first order formula constructed from
% true, false, atomic formulas,
% and, or, ->, <->, e (exists), a (forall).

fo_formula(true).
fo_formula(false).
fo_formula(F) :- atomic_formula(F), !.
fo_formula(\+F) :- fo_formula(F).
fo_formula((F1 and F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 or F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 -> F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula((F1 <-> F2)) :- fo_formula(F1), fo_formula(F2).
fo_formula(e(X,F)) :- nonvar(X), list_of_vars(X), fo_formula(F).
fo_formula(a(X,F)) :- nonvar(X), list_of_vars(X), fo_formula(F).

atomic_formula(F) :-
        F\=(\+ _),
        F\=(_ and _),
        F\=(_ or _),
        F\=(_ -> _),
        F\=(_ <-> _),
        F\=e(_,_),
        F\=a(_,_).

% list_of_vars(Vs) :-  Vs is a list of variable names

list_of_vars([]) .
list_of_vars([V|Vs]) :- var(V), list_of_vars(Vs).


%fv(+F,?Vs): F -> Vs
%fv(F,Vs) :- Vs is the set of free variables in F

fv(true,[]).
fv(false,[]).
fv(A,V) :-
        atomic_formula(A), !,
        term_variables(A,V).
fv((\+ F),V) :- fv(F,V).
fv((F1 and F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 or F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 -> F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv((F1 <-> F2),V) :-
        fv(F1,V1),
        fv(F2,V2),
        term_variables((V1,V2),V).
fv(e(X,F),V) :-
        fv(F,V1),
        delete_all(X,V1,V).
fv(a(X,F),V) :-
        fv(F,V1),
        delete_all(X,V1,V).


% delete_all(+Xs,+Ys,?Zs): Xs,Ys -> Zs
% Zs is obtained by deleting from Ys all
% occurrences of the terms in Xs.

delete_all([],L,L).
delete_all([X|Xs],Ys,Zs) :-
        delete_one(X,Ys,Zs1),
        delete_all(Xs,Zs1,Zs).

delete_one(_,[],[]).
delete_one(X,[Y|Ys],Zs) :- X==Y, !, delete_one(X,Ys,Zs).
delete_one(X,[Y|Ys],Zs) :- Zs=[Y|Zs1], delete_one(X,Ys,Zs1).


% drop_quantifier(e(X,F), NewF) :- NewF is the formula obtained from
% F by renaming the variables in the list X

drop_quantifier(e(X,F), NewF) :-
        fv(e(X,F),V),
        assert(tmp_clause(V,F)),
        tmp_clause(V,NewF),
        retract(tmp_clause(_,_ )).


% initialize_tmp_program :-
% (i) check that a transformation session has been
% started by using the st command;
% (ii) discard all temporary program clauses and type definitions;
% (iii) get the clauses of the current program (represented as
% tmp_program_clause/2 facts) together with the type definitions of the
% predicates which occur in it (represented as tmp_type_definition/3
% facts).

initialize_tmp_program :-
      check_exists_transf_tree,  % from struct.pl
      retractall(tmp_program_clause(_,_)),
      retractall(tmp_type_definition(_,_,_)),
      assert_current_program,
      assert_types.

% assert_current_program :- assert as tmp_program_clause/2 facts all
% clauses of the current program.

assert_current_program :-
        findall((H,B),(
        current_prog(_,Ns),
        member(N,Ns),
        clause3(N,(H,B),_),
        assert(tmp_program_clause(H,B))),_).

assert_types :-
        findall((P,A,L),
        (type_definition(P,A,L),assert(tmp_type_definition(P,A,L))),_),
        findall(T,
        (tmp_type_definition(_,_,L), member(T,L)),
        List_of_Types),
        list_to_ord_set(List_of_Types,Set_of_Types),
        findall(T1,(member(T1,Set_of_Types),assert(is_type(T1))),_).

% read_formula(Phi) :-  read Phi from the user and
% check that Phi is a first order formula (not necessarily closed)

read_formula(Phi) :-
        write('enter a first order formula'), nl,
        read(Phi).

check_fo_formula(Phi) :-
        (fo_formula(Phi) -> true;
        (write('not a first order formula'), nl, fail)).
check_fo_formula(Phi) :-
        (fo_formula(Phi) -> true;
        (write('not a first order formula'), nl, fail)).


% prove(Phi,Answer) :-
% Step 1. apply Lloyd-Topor Transformation to Phi and derive a
% new nullary predicate Pred and a set of definitions Defs;
% Step 2. apply the ufs strategy.
% inspect the final program:
% if Pred. belongs to the final program then Answer is true else Answer
% is false.

prove(Phi,Answer) :-
        new_statement(Pred, Phi),
        lloyd_topor_transf(Pred,Phi,Defs),
        (derivation_trace_on ->
          (nl, write('LT_transformation '), nl, nl, show_all(Defs), nl,
           write('UF proof strategy'), nl);
          true),
        ufs(Defs),
        delete_useless_and_subsumed_by_facts([Pred]),
        inspect(Pred,Answer).


show_all([]).
show_all([D|Ds]) :- show_clause(0,D), nl, show_all(Ds).


% synth3(+Newp,+Phi,Program) :-
% Step 1. apply Lloyd-Topor Transformation to Newp :- Phi and derive a
% set of definitions Defs;
% Step 2. apply the ufs strategy and derive the final program Program.

synth3(Newp,Phi,Program) :-
        lloyd_topor_transf(Newp,Phi,Defs),
        (derivation_trace_on ->
          (nl, write('LT_transformation '), nl, nl, show_all(Defs), nl,
           write('UF synthesis strategy'), nl);
          true),
        ufs(Defs),
        delete_useless_and_subsumed_by_facts([Newp]),
        synthesized_prog(Newp,Program).


% lloyd_topor_transf(+Newp,+Phi,?Defs) :-
% apply Lloyd-Topor Transformation to the statement Newp :- Phi
% and derive a list Defs of single-clause definitions.
% In the list Defs = [D1,...,Dn] a predicate occurring in the
% body of Dj is defined in Di (i<j) or in the initial program P.
% We assume that Phi is typable by using the type definitions
% declared in the program by the type/3 predicate and stored as
% tmp_type_definition/3 facts.
% Definitions are asserted as facts of
% the tmp_definition_clause/2 % predicate.

lloyd_topor_transf(Newp,Phi,Defs) :-
        check_statement(Newp,Phi),
        rewrite_formula(Phi,Phi1),
        statement_into_clauses((Newp,Phi1),Defs).

% check_statement(+Newp,+F):-  Newp :- F is a statement of the form
% where Newp and F have the same set of free variables.

check_statement(Newp,F) :-
        ((fv(Newp,Vars1), fv(F,Vars2), sort(Vars1,V1), sort(Vars2,V1)) ->
                true ;
                (error_msg(not_statement,[Newp,F]),
                 discard_tmp_program,
                 fail)).


% rewrite_formula(Phi,Phi1) :- Phi1 is logically equivalent to Phi and uses
% only true, \+ (negation), and (conjunction), e (exists).

rewrite_formula(true,(\+ false)).
rewrite_formula(false,false).
rewrite_formula(F1,F2) :- atomic_formula(F1), !, F1=F2.
rewrite_formula((\+ F1),(\+ F2)) :- rewrite_formula(F1,F2).
rewrite_formula((F1 and F2), (R1 and R2)) :-
        rewrite_formula(F1,R1),
        rewrite_formula(F2,R2).
rewrite_formula((F1 or F2), (\+ ((\+ R1) and (\+ R2)))) :-
        rewrite_formula(F1,R1),
        rewrite_formula(F2,R2).
rewrite_formula((F1 -> F2), (\+ (R1 and (\+ R2)))) :-
        rewrite_formula(F1,R1),
        rewrite_formula(F2,R2).
rewrite_formula((F1 <-> F2),
                ((\+ (R1 and (\+ R2))) and (\+ ((\+ R1) and R2)))) :-
        rewrite_formula(F1,R1),
        rewrite_formula(F2,R2).
rewrite_formula(e(X,F),e(X,R)) :- rewrite_formula(F,R).
rewrite_formula(a(X,F),(\+ e(X,(\+ R)))) :- rewrite_formula(F,R).


% statement_into_clauses(Stat,Defs) :- Defs is the list of clauses derived by
% applying the typed LT transformation starting from the statement Stat.
% Conjunctions are eliminated in favour of flat lists.
% Clauses with false body are duplicate literals are deleted.
% Type atoms are added.

statement_into_clauses(Stat,Defs) :-
        ltt(Stat,[],Cls,_),
%        ltt(Stat,Cls),
        and_flatten_cls(Cls,Cls1),
        simplify(Cls1,Cls2),
        t_sort(Cls2,Cls3),
        types_of_defs(Cls3),
        add_types_to_all(Cls3,Defs).


% ltt(Stat,Defs) :- Defs is the list of clauses derived by
% applying the LT transformation starting from the statement Stat.
% Distinct clauses for different occurrences of the same subformula
% may be introduced.

ltt((H,true), [(H,true)]).

ltt((H,L), Sts) :- lit(L), !, Sts=[(H,L)].

ltt((_,(\+ true)), []).

ltt((H,(\+ (\+ F))), Sts) :-
        ltt((H,F), Sts).

ltt((H,(\+ (F1 and F2))),  Sts):-
         new_statement(Newp, (F1 and F2)),
         Sts=[(H,(\+ Newp)) | Sts12],
         ltt((Newp, (F1 and F2)),Sts12).

ltt((H, (\+ e(X,F))), Sts) :-
           new_statement(Newp, e(X,F)),
           Sts=[(H,(\+ Newp)) | Sts12],
           ltt((Newp,e(X,F)), Sts12).

ltt((H, (F1 and F2)), [(H, (G1 and G2)) | NewSts]) :-
        ltt((H,F1), [(H,G1) | NewSts1]),
        ltt((H,F2), [(H,G2) | NewSts2]),
        append(NewSts1,NewSts2,NewSts).

ltt((H, e(X,F)), Sts) :-
        drop_quantifier(e(X,F), NewF),
        ltt((H, NewF), Sts).


% ltt(+Stat,+Aux_Sts1,?Defs,?Aux_Sts2)
% ltt(Stat,[],Defs,_) :- Defs is the list of clauses derived by
% applying the LT transformation starting from the statement Stat.
% Distinct clauses for different occurrences of the same subformula are
% not introduced.
% The second and fourth arguments are used to keep track
% of all statements introduced during the LT transformation.

ltt((H,true), Aux_Sts, [(H,true)], Aux_Sts).
ltt((H,L), Aux_Sts1, Sts, Aux_Sts2) :-
        lit(L),
        !,
        Sts=[(H,L)],
        Aux_Sts1=Aux_Sts2.

ltt((_,(\+ true)), Aux_Sts, [], Aux_Sts).

ltt((H,(\+ (\+ F))), Aux_Sts1, Sts, Aux_Sts2) :-
        ltt((H,F), Aux_Sts1, Sts, Aux_Sts2).

ltt((H,(\+ (F1 and F2))), Aux_Sts1, Sts, Aux_Sts2):-
        member((A,Phi), Aux_Sts1), variant(Phi, (F1 and F2)),
        !,
        functor(A,Pred,_),
        fv((F1 and F2),Vars),
        Newp=..[Pred|Vars],
        Sts=[(H,(\+ Newp))],
        Aux_Sts2=Aux_Sts1.

ltt((H,(\+ (F1 and F2))), Aux_Sts1, Sts, Aux_Sts2):-
        Sts=[(H,(\+ Newp)) | Sts12],
        new_statement(Newp, (F1 and F2)),
        ltt((Newp, (F1 and F2)),
            [(Newp, (F1 and F2))|Aux_Sts1], Sts12, Aux_Sts2).

ltt((H, (\+ e(X,F))), Aux_Sts1, Sts, Aux_Sts2) :-
        member((A,Phi), Aux_Sts1), variant(Phi, e(X,F)),
        !,
        functor(A,Pred,_),
        fv(e(X,F),Vars),
        Newp=..[Pred|Vars],
        Sts=[(H,(\+ Newp))],
        Aux_Sts2=Aux_Sts1.

ltt((H, (\+ e(X,F))), Aux_Sts1, Sts, Aux_Sts2) :-
        new_statement(Newp, e(X,F)),
        Sts=[(H,(\+ Newp)) | Sts12],
        ltt((Newp,e(X,F)), [(Newp,e(X,F))|Aux_Sts1], Sts12,Aux_Sts2).

ltt((H, (F1 and F2)), Aux_Sts1, [(H, (G1 and G2)) | NewSts], Aux_Sts3) :-
        ltt((H,F1), Aux_Sts1, [(H,G1) | NewSts1], Aux_Sts2),
        ltt((H,F2), Aux_Sts2, [(H,G2) | NewSts2], Aux_Sts3),
        append(NewSts1,NewSts2,NewSts).

ltt((H, e(X,F)), Aux_Sts1, Sts, Aux_Sts2) :-
        drop_quantifier(e(X,F), NewF),
        ltt((H, NewF), Aux_Sts1, Sts, Aux_Sts2).


lit(L) :- atomic_formula(L), !.
lit(\+ L) :- atomic_formula(L).


% new_statement(Newp, F) :-  Newp :- F is a statement of the form
% Pred(V1,...,Vn) :- F, where:
% (i) Pred is a new predicate name, and
% (ii) V1,...,Vn  are the free variables of F.

new_statement(Newp, F) :-
        fv(F,Vars),
        gen_new_name_functor(Pred),
        Newp=..[Pred|Vars].


% and_flatten_cls(Cs,Ds) :-  transform the bodies of the
% statements in Cs from a conjunction (which uses the connective and) to a
% flat list.

and_flatten_cls([],[]).
and_flatten_cls([(H,B)|Cs],[(H,FB)|Ds]) :-
        and_flatten(B,FB),
        and_flatten_cls(Cs,Ds).

and_flatten(true,[]).
and_flatten((C1 and C2), F) :-
        !,
        and_flatten(C1,F1),
        and_flatten(C2,F2),
        append(F1,F2,F).
and_flatten(L,F) :- F=[L].



% simplify(Cs,Ds) :- simplify the clauses in the list Cs and derive
% the list Ds by:
% (i) Deleting subsumed clauses
% (ii) Deleting clauses with false body, and
% (iii) deleting duplicate literals.
% The predicate drop_useless_types which deletes type atoms for variables
% not occurring in the body will act as the identity when called
% from statement_into_clauses. However, it may not be the identity
% when called by apply_strat (see below).

simplify([],[]).
simplify([C1|Cs],Ds) :-
        del_elem(C2,Cs,Cs1),
        subsumes_clause_chk(C1,C2),
        !,
        simplify([C1|Cs1],Ds).
simplify([C1|Cs],Ds) :-
        member(C2,Cs),
        subsumes_clause_chk(C2,C1),
        !,
        simplify(Cs,Ds).
simplify([C|Cs],Ds) :-
        false_body(C),
        !,
        simplify(Cs,Ds).
simplify([(H,B)|Cs],[D|Ds]) :-
        delete_duplicates((H,[],B),C1),
        drop_useless_types(C1,D),
        simplify(Cs,Ds).

subsumes_clause_chk(C1,C2) :-
        copy_term(C1,(H1,B1)),
	copy_term(C2,(H2,B2)),
        subsumes(H1,H2),
        subsume(B1,B2),
	variant(C2,(H2,B2)). %needed. counterexample:  subsumes_clause_chk((h,[p(X1,Y1),p(Y1,X1)]),(h,[p(X,Y),p(Y,Z)])).

subsume([],_Us).
subsume([T|Ts],Us) :-
	member(U,Us),
	subsumes(T,U),
	subsume(Ts,Us).

/*
ground_representation(T,G) :-
	copy_term(T,G),
	term_variables(G,Vars),
	instantiate_vars(Vars,0).

instantiate_vars([],_).
instantiate_vars([X|Xs],N) :-
	X=var(N),
	N1 is N+1,
	instantiate_vars(Xs,N1).

subsumes_body([],_).
subsumes_body([L1|B1],B2) :-
        del_elem(L2,B2,B3),
        subsumes(L1,L2),
        subsumes_body(B1,B3).
*/

% false_body((H,B)) :- the body B of H:-B contains an atom A
% and its negation \+A

false_body((_,B)) :-
        member(A,B),
        member_chk((\+ A),B).

% member_chk(+X,+Ys) :- there is a member Y of the list of terms Ys
% (possibly with variables) such that X==Y. No variable is bound.

member_chk(X,[Y|_]) :- X==Y, !.
member_chk(X,[_|Ys]) :- member_chk(X,Ys).


% delete rightmost literal duplicates

delete_duplicates((H,B,[]),(H,B)).
delete_duplicates((H,B,[L|Ls]),(H,NewB)) :-
        member_chk(L,B),
        !,
        delete_duplicates((H,B,Ls),(H,NewB)).
delete_duplicates((H,B,[L|Ls]),(H,NewB)) :-
        append(B,[L],BL),
        delete_duplicates((H,BL,Ls),(H,NewB)).


% drop_useless_types((H,B),(H,B1)) :- B1 is obtained from B
% by deleting type atoms for variables which do not
% occur in B.
% NB: a type atom t(X) occurs to the left of any other occurrence of X
% in the body of a clause.

drop_useless_types((H,[]),(H,[])).
drop_useless_types((H,[L|Ls]),(H,Ms)) :-
        is_type_atom(L),
        \+ useful_type_atom(L,Ls),
        !,
        Ms=Ls1,
        drop_useless_types((H,Ls),(H,Ls1)).
drop_useless_types((H,[L|Ls]),(H,Ms)) :-
        Ms=[L|Ls1],
        drop_useless_types((H,Ls),(H,Ls1)).

is_type_atom(A) :-
        functor(A,P,1),
        is_type(P).

useful_type_atom(A,Ls) :-
        A=..[_,X],
        var(X),
        occurs_in_nonvar_or_neg_chk(X,Ls).

occurs_in_nonvar_or_neg_chk(V,B) :-
        member(L,B),
        L=..[_|Args],
        member(Arg,Args),
        nonvar(Arg),
        term_variables(Arg,Vs),
        member_chk(V,Vs).


% occurs_in_lit(+V,+B,?L) :- L is the leftmost literal of goal B
% such that variable V occurs in L

occurs_in_lit(V,[L|_],L1) :-
        term_variables(L,Vs),
        member_chk(V,Vs),
        !,
        L1=L.
occurs_in_lit(V,[_|Ls],L1) :-
        occurs_in_lit(V,Ls,L1).


% t_sort(+Cs,?S) :-  S is a list of clauses derived by reordering
% the clauses in Cs s.t. if C1 precedes C2 in S then
% no literal in the body of C1 depends on the head of C2

t_sort([],[]).
t_sort([C|Cs],[M|S]) :-
        del_minimal([C|Cs],M,Cs1),
        t_sort(Cs1,S).

del_minimal([M],M,[]).
del_minimal([C1,C2|Cs],M,Cs1) :-
        dep(C1,[C2|Cs]),
        !,
        Cs1=[C1|Cs2],
        del_minimal([C2|Cs],M,Cs2).
del_minimal([C1,C2|Cs],M,Cs1) :-
        M=C1,
        Cs1=[C2|Cs].


% dep(+C,+Cs) :- a predicate in the body of C depends on
% the head of a clause in Cs.

dep((_,B),Cs) :-
        member(L,B),
        pred(L,P/A),
        member((H1,_),Cs),
        pred(H1,P/A).


% pred(+L,?P/A) :-
% literal L has predicate symbol P and with arity A

pred(L,P/A) :-
        atomic_formula(L),
        !,
        functor(L,P,A).
pred(L,P/A) :-
        L=(\+ B),
        functor(B,P,A).


% types_of_defs(Defs) :- infer the types of all new predicates
% introduced by the definitions Defs and assert them as
% tmp_type_definition/3 facts

types_of_defs([]).
types_of_defs([(H,B)|Cs]) :-
        types_of_def((H,B)),
        types_of_defs(Cs).

types_of_def((H,B)) :-
        H=..[Pred|Vars],
        length(Vars,Arity),
        types_of_vars(Vars,B,List_of_types),
        assert(tmp_type_definition(Pred,Arity,List_of_types)).


% types_of_vars(Vars,G,Ts) :-  the variables Vars
% occurring in the goal G have types Ts

types_of_vars([],_,[]).
types_of_vars([V|Vars],G,[T|Ts]) :-
        occurs_in_lit(V,G,L),
        type_of_var(V,L,TV),
        TV=..[T,V],
        types_of_vars(Vars,G,Ts).


% type_of_var(V,L,TV) :- TV is the type atom of variable V
% occurring in literal L

type_of_var(V,L,TV) :-
        type_atoms(L,Tatoms),
        unfold_instantiated_types(Tatoms,Types_of_vars),
        occurs_in_lit(V,Types_of_vars,TV).


% type_atoms(+L,?Tatoms) :- Tatoms are the type atoms for the arguments
% of literal L, as declared by the type/3 predicate
% (stored as tmp_type_definition/3 facts)

type_atoms(L,Tatoms) :-
        is_type_atom(L),
        !,
        Tatoms=[L].
type_atoms(L,Tatoms) :-
        atomic_formula(L),
        !,
        functor(L,Pred,Arity),
        L=..[Pred|Args],
        (tmp_type_definition(Pred,Arity,List_of_Types) ->
                true;
                (error_msg(no_type,[Pred]),discard_tmp_program,fail)),
        map_types(List_of_Types,Args,Tatoms).
type_atoms(L,Tatoms) :-
        L=(\+ A),
        !,
        type_atoms(A,Tatoms).
type_atoms(L,_) :- error_msg(not_lit,[L]),discard_tmp_program,fail.

% map_types(List_of_Types,Args,Tatoms) :- Tatoms is the list of
% type atoms obtained by applying each type predicate in List_of_Types
% to the arguments in Args.

map_types([],[],[]).
map_types([T|Ts],[Arg|Args],[TA|TAs]) :-
        TA=..[T,Arg],
        map_types(Ts,Args,TAs).


% unfold_instantiated_types(Tatoms,Types_of_vars) :-
% Tatoms is a list of type atoms and
% Types_of_vars is a list of type atoms of the form t(X),
% where X is a variable. Types_of_vars is obtained by
% unfolding the atoms with non-variable arguments in Tatoms

unfold_instantiated_types([],[]).
unfold_instantiated_types([A|As],Tatoms) :-
        instantiated_type(A),
        !,
        tmp_program_clause(A,B),
        append(B,As,BAs),
        unfold_instantiated_types(BAs,Tatoms).
unfold_instantiated_types([A|As],[A|Tatoms]) :-
        unfold_instantiated_types(As,Tatoms).

instantiated_type(TA) :-
        TA=..[P,Arg],
        is_type(P),
        nonvar(Arg).


%%% add_types_to_all(Cs,Ds) :-  Ds is a list of clauses obtained by adding
%%% type atoms  to all clauses of Cs

add_types_to_all([],[]).
add_types_to_all([C|Cs],[D|Ds]) :-
        add_types(C,D),
        add_types_to_all(Cs,Ds).

% add_types(C,D) :- D is derived from C by adding a type atom
% for each variable occurring in an instantiated or negated atom

add_types((H,B),(H,NewB)) :-
        vars_to_be_typed(B,Vars),
        add_type_of_vars(Vars,B,NewB).


% vars_to_be_typed(B,Vars) :-  Vars is the set of variables
% which occur in an nonvariable argument of a literal or in a
% negated atom in B


vars_to_be_typed(B,Vars) :-
        setof(V, occurs_in_nonvar_or_neg(V,B), Vars) -> true ; Vars=[].
%        term_variables(B,Vars). %type all vars

occurs_in_nonvar_or_neg(V,B) :-
        member(L,B),
        L=..[_|Args],
        member(Arg,Args),
        nonvar(Arg),
        term_variables(Arg,Vs),
        member(V,Vs).


% add_type_of_vars(Vars,B,NewB) :- NewB is a goal derived from goal B
% by adding to B a type atom for each variable occurring in
% Vars

add_type_of_vars([],B,B).
add_type_of_vars([V|Vars],B,[TV|NewB]) :-
        occurs_in_lit(V,B,L),
        type_of_var(V,L,TV),
        add_type_of_vars(Vars,B,NewB).


% ufs(Defs) :- apply the transformation strategy UFS to P U Defs.
% In the list of definitions Defs = [D1,...,Dn] a predicate occurring in
% the body of Dj is defined in Di (i<j) or in the initial program P.

ufs(Defs) :-
        initialize_definitions(Defs),
        forall_defs(Defs,usdf),
        discard_definitions.


% initialize_definitions(Defs) :-  the set of definitions, represented
% as tmp_definition_clause/2 facts, is initialized to Defs

initialize_definitions(Defs) :-
      retractall(tmp_definition_clause(_,_)),
      assertall_defs(Defs).

assertall_defs([]).
assertall_defs([(H,B)|Defs]) :-
%        sort(B,B1),
        goal_hash(B,N),
        assert(tmp_definition_clause(N,(H,B))),
        assertall_defs(Defs).


% forall_defs(Cs,S) :-  for all clauses C in Cs
% (i) delete useless clauses and clauses subsumed by facts
% (ii) apply the strategy S starting from C

forall_defs([],_).
forall_defs([(H,B)|Ds],Strategy) :-
        delete_useless_and_subsumed_by_facts(B),
        apply_strat(Strategy,[(H,B)]),
        forall_defs(Ds,Strategy).


% delete_useless_and_subsumed_by_facts(Goal) :-
% for all predicates p/n in Goal transform
% the definition of p into the empty set of clauses if p is useless
% or into the single clause p(X1,...,Xn).  if in the program there is
% one clause of the form p(X1,...,Xn). 

delete_useless_and_subsumed_by_facts(B) :-
        list_of_preds(B,Ps),
        refl_depend_on(Ps,Preds),
        useless(Preds,UselessPreds),
        delete_defs(UselessPreds),
        replace_by_facts(Preds,UselessPreds,TruePreds),
        (derivation_trace_on ->
          (nl, write(' -delete useless and subsumed clauses- '), nl,
          write('inferred falsity of '),
          (UselessPreds=[] ->
                write('no predicate') ; write_all(UselessPreds)),
          nl,
          write('inferred truth of '),
          (TruePreds=[] ->
                write('no predicate') ; write_all(TruePreds)),
          nl) ; true).


% refl_depend_on(P1,P2) :- P2 is the set of predicates (with arity) which
% are in the reflexive closure of the depends relation
% of a predicate (with arity) in P1

refl_depend_on(P1,P2) :-
        immediately_depend_on(P1,P),
        (P1=P -> P1=P2 ; refl_depend_on(P,P2)).

immediately_depend_on(P1,P2) :-
        setof(Pred/Arity,d(Pred/Arity,P1),Preds) -> P2=Preds ; P2=[].

d(Pred/Arity,P1) :-
        member(Pred/Arity,P1);
        (member(Pred1/Arity1,P1),
        functor(H,Pred1,Arity1),
        tmp_program_clause(H,Bd),
        member(A,Bd),
        pred(A,Pred/Arity)).


% useless(P,UP) :- P is a set of predicates occurring in the
% program whose clauses are stored as tmp_program_clause(H,B) facts
% and UP is the subset of useless predicates

useless(P0,Gfp) :-
        transf(P0,P),
        (P0=P -> Gfp=P0 ; useless(P,Gfp)).

transf(P1,P2) :-
        setof(P/A,
        (member(P/A,P1),
         \+ has_useful_body(P/A,P1)),
        Preds) -> P2=Preds ; P2=[].

has_useful_body(P/A,P1) :-
        functor(H,P,A),
        tmp_program_clause(H,Bd),
        \+ has_useless_atom(Bd,P1).

has_useless_atom(Bd,P1) :-
        member(Q,Bd),
	  atomic_formula(Q),
        pred(Q,Pred/A),
        member(Pred/A,P1).


% delete_defs(Ps) :- The definition of every predicate P/A in Ps is
% deleted

delete_defs([]).
delete_defs([P/A|Ps]) :-
        functor(H,P,A),
        retractall(tmp_program_clause(H,_)),
        delete_defs(Ps).


% replace_by_facts(+Preds,+UselessPreds,?TruePreds) :-
% replace the definition of p by a single fact p.
% for every nullary predicate p in Pred which is not useless and
% replace the definition of p by a single fact (p(X1,...,Xn).
% for every predicate p in Pred such that in the program there is
% a clause of the form p(X1,...,Xn).

replace_by_facts([],_,[]).
replace_by_facts([Pred/A|Preds],UselessPreds,[Pred/A|TruePreds]) :-
        A=0,
        \+member(Pred/0,UselessPreds),
        !,
        retractall(tmp_program_clause(Pred,_)),
        assert(tmp_program_clause(Pred,[])),
        replace_by_facts(Preds,UselessPreds,TruePreds).
replace_by_facts([Pred/A|Preds],UselessPreds,[Pred/A|TruePreds]) :-
        functor(H,Pred,A),
        tmp_program_clause(H,[]),
        functor(H1,Pred,A),
        variant(H,H1),
        !,
        retractall(tmp_program_clause(H,_)),
        assert(tmp_program_clause(H,[])),
        replace_by_facts(Preds,UselessPreds,TruePreds).
replace_by_facts([_|Preds],UselessPreds,TruePreds) :-
        replace_by_facts(Preds,UselessPreds,TruePreds).


% apply_strat(usdf,Ds) :- the strategy usdf is applied to
% every definition in Ds

apply_strat(usdf,[]).
apply_strat(usdf,[D|Ds]) :-
        (derivation_trace_on ->
          (nl, write(' -define- '), nl, show_clause(0,D),  nl, nl);
          true),
        bf_unfold(D,UnfoldedD),
        (derivation_trace_on ->
          (write(' -unfold- '), nl, show_all(UnfoldedD), nl);
          true),
        simplify(UnfoldedD,SimplifiedD),
        (derivation_trace_on ->
          (write(' -simplify- '), nl, show_all(SimplifiedD), nl);
          true),
        define_fold(SimplifiedD,NewDefs,FoldedD),
        (derivation_trace_on ->
          (write(' -fold- '), nl, show_all(FoldedD), nl);
          true),
        append(NewDefs,Ds,NewDs),
        apply_strat(usdf,NewDs).


% UNFOLDING STRATEGY:  breadth-first unfold
% unfold all clauses wrt all literals, from left to
% right (so that type atoms are unfolded first)

bf_unfold((H,B), UnfoldedC) :-
        unfold_all_clauses([(H,[],B)], UnfoldedC).


% unfold_all_clauses(Cs,U) :- U is the result of unfolding all clauses
% in the lists Cs. A clause in Cs is represented as (H,B1,B2),
% where B1@B2 is the body. The first element of B2 is the literal selected
% for unfolding (if B2=[] no unfolding is performed)

unfold_all_clauses([],[]).
unfold_all_clauses([(H,B,[])|Cs],[(H,B)|U]) :-
        unfold_all_clauses(Cs,U).
unfold_all_clauses([(H,B,[L|Ls])|Cs],UCs) :-
        unfold_one_clause((H,B,[L|Ls]),UC),
        append(UC,Cs,NewCs),
        unfold_all_clauses(NewCs,UCs).

unfold_one_clause((H,B,[A|Ls]), UC) :-
        atomic_formula(A),
        !,
        findall(C1,pos_unfold((H,B,[A|Ls]),C1),UC).
unfold_one_clause((H,B,[(\+ A)|Ls]), UC) :-
        neg_unfoldable(A),
        !,
        findall(C1,neg_unfold((H,B,[(\+ A)|Ls]),C1),UC).
unfold_one_clause((_,_,[(\+ A)|_]),_) :-
        error_msg(neg_lit,[(\+ A)]),
        discard_tmp_program,
        fail.

neg_unfoldable(A) :-
        \+ exists_head_which_binds_or_body_with_existential_vars(A).

exists_head_which_binds_or_body_with_existential_vars(A) :-
        functor(A,P,Arity),
        functor(H,P,Arity),
        tmp_program_clause(H,B),
        unifiable(A,H),
        binds_or_body_with_existential_vars(A,H,B).

binds_or_body_with_existential_vars(A,H,_):-
        \+ subsumes_chk(H,A),
        !.
binds_or_body_with_existential_vars(_,H,B):-
        existential_variable((H,B),_),
        !.

%existential_variable(+(H,B),?X) not functional

existential_variable((H,B),X) :-
          term_variables(H,VH),
          term_variables(B,VB),
          member(X,VB),
          \+ member_chk(X,VH).

% unifiable(A,B) :-
%         copy_term(A,A1),
%         copy_term(B,B1),
%         A1=B1.

pos_unfold((H,B,[A|Ls]),(H,B1,Ls)) :-
        tmp_program_clause(A,Body),
        append(B,Body,B1).

neg_unfold((H,B,[(\+ A)|Ls]), (H,B1,Ls)) :-
        complete_def(A,Def),
        negated(Def,Body),
        append(B,Body,B1).

complete_def(A,Def) :-
        setof(Body,tmp_program_clause(A,Body),D) -> Def=D ; Def=[].

% negated(Gs,NegG) :- Gs is a list of goals G1,...,Gn, and NegG is a goal
% consisting of the complement of a literal L in Gk, for k=1,...,n

negated([],[]).
negated([G|Gs],[L1|NegG]) :-
        member(L,G),
        complement(L,L1),
        negated(Gs,NegG).

complement(L,C) :- L= (\+ A) -> C=A ; C=(\+ L).


% SIMPLIFY STRATEGY
% Delete subsumed clauses, delete clauses with false body,
% delete duplicate literals, % delete types for variable which do not occur
% in the body.
% The predicate simplify(Cs,Ds) is defined above


% DEFINE & FOLD strategy
% Single clause, block-folding

define_fold([],[],[]).
define_fold([C|Cls],NewDefs,[FoldedC|FoldedCls]) :-
        define_fold_one_clause(C,NewDC,FoldedC),
        append(NewDC,NewDCls,NewDefs),
        define_fold(Cls,NewDCls,FoldedCls).

define_fold_one_clause((H,B),NewDC,(H,FldB)) :-
        blocks(B,Blocks),
        foldall_blocks(Blocks,(H,FldB),NewDC),
        add_to_program((H,FldB)).


% blocks(G,Bs) :- goal G is partitioned into blocks Bs, i.e.,
% a list of subgoals without common variables

blocks([],[]).
blocks([L|Ls],[B|Bs]) :-
        blocks(Ls,Blocks),
        partition(L,Blocks,Bs1,Bs),
        flatten([[L]|Bs1],B).


% partition(L,Blocks,Bs1,Bs) :- Blocks is a list of blocks,
% Bs1 is the sublist of all blocks in Blocks which share
% variables with L, and Bs is the sublist of all other blocks

partition(_,[],[],[]).
partition(A,[B|Bs],NewBs1,NewBs2) :-
        share_var(A,B) ->
        (NewBs1=[B|Bs1], NewBs2=Bs2, partition(A,Bs,Bs1,Bs2)) ;
        (NewBs1=Bs1, NewBs2=[B|Bs2], partition(A,Bs,Bs1,Bs2)).


%share_var(A,B) :- terms A and B have a variable in common

share_var(A,B) :-
        term_variables(A,VA),
        term_variables(B,VB),
        member(X,VA),
        member_chk(X,VB).

/*
goal_hash(G,H) :-
         skeletons(G,Ps),
         length(G,N),
         N1 is N+5,
         term_hash(Ps,N1,10000,H).  % imported from library(term)
*/

goal_hash([],0).
goal_hash([L|Ls],N) :-
        skeleton(L,S),
        term_hash(S,4,10000,H),
        goal_hash(Ls,N1),
        N is N1+H.

list_of_preds([],[]).
list_of_preds([L|Ls],[P/A|Ps]) :-
        pred(L,P/A),
        list_of_preds(Ls,Ps).

skeleton(X,S) :- var(X), !, S=variable_name.
skeleton(A,S) :- atomic(A), !, S=A.
skeleton(T,S) :-
        T=..[F|Args],
        skeletons(Args,Ss),
        S=..[F|Ss].

skeletons([],[]).
skeletons([T|Ts],[S|Ss]) :-
        skeleton(T,S),
        skeletons(Ts,Ss).


% foldall_blocks(Bs,(H,FldBs),NewDCs) :-  we introduce the new
% definitions NewDCs and by folding all blocks
% Bs we derive the clause H:-FldBs

foldall_blocks([],(_,[]),[]).
foldall_blocks([B|Bs],(H,[H1|FldBs]),NewDCs) :-
        foldable((H,B),C), !,
        head(C,H1),
        foldall_blocks(Bs,(H,FldBs),NewDCs).
foldall_blocks([B|Bs],(H,[H1|FldBs]),[(H1,B1)|NewDCs]) :-
        introduce_definition((H,B),(H1,B1)),
        foldall_blocks(Bs,(H,FldBs),NewDCs).

% foldable((H,B),(H1,B1)) :-  H1:-B1 is a definition clause
% such that B is a variant of B1 and, after unifying B and B1,
% forall X in B, X existential in H:-B iff X existential in H1:-B

foldable((H,B),(H1,B1)) :-
        goal_hash(B,N),
        tmp_definition_clause(N,(H1,B1)),
        v_sort(B,B1,B2), %B=B2, %
        variant(B2,B1),
        B2=B1,
        linking_vars((H,B2),LV),
        linking_vars((H1,B1),LV1),
        LV==LV1.    %LV and LV1 are sorted because are computed by setof

head((H,_),H).

v_sort([],[],[]).
v_sort(G,[L|Ls],[L1|Ls1]) :-
        del_elem(L1,G,G1),
        variant(L,L1),
        v_sort(G1,Ls,Ls1).

del_elem(X,[X|Xs],Xs).
del_elem(X,[Y|Xs],[Y|Ys]) :- del_elem(X,Xs,Ys).


introduce_definition((H,B),(Newp,B)) :-
        gen_new_name_functor(Pred),  % imported from determ_prog_util.pl
        linking_vars((H,B),LinkingVars),
        Newp=..[Pred|LinkingVars],
        goal_hash(B,N),
        assert(tmp_definition_clause(N,(Newp,B))),
        types_of_def((Newp,B)).

linking_vars((H,B),LinkingVars) :-
        setof(X,(
        term_variables(B,VB),
        term_variables(H,VH),
        member(X,VB),
        member_chk(X,VH)), LV) -> LinkingVars=LV ; LinkingVars=[].

add_to_program((H,FldB)) :-
        assert(tmp_program_clause(H,FldB)).


% inspect(+Pred,?A):-
% Pred is a nullary atom.
% if the fact Pred. belongs to the final program then A is true;
% If no clause Pred :- Body. belongs to the final program then A is false.

inspect(Pred,A) :-
        tmp_program_clause(Pred,Body) ->
        (Body=[] -> A=true ; A='proof failed') ;  A=false.

synthesized_prog(Atom,Program) :-
        functor(Atom,P,N),
        refl_depend_on([P/N],Ps),
        select_clauses_for(Ps,Program),
        select_types_for(Ps).

select_clauses_for(Ps,Program) :-
        findall((H,B),
        (member(P/N,Ps),
         functor(H,P,N),
         tmp_program_clause(H,B)),
        Program).

select_types_for(Ps) :-
        findall(P,
        (member(P/A,Ps),
         tmp_type_definition(P,A,L),
         assert(type_definition(P,A,L))),
        _).

discard_tmp_program :-
        retractall(tmp_program_clause(_,_)),
        retractall(tmp_type_definition(_,_,_)),
        retractall(is_type(_)).

discard_definitions :-
%        findall(C,tmp_definition_clause(N,C),Ds),
%        nl, write_all(['Defs: '|Ds]),
%        show_all(Ds),
%        length(Ds,L),
%        nl, write('# of defs: '), write(L), nl,
        retractall(tmp_definition_clause(_,_)).

write_all([]).
write_all([X|Xs]) :- nl, write(X), write_all(Xs).


update_transf_tree(Ns,SynthesizedP) :-
        add_clauses(Ns,SynthesizedP,synth),  %from environment.pl
        update_programs((synth,Ns),_).       %from environment.pl


error_msg(not_statement,[Newp,F]) :-
        write('ERROR: '), write(Newp), write(' and '), write(F),
        write(' have different free variables').
error_msg(not_atom,[Newp]) :-
        write('ERROR: '), write(Newp), write(' is not an atom').
error_msg(not_new_pred,[P]) :-
        write('ERROR: '), write(P), write(' is not a new predicate').
error_msg(no_type,[Pred]) :-
        write('ERROR: no type definition for '), write(Pred).
error_msg(not_lit,[L]) :-
        write('ERROR: '), write(L), write(' is not a literal').
error_msg(neg_lit,[L]) :-
        write('ERROR: negative literal'), write(L),
        write(' cannot be unfolded').

ask_user(redefine,[P/A]) :-
        write(P/A),
        write(' is not a new predicate. Redefine? (y. or n.)'), nl,
        read(Answer),
        (Answer=y;Answer=n), !,
        ((Answer = y) ->
                (delete_from_current_prog(P/A)) ;
                 discard_tmp_program, fail).

delete_from_current_prog(P/A) :-
        findall(N,cl_number_in_current_prog(P/A,N),Cls),
        del_cl(Cls).

cl_number_in_current_prog(P/A,N) :-
        current_prog(_,Ns),
        member(N,Ns),
        functor(H,P,A),
        clause3(N,(H,_),_).

