% unfolding_rules.pl
%
% This file contains the predicates which specify the effects of applying the
% transformation rules associated with the predicates unfold, unfold_d,
% unfold_d_i, unfold_d_unit, unfold_d_eq, defined in top_level.pl, after
% selection and checking of the input parameters.

:- use_module(library(lists), [ append/3, member/2 ]).
:- use_module(library(terms), [ variant/2]). %, unify_with_occurs_check/2 ]).

/*-------
apply_unfold(C,A): unfold clause C w.r.t. atom A using clauses in initial
program.
---------*/
apply_unfold(Clause_Number,Atom_Number) :- 
   current_prog(PN,CNs),
   (label_arc(PN,(Clause_Number,(unf,Atom_Number),NewCNs),PNSucc) -> 
   (label_node(PNSucc,CNsSucc),
    update_current_prog((PN,CNs),(PNSucc,CNsSucc))); 
   (apply_unfold(Clause_Number,Atom_Number,Clauses),
   add_clauses(NewCNs,Clauses,nil),
   update_programs((Clause_Number,(unf,Atom_Number),NewCNs), PNSucc))), 
   (derivation_trace_on ->
        new_clauses(unfold,Clause_Number,Atom_Number,NewCNs)
        ;true).

apply_unfold(Clause_Number,_,[Clause]) :- 
    clause_number_to_clause(Clause_Number,Clause),
    is_a_unit_clause(Clause), !.

apply_unfold(Clause_Number,Atom_Number,Clauses) :- 
    clause_number_to_clause(Clause_Number,(Head,Body1)),
    split(Body1,Atom_Number,Before,Lit,After),
    (is_an_equality(Lit,LHS,RHS) ->
        (unify_with_occurs_check(LHS,RHS),
         append(Before,After,Body),
         Clauses = [(Head,Body)]) ;
    (is_a_negative_literal(Lit,Atom) -> 
        apply_unfold_neg(Clause_Number,Head,Before,Atom,After,
			 Atom_Number,Clauses) ;
        apply_unfold_pos(Clause_Number, Atom_Number,Clauses))).

apply_unfold_pos(Clause_Number, Atom_Number,Clauses) :-
   bagof((Hd,Bd), one_unf(Clause_Number,Atom_Number,(Hd,Bd)), Clauses), !.
apply_unfold_pos(Clause_Number, Atom_Number,_) :-
   undefined_predicate(1,Clause_Number,Atom_Number,F/A), !,
   msg_undefined_predicate_in_clause(F/A,1,Clause_Number), fail.
    
one_unf(Clause_Number,Atom_Number,(Head,Bd)) :-
    clause_number_to_clause(Clause_Number,(Head,Body1)),
    split(Body1,Atom_Number,Before,Atom,After),
    clause3(_,(Atom2,R),prog),
    unify_with_occurs_check(Atom2,Atom),
    append(Before,R,After,Bd).

apply_unfold_neg(C,Head,Before,Atom,After,Atom_Number,Clauses) :-
    bagof((Hd,Bd), clause_with_head_unifiable_with(Atom,Hd,Bd), L), !,
    check_instantiation_head_before_unfold_neg(C,Atom,Atom_Number,L,Bodies),
     apply_unfold_neg(Head,Before,After,Bodies,Clauses).

apply_unfold_neg(C,Head,Before,Atom,After,_,[(Head,Bd)]) :-
   (undefined_predicate(1,Atom) ->
         (predname_and_arity(Atom,F,N),
      msg_undefined_predicate_in_clause(F/N,1,C), fail)
    ; append(Before,After,Bd)).

apply_unfold_neg(Head,Before,After,Bodies,Clauses) :-
    apply_de_morgan(Bodies,L2),
    apply_unfold_neg_(Head,Before,After,L2,Clauses).

apply_unfold_neg_(_,_,_,[],[]).
apply_unfold_neg_(Head,Before,After,[E|L2],[(Head,Bd)|Clauses]) :-
    append(Before,E,After,Bd),
    apply_unfold_neg_(Head,Before,After,L2,Clauses).

apply_de_morgan([Conj],Disj) :- !,
   apply_de_morgan2(Conj,Disj).
apply_de_morgan([Conj|Disj],Disj2) :-
   apply_de_morgan(Disj,Conj1),
   apply_de_morgan_(Conj,Conj1,Disj2).

apply_de_morgan_([],_,[]).
apply_de_morgan_([E|Conj],Conj1, Conj3) :-
   add_to_all_disjs(E,Conj1,Conj2),
   apply_de_morgan_(Conj,Conj1,L),
   append(Conj2,L,Conj3).

apply_de_morgan2([],[]).
apply_de_morgan2([\+(E)|Conj],[[E]|Disj]) :-!,
   apply_de_morgan2(Conj,Disj).
%  addition of the following two clauses 03.10.97
apply_de_morgan2([false|Conj],[[]|Disj]) :-!,
   apply_de_morgan2(Conj,Disj).
apply_de_morgan2([fail|Conj],[[]|Disj]) :-!,
   apply_de_morgan2(Conj,Disj).
apply_de_morgan2([E|Conj],[[\+ E]|Disj]) :-
   apply_de_morgan2(Conj,Disj).


add_to_all_disjs(_,[],[]).
add_to_all_disjs(\+(E),[Disj|Disjs],[[E|Disj]|L]) :-!,
   add_to_all_disjs(E,Disjs,L).
%  addition of the following two clauses 03.10.97
add_to_all_disjs(false,L,L).
add_to_all_disjs(fail,L,L).
add_to_all_disjs(E,[Disj|Disjs],[[\+ E|Disj]|L]) :-
   add_to_all_disjs(E,Disjs,L).

/*-------
apply_unfold_d(C,A): unfold_d clause C w.r.t. atom A, i.e. unfold C w.r.t. A and delete C
if A corresponds to a failing goal.
---------*/
apply_unfold_d(Clause_Number,Atom_Number) :-
    current_prog(PN,CNs),
    (label_arc(PN,(Clause_Number,(uunf,Atom_Number),NewCNs),PNSucc) -> 
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)),
    (derivation_trace_on ->
        new_clauses(unfold_d,Clause_Number,Atom_Number,NewCNs);true));
    (apply_unfold(Clause_Number,Atom_Number,Clauses) ->
    (add_clauses(NewCNs,Clauses,nil),
    update_programs((Clause_Number,(uunf,Atom_Number),NewCNs), PNSucc),
    (derivation_trace_on ->
        new_clauses(unfold_d,Clause_Number,Atom_Number,NewCNs);true)) ; 
     (undefined_predicate(1,Clause_Number,Atom_Number,_) -> 
          (is_a_negative_goal(Clause_Number,Atom_Number) -> 
	     (yesno('Do you want to apply goal deletion',n) -> 
	      apply_delete_goal(Clause_Number,[Atom_Number]) ; fail) ;
	     (yesno('Do you want to apply clause deletion',n) -> 
              apply_del_cl_failing_bd(Clause_Number) ; fail))
          ; (is_a_negative_goal(Clause_Number,Atom_Number,NGoal) -> 
             (exists_clause_with_head_unifiable_with(1,NGoal) -> fail ;
              apply_delete_goal(Clause_Number,[Atom_Number])) ;
           apply_del_cl_failing_bd(Clause_Number))))).

/*-------
apply_unfold_d_p(C,A): unfold_d clause C w.r.t. A, using clauses in current program.
---------*/
apply_unfold_d_p(Clause_Number,Atom_Number) :- current_prog(PN,CNs),
    (label_arc(PN,(Clause_Number,(uunf_i,PN,Atom_Number),NewCNs),PNSucc) -> 
    (label_node(PNSucc,CNsSucc), update_current_prog((PN,CNs),(PNSucc,CNsSucc)),
    (derivation_trace_on ->
        new_clauses(unfold_d_p,Clause_Number,Atom_Number,NewCNs);true)); 
    (apply_unfold_d_p(Clause_Number,Atom_Number,Clauses) ->
    (add_clauses(NewCNs,Clauses,nil),
    update_programs((Clause_Number,(uunf_i,PN,Atom_Number),NewCNs), PNSucc),
    (derivation_trace_on ->
        new_clauses(unfold_d_p,Clause_Number,Atom_Number,NewCNs);true)) ;

    (undefined_predicate(PN,Clause_Number,Atom_Number,_) -> 
          (is_a_negative_goal(Clause_Number,Atom_Number) -> 
	     (yesno('Do you want to apply goal deletion',n) -> 
	      apply_delete_goal(Clause_Number,[Atom_Number]) ; fail) ;
	     (yesno('Do you want to apply clause deletion',n) -> 
              apply_del_cl_failing_bd(Clause_Number) ; fail))
         ; (is_a_negative_goal(Clause_Number,Atom_Number,NGoal) -> 
             (exists_clause_with_head_unifiable_with(1,NGoal) -> fail ;
	   apply_delete_goal(Clause_Number,[Atom_Number])) ;
           apply_del_cl_failing_bd(Clause_Number) ) ) )).

apply_unfold_d_p(Clause_Number,_,[Clause]) :- 
    clause_number_to_clause(Clause_Number,Clause),
    is_a_unit_clause(Clause), !.

apply_unfold_d_p(Clause_Number,Atom_Number,Clauses) :- 
    clause_number_to_clause(Clause_Number,(Head,Body1)),
    split(Body1,Atom_Number,Before,Lit,After),
   current_program(PN),
    (is_an_equality(Lit,LHS,RHS) ->
        (unify_with_occurs_check(LHS,RHS),
         append(Before,After,Body),
         Clauses = [(Head,Body)]) ;
    (is_a_negative_literal(Lit,Atom) -> 
        apply_unfold_d_i_neg(PN,Clause_Number,Head,Before,Atom,After,
			     Atom_Number,Clauses)
       ; apply_unfold_d_i_pos(PN,Clause_Number, Atom_Number,Clauses))).

apply_unfold_d_i_pos(PN,Clause_Number, Atom_Number,Clauses) :-
    bagof((Hd,Bd), one_unf_i(PN,Clause_Number,Atom_Number,(Hd,Bd)), Clauses), !.
apply_unfold_d_i_pos(PN,Clause_Number, Atom_Number, _) :-
   undefined_predicate(PN,Clause_Number,Atom_Number,F/A), !,
   msg_undefined_predicate_in_clause(F/A,PN,Clause_Number), fail.

apply_unfold_d_i_neg(Pi,C,Head,Before,Atom,After,Atom_Number,Clauses) :-
    bagof((Hd,Bd), clause_with_head_unifiable_with(Pi,Atom,Hd,Bd), L), !,
    check_instantiation_head_before_unfold_neg(C,Atom,Atom_Number,L,Bodies),
    apply_unfold_neg(Head,Before,After,Bodies,Clauses).

apply_unfold_d_i_neg(Pi,C,Head,Before,Atom,After,_,[(Head,Bd)]) :-
   (undefined_predicate(Pi,Atom) ->
         (predname_and_arity(Atom,F,A),
         msg_undefined_predicate_in_clause(F/A,Pi,C),
         fail)
    ; append(Before,After,Bd)).

/*-------
apply_unfold_d_i(Pi,C,A): unfold_d clause C w.r.t. atom A, using clauses in program Pi.
---------*/
apply_unfold_d_i(Pj,Clause_Number,Atom_Number) :- 
    current_prog(PN,CNs), 
    (label_arc(PN,(Clause_Number,(uunf_i,Pj,Atom_Number),NewCNs),PNSucc) -> 
    (label_node(PNSucc,CNsSucc),  update_current_prog((PN,CNs),(PNSucc,CNsSucc)),
    (derivation_trace_on ->
        new_clauses(unfold_d_i,Pj,Clause_Number,Atom_Number,NewCNs);true));
    (apply_unfold_d_i(Clause_Number,Atom_Number,Pj,Clauses) ->
    (add_clauses(NewCNs,Clauses,nil),
    update_programs((Clause_Number,(uunf_i,Pj,Atom_Number),NewCNs), PNSucc), 
    (derivation_trace_on ->
        new_clauses(unfold_d_i,Pj,Clause_Number,Atom_Number,NewCNs);true)) 
;     (undefined_predicate(Pj,Clause_Number,Atom_Number,_) ->
          (is_a_negative_goal(Clause_Number,Atom_Number) ->
	     (yesno('Do you want to apply goal deletion',n) ->
       apply_delete_goal(Clause_Number,[Atom_Number]) ; fail) ;
	     (yesno('Do you want to apply clause deletion',n) ->
              apply_del_cl_failing_bd(Clause_Number) ; fail))
         ; (is_a_negative_goal(Clause_Number,Atom_Number,NGoal) ->
             (exists_clause_with_head_unifiable_with(1,NGoal) -> fail ;
         apply_delete_goal(Clause_Number,[Atom_Number])) ;
           apply_del_cl_failing_bd(Clause_Number) ) ) )).

apply_unfold_d_i(Clause_Number,_,_,[Clause]) :- 
    clause_number_to_clause(Clause_Number,Clause),
    is_a_unit_clause(Clause), !.

apply_unfold_d_i(Clause_Number,Atom_Number,Pj,Clauses) :- 
   clause_number_to_clause(Clause_Number,(Head,Body1)),
   split(Body1,Atom_Number,Before,Lit,After),
    (is_an_equality(Lit,LHS,RHS) ->
        (unify_with_occurs_check(LHS,RHS),
         append(Before,After,Body),
         Clauses = [(Head,Body)]) ;
    (is_a_negative_literal(Lit,Atom) -> 
        apply_unfold_d_i_neg(Pj,Clause_Number,Head,Before,Atom,After,
           Atom_Number,Clauses) ;
    apply_unfold_d_i_pos(Pj,Clause_Number, Atom_Number,Clauses) )).

one_unf_i(_,Clause_Number,_,(Head,[])) :-
    clause_number_to_clause(Clause_Number,(Head,[])),!.

one_unf_i(_,Clause_Number,Atom_Number,(Head,Body)) :-
    clause_number_to_clause(Clause_Number,(Head,Body1)),
    split(Body1,Atom_Number,Before,Atom,After),
    is_an_equality(Atom,LHS,RHS), !,
    unify_with_occurs_check(LHS,RHS),
    append(Before,After,Body).

one_unf_i(Pj,Clause_Number,Atom_Number,(Head,Body)) :-
    clause_number_to_clause(Clause_Number,(Head,Body1)),
    split(Body1,Atom_Number,Before,Atom,After),
    label_node(Pj,Clauses_Pj), 
    member(D,Clauses_Pj),
    clause_number_to_clause(D,(HeadD,BodyD)),
    unify_with_occurs_check(HeadD,Atom), append(Before,BodyD,After,Body).

/*-----------
apply_simpl_body(C): simplify the body of clause C by unfolding instances of facts,
deleting clauses with a failing body, and simplifying equalities and inequalities if any. This
predicates is used by the top level predicate unfold_d_units.
------------*/
apply_simpl_body(C) :- clause_number_to_clause(C,HB), apply_simpl_body(C,HB,_).
apply_simpl_body(C,(H,B), X):-
     apply_simpl_body(B,B2,0,[],Lit_Numbers),
     (Lit_Numbers==failing -> (apply_delete_clause(C), X=removed(C));
     (Lit_Numbers==[] -> X=C; (add_clauses(NewCN,[(H,B2)],nil), NewCN=[X],
     update_programs((C, (unfold_facts,Lit_Numbers),NewCN),_), 
     (derivation_trace_on ->
        new_clause(unfold_facts,C,Lit_Numbers,NewCN);true)))).
apply_simpl_body([\+(A=B)|_],_,_,_,failing) :-    A==B,    !.
apply_simpl_body([\+(A=B)|R],[\+(A=B)|R2],N,L,L2) :-
    \+ \+ unify_with_occurs_check(A,B), !, N2 is N + 1,  apply_simpl_body(R,R2,N2,L,L2).
apply_simpl_body([\+(_=_)|R],R2,N,L,L2) :- !, N2 is N + 1, apply_simpl_body(R,R2,N2,[N2|L],L2).
apply_simpl_body([G|R],R2,N,L,L2) :-
    current_clauses(CNs), member(Clause,CNs), clause_number_to_clause(Clause,(H,[])),
    anti_unify(G,H,T), variant(T,H),!, N2 is N + 1, apply_simpl_body(R,R2,N2,[N2|L],L2).
apply_simpl_body([G|R],[G|R2], N,L,L2) :-
    current_clauses(CNs), member(Clause,CNs), head_of_clause(Clause, H),
    \+ \+ G = H, !, N2 is N + 1, apply_simpl_body(R,R2,N2,L,L2).
apply_simpl_body([G|R],[G|R2], N,L,L2) :-
    current_program(PN),
    undefined_predicate(PN,G), !, N2 is N + 1, apply_simpl_body(R,R2,N2,L,L2).
apply_simpl_body([G|R],[G|R2], N,L,L2) :-
    (predicate_property(G,_) ; functor(G,\+,_) ; functor(G,not,_)),!,
    N2 is N + 1, apply_simpl_body(R,R2,N2,L,L2).
apply_simpl_body([_|_],_,_,_,failing) :-!.
apply_simpl_body([],[],_,L,L).

/*--------
apply_unfold_d_eq(C): unfolds equalities and inequalities in the body of C.
----------*/ 
apply_unfold_d_eq(C) :- clause_number_to_clause(C, HB), apply_unfold_d_eq(C,HB,_).
apply_unfold_d_eq(C, (H,B), X) :- apply_unfold_d_eq(B,B2,0,[],Lit_Numbers),
    (Lit_Numbers==failing -> (delete_clause(C), X=removed(C)); (Lit_Numbers==[] -> X=C ; 
    (add_clauses(NewCN,[(H,B2)],nil), NewCN=[X], 
    update_programs((C, (ud_eq,Lit_Numbers),NewCN),_),
    (derivation_trace_on ->
        new_clause(ud_eq,C,Lit_Numbers,NewCN);true)))).
apply_unfold_d_eq(_,_,_,_,Failing) :- Failing == failing, !.
apply_unfold_d_eq([\+(A=B)|_],_,_,_,failing) :- A==B, !.
apply_unfold_d_eq([\+(A=B)|R],[\+(A=B)|R2],N,L,L2) :- 
    \+ \+ unify_with_occurs_check(A,B), !, N2 is N + 1, apply_unfold_d_eq(R,R2,N2,L,L2).
apply_unfold_d_eq([\+(_=_)|R],R2,N,L,L2) :- N2 is N + 1, apply_unfold_d_eq(R,R2,N2,[N2|L],L2).
apply_unfold_d_eq([(A=B)|_],_,_,_,failing) :- \+ unify_with_occurs_check(A,B), !.
apply_unfold_d_eq([A=B|R],R2,N,L,L2) :-
   (var(A) ; var(B)), !, unify_with_occurs_check(A,B), N2 is N + 1, apply_unfold_d_eq(R,R2,N2,[N2|L],L2).
apply_unfold_d_eq([X=Y|R],R3,N,L,L2) :-
   X =.. [_|LX], Y =.. [_|LY], distribute_equality(LX,LY,LZ,L2), (L2 == failing -> ! ;
   (N2 is N + 1, apply_unfold_d_eq(R,R2,N2,[N2|L],L2), append(LZ,R2,R3))).
apply_unfold_d_eq([G|R],[G|R2], N,L,L2) :- N2 is N + 1, apply_unfold_d_eq(R,R2,N2,L,L2).
apply_unfold_d_eq([],[],_,L,L).
distribute_equality([],[],[],_).
distribute_equality([X|_],[Y|_],[],failing) :- \+ X = Y, !.
distribute_equality([X|LX],[Y|LY],LZ,L2) :-
    (var(X) ; var(Y)), !, unify_with_occurs_check(X,Y), distribute_equality(LX,LY,LZ,L2).
distribute_equality([X|LX],[Y|LY],LZ,L2) :-
     X =.. [F|LX2], Y =.. [F|LY2], append(LX2,LX,LX3),
     append(LY2,LY,LY3), distribute_equality(LX3,LY3,LZ,L2).
distribute_equality([_|LX],[_|LY],LZ,L2) :- distribute_equality(LX,LY,LZ,L2).

