%
% uve.pl
%
% This program implements the procedure for the elimination of unnecessary
% variables described in:
%       Proietti M. and Pettorossi A.: Unfolding-Definition-Folding, in This Order,
%       for Avoiding Unnecessary Variables in Logic Programs.
%       Theoretical Computer Science 142(1):89-124, 1995.
% The sets InDefs and OutDefs occurring in that procedure are represented  
% respectively as a set of facts of the form clause3(N,(A,[A1,...,An]),[def,1]) 
% and as a set of facts of the form clause3(N,(A,[A1,...,An]),[def,0]), in which
% A :- A1,...,An is the clause in C identified by the positive integer N.

:- use_module(library(lists), [ append/3 ]).
:- use_module(library(terms), [ variant/2 ]).

uve :-	 
       check_exists_transf_tree, 
       write('input clause number: '), read(N),
       check_input_clause(N), clause3(N,(H,B),_),
       insert_definition((H,B),B),
       while, nl,
       delete_clause(N),
       write('new_program: '), nl, p.

insert_InDef(C) :- define(C,[1]).

define(D,Aux) :-
       (D=(Head_def :- Body_def) ->
             conj_to_list((Body_def),Body_list);D=(Head_def,Body_list)),
       check_head_variable(Head_def,Body_list),
       functor(Head_def,Functor_Head,Arity),
       check_new_pred(Head_def,Functor_Head,Arity), current_prog(PN,CNs),
       ((label_arc(PN,(def,NewCN),PNSucc), NewCN=[N], clause3(N,(H,B),[def|_]),
       variant((Head_def,Body_list),(H,B))) ->
             (label_node(PNSucc,CNsSucc),
              update_current_prog((PN,CNs),(PNSucc,CNsSucc)));
             (add_clauses(NewCN,[(Head_def,Body_list)],[def|Aux]),
             update_programs((def,NewCN),PNSucc))).

while :- while_test -> true; (body_while, while).
while_test :- indefs(InDefs), InDefs=[].

indefs(InDefs) :- setof(N,indef(N),InDefs) -> true; InDefs=[].

indef(N) :- clause3(N,_,[def,1]).  
           
body_while :- unfolding_step(UD), definition_steps(UD), folding_steps(UD).

unfolding_step(UD) :-
       indefs(InDefs), write('Clause Number in '), nl,
       cc(InDefs), nl, read(Clause_Number), write('Atom Number'), nl,
       read(Atom_Number), remove(Clause_Number),
       unfolding_step(Clause_Number,Atom_Number,UD1),
       not_final_clauses(UD1,UD).

not_final_clauses([NC|NCs],UD) :-
       clause3(NC,C,_), (not(failing(C)), not(body_empty(C))) ->
       (not_final_clauses(NCs,UD1), UD=[NC|UD1]); not_final_clauses(NCs,UD).
not_final_clauses([],[]).

body_empty((_,B)) :- B=[].

definition_steps([UD|UDs]) :- definition_step(UD), definition_steps(UDs).
definition_steps([]).

unfolding_step(Clause_Number,Atom_Number,UD) :- 
%       unfold(Clause_Number,Atom_Number), 
       unfold_d_p(Clause_Number,Atom_Number), 
       current_prog(PN,_), label_arc(PN,(_,_,UD)).

definition_step(E) :-
       clause3(E,(H,B),_), split_in_blocks(B,PartB),
       blocks_with_unvar((H,B),PartB,Blocks),
       (not(Blocks=[])-> insert_definitions((H,B),Blocks);true).

% blocks_with_unvar(C,Blocks,Blocks1): Blocks is a list of sets of atoms
% and Blocks1 is the list of all elements in Blocks having at least an
% unnecessary variable of the clause C. 
% Ex: blocks_with_unvar((p(X,Y),[h(X),r(X,Z),q(Y)]),
% [[h(X),r(X,Z)],[q(Y)]],[[h(X),r(X,Z)]]) is true.

blocks_with_unvar((H,B),[Block|Blocks],Blocks1) :- 
       blocks_with_unvar((H,B),Blocks,Blocks2),
       (un_var((H,Block)) -> Blocks1=[Block|Blocks2]; Blocks1=Blocks2).
blocks_with_unvar((_,_),[],[]).

un_var((H,B),V) :-
       free_variables(B,[],[],VB), free_variables(H,[],[],VH),
       member(V,VB), \+(belongs(V,VH)).
%un_var((_,B),V) :- free_variables(B,[],[],VB), pos1(V,VB,N), pos(V,VB,M), \+(N=M).
% Bug? correction: the above line is replaced by the following
% (my_free_variables does not remove dupplicate elts).
% un_var((_,B),V) :- my_free_variables(B,VB), pos1(V,VB,N), pos(V,VB,M),
% done by Sophie. It does not seem to work. changed in:  (Mau 21/02/2002)
un_var((_,B),V) :- bag_of_vars(B,VB), pos1(V,VB,N),
        pos(V,VB,M), \+(N=M).
un_var((H,B)) :- un_var((H,B),_).



% split_in_blocks(Atoms,Blocks): Blocks is the list of blocks in the 
% set Atoms of atoms.
% Ex: split_in_blocks([r(X),p(X,Y),q(Y),r(Z)],[[r(X),p(X,Y),q(Y)],[r(Z)]])
% is true.
split_in_blocks([A],[[A]]).
split_in_blocks([Atom|Atoms],Blocks) :-
       split_in_blocks(Atoms,Old_Blocks), update_blocks(Atom,Old_Blocks,Blocks).
update_blocks(Atom,Blocks,NewBs) :-
       rels(Atom,Blocks,Blocks1), delete(Blocks1,Blocks,Blocks3),
       merge([[Atom]|Blocks1],Blocks2), NewBs=[Blocks2|Blocks3].
 
rels(Atom,[Block|Blocks],Blocks1) :-
       rels(Atom,Blocks,Blocks2),
       (rel_atom_block(Atom,Block) -> Blocks1=[Block|Blocks2]; Blocks1=Blocks2). 
 
rels(_,[],[]).
 
rel_atom_block(A,As) :- member(B,As), rel(A,B).
 
rel(A1,A2) :- free_variables(A1,[],[],V1), free_variables(A2,[],[],V2),
              member(X,V1), belongs(X,V2).
 
merge([H|[H1|T]],R) :- merge([H1|T],L), append(H,L,R). 
merge([H],H).
 
insert_definitions((H,B),[Block|Blocks]) :-
       (test_def(H,Block) -> insert_definition((H,B),Block);true),
                             insert_definitions((H,B),Blocks).
insert_definitions((_,_),[]).

test_def(H,Block) :- defs(Ds), not_faithful_variants(H,Block,Ds).

defs(Defs) :- setof(D,N^def(N,D),Defs) -> true; Defs=[].
def(N,D) :- clause3(N,D,[def|_]).

not_faithful_variants(H,Block,[(HD,BD)|Ds]) :- 
       not(faithful_variant(((H,Block),Block),((HD,BD),BD))),
       not_faithful_variants(H,Block,Ds).
not_faithful_variants(_,_,[]).

% faithful_variant((C,B),(C1,B1)): the set B of atoms in clause C is a 
% faithful variant of the set B1 of atoms in clause C1.
% Ex. faithful_variant(((p(X),[h(X,Y),r(Z)]),[h(X,Y)]),
% ((q(X,Z),[h(X,Y),r(V)]),[h(X,Y)])) is true.

%faithful_variant(((H1,B1),Block1),((H2,B2),Block2)) :- 
faithful_variant(((H1,_),Block1),((H2,_),Block2)) :- 
       variant_set(Block1,Block2), free_variables(H1,[],[],VH1),
       free_variables(H2,[],[],VH2),
       free_variables(Block1,[],[],VB1), free_variables(Block2,[],[],VB2),
       check_vars((VH1,VB1),(VH2,VB2)).

% variant_set(T,T1): it is equal to variant_set_0.
variant_set(T,T1) :-
                  variant_split(T,VT), variant_split(T1,VT1),
                  copy_term(VT,CT), copy_term(VT1,CT1),
                  variant_sets(CT,CT1).

variant_sets([],[]).
variant_sets([H|T],L) :- find_variant_set(H,L,H1), del_2(H1,L,L1),
                      permutation(H,P), free_variables(P,[],[],V),
                      free_variables(H1,[],[],V1), variant(P,H1), V=V1, 
                      numbervars((V,V1),0,_),
                      variant_sets(T,L1).

find_variant_set([A|As],[H1|T],H2) :- H1=[B|_],
                  variant(A,B) -> H1=H2; find_variant_set([A|As],T,H2).

variant_split([T|Ts],[[T|B]|Bs]) :- find_variant(T,Ts,B), 
                delete_3(B,Ts,Ts1), variant_split(Ts1,Bs).
variant_split([],[]).

find_variant(T,[T1|Ts],R) :- (variant(T,T1) -> R=[T1|LTs];R=LTs),
                  find_variant(T,Ts,LTs).
find_variant(_,[],[]).

check_vars((VH,[V|Vs]),(VH1,[V1|Vs1])) :-
       belongs(V,VH), belongs(V1,VH1), check_vars((VH,Vs),(VH1,Vs1)).

check_vars((VH,[V|Vs]),(VH1,[V1|Vs1])) :-
       \+(belongs(V,VH)), \+(belongs(V1,VH1)), check_vars((VH,Vs),(VH1,Vs1)).

check_vars((_,[]),(_,[])).

insert_definition((H,B),Block) :- 
       link_vars((H,B),Block,Vars), create_new_atom(NewA,Vars),
       insert_InDef((NewA,Block)).
create_new_atom(NewA,Vars) :- create_new_def_symbol(1,NewN), NewA =.. [NewN|Vars].

create_new_def_symbol(N,NewN) :-
    name(N,L), append([110,101,119],L,NewI), name(Pred,NewI),
    (already_used_symbol(Pred) -> 
           N2 is N + 1, create_new_def_symbol(N2,NewN) ; NewN = Pred).

already_used_symbol(Pred) :- 
       definition(_,(Head_Def,_)), functor(Head_Def,Pred,_), !.
%already_used_symbol(Pred) :- clause3(_,(Head,_),prog), functor(Head,Pred,_), !.
already_used_symbol(Pred) :- clause3(_,(Head,_),_), functor(Head,Pred,_), !.

folding_steps([E|Es]) :- (folding_step(E)->true;true), folding_steps(Es).
folding_steps([]).

folding_step(N) :-
       clause3(N,(H,B),_), split_in_blocks(B,PartB),
       foldings(N,(H,B),(H,B),PartB).

foldings(N,(H,B),(H,B1),[Block|Blocks]) :-
       def(M,(HD,BD)), faithful_variant(((H,B),Block),((HD,BD),BD)),
       atoms(B1,Atom_Numbers,Block),
       ts_fold(N,Atom_Numbers,M), current_prog(PN,_),
       label_arc(PN,(_,_,[NewN])),
       clause3(NewN,(H2,B2),_),
       foldings(NewN,(H,B),(H2,B2),Blocks).

foldings(_,(_,_),(_,_),[]).

remove(Clause_Number) :-
       clause3(Clause_Number,C,[def,1]),
       retract(clause3(Clause_Number,C,[def,1])),
       assert(clause3(Clause_Number,C,[def,0])).
                
new_number(N,NN) :- test_number(N) -> (M is N+1, new_number(M,NN)); NN=N.
new_number(N) :- new_number(1,N).

test_number(N) :-
       clause3(_,(H,_),[def|_]), functor(H,F,_), name(N,L),
       append([110,101,119],L,L1), name(F,L1).

% link_vars(C,Atoms,Vs): Vs is the list of linking variables of the set 
% Atoms of atoms in clause C.

link_vars((H,B),Atoms,Vs) :-
       delete_2(Atoms,B,B1),
       (setof(V,link_var(H,Atoms,B1,V),Vs) -> true; Vs=[]).
 
link_var(H,Atoms,B,V) :-
       free_variables(Atoms,[],[],VAs), free_variables((H,B),[],[],VCs),
       member(V,VAs), belongs(V,VCs).

% create_atom(Functor,Args,Atom): Atom is a term having 'Functor' as
% principal functor and the elements in the list Args as arguments.

create_atom(Functor,Args,Atom) :-
       length(Args,L), functor(Atom,Functor,L), insert_args(Args,1,Atom).
 
insert_args([A|As],N,Atom) :-
       insert_arg(A,N,Atom), M is N+1, insert_args(As,M,Atom).
insert_args([],_,_).
insert_arg(A,N,Atom) :- arg(N,Atom,A).

not(X) :- \+ X.

% failing(C): C is a failing clause.            
failing((_,B)) :-
       member(A,B), \+(setof((L,R),(clause3(_,(L,R),prog),A=L),_)).
 
atoms(_,[],[]).
atoms(B,[H|T],[A|L]) :- pos1(A,B,H), atoms(B,T,L).
