% :- module(utilities).%,[def/3,def_star/3,union/3,subtract/3,extract/3,is_member/2,is_not_member/2,member_chk/2,unifiable/2,existential_variable/2,linking_variables/3,test/3]).

%=========================================================
% Utilities

% CLAUSE PRINTING

show_clause_constr(InitPos, (H,bd(C,B))):-
	copy_term((H,bd(C,B)),(H1,bd(C1,B1))),
	numbervars((H1,bd(C1,B1)),0,_),
	writeq(H1), (((C1 == [1=1] ; C1 == []), B1=[]) -> write('.');
		(print_length(write(H1),LH1),LH2 is LH1 + 4 + InitPos,
			write(' :- '), ((C1 == [1=1] ; C1 == []) ->
						show_body(B1,LH2) ;
						print(C1), (B1=[] ->
									write('.');
									write(', '), show_body(B1,LH2)))
	)).

show_clause_constr2(InitPos, (H,bd(C,CEx,B))):-
	copy_term((H,bd(C,CEx,B)),(H1,bd(C1,CEx1,B1))),
	numbervars((H1,bd(C1,CEx1,B1)),0,_),
	writeq(H1), (((C1 == [1=1] ; C1 == []), (CEx1 == [1=1] ; CEx1 == []), B1=[]) -> write('.');
		( print_length(write(H1),LH1),LH2 is LH1 + 4 + InitPos,
		  write(' :- '), print(C1), write(', '), print(CEx1),
		  (B1==[] -> write('.') ; (write(', '), show_body(B1,LH2))) )).

show_all_constr([]).
show_all_constr([D|Ds]) :- show_clause_constr(0,D), nl, show_all_constr(Ds).

show_def((_,[])).
show_def((_,[D|Ds])) :- show_clause_constr(0,D), nl, show_def((_,Ds)).

show_def2((_,[])).
show_def2((_,[D|Ds])) :- show_clause_constr2(0,D), nl, show_def2((_,Ds)).

show_all_defs([]).
show_all_defs([D|Ds]) :- show_def(D), show_all_defs(Ds).


%-------------------------------------------
%

% MODE: def(+Pred,+Prog,-Def)
% TYPE: def(predicate/arity,list(def),def)
def(P,[(P1,Def1)|Defs],Def2) :-
	P=P1 -> (P1,Def1)=Def2 ; def(P,Defs,Def2).

% MODE: delete_def(P,Prog,NewProg)
% TYPE:
% SEMANTICS:
delete_def(_,[],[]).
delete_def(P,[(P,_)|Defs],Defs).
delete_def(P,[(Q,DefQ)|Defs],[(Q,DefQ)|NewDefs]) :-
	P\==Q,
	delete_def(P,Defs,NewDefs).

% MODE: def_star(+P/A,+Prog1,-Prog2)
% TYPE: def_star(predicate/arity,list(def),list(def))
% we suppose Preds has no duplicates
def_star(P/A,Prog,Defs) :-
	defall_star([P/A],Prog,[],Defs).

% MODE: defall_star(+Predicates,Program,Done,ProgOut).
defall_star([],_,_,[]).
defall_star([P/A|Preds],Prog,Done,Defs) :-
	member(P/A,Done) ->
		( defall_star(Preds,Prog,Done,NewDefs),
		  NewDefs=Defs
		)
		;
		( def(P/A,Prog,Def),
		  find_predicates(Def,DefPreds),
		  Defs=[Def|NewDefs],
		  union(DefPreds,Preds,NewPreds),
		  append([P/A],Done,NewDone),
		  defall_star(NewPreds,Prog,NewDone,NewDefs)
		).

find_predicates((_,[]),[]).
find_predicates((_,[(_,bd(_,B))|Cls]),NewPreds) :-
	find_pred_names(B,Ps),
	find_predicates((_,Cls),Preds),
	union(Ps,Preds,NewPreds).

find_pred_names([],[]).
find_pred_names([L|Ls],[P/A|Preds]) :-
	(functor(L,\+,_) -> (L=..[\+,Atom]); (L=Atom)),
	functor(Atom,P,A),
	find_pred_names(Ls,Preds).

% occurs_in_lit(+V,+B,?L) :- L is the leftmost literal of goal B
% such that variable V occurs in L

occurs_in_lit(V,[L|_],L1) :-
        term_variables(L,Vs),
        is_member(V,Vs),
        !,
        L1=L.
occurs_in_lit(V,[_|Ls],L1) :-
        occurs_in_lit(V,Ls,L1).

% MODE: complement_lit(+L,-NotL)
% TYPE: complement_lit(literal,literal)
% SEMANTICS: returns the complement of a literal
complement_lit(L,NotL) :- L=..[\+,NotL].
complement_lit(L,NotL) :- \+ functor(L,(\+),1), NotL = (\+ L).

% MODE: negative_literal(+Literal)
% TYPE: negative_literal(+literal)
% SEMANTICS: tests whether A is a negated literal or not
negative_literal(\+ _).	%  :- functor(L,P,A), P==(\+).

lit(L) :- atomic_formula(L), !.
lit(\+ L) :- atomic_formula(L).

% MODE: remove_nat_predicates(+List,-NewList).
% TYPE: remove_nat_predicates(list(literal),list(literal)).
% SEMANTICS:
remove_nat_predicates([],[]).
remove_nat_predicates([nat(_)|Ls],NewLs) :- remove_nat_predicates(Ls,NewLs).
remove_nat_predicates([L|Ls],[L|NewLs]) :- L\=nat(_), remove_nat_predicates(Ls,NewLs).

% pred(+L,?P/A) :-
% literal L has predicate symbol P and with arity A

pred(L,P/A) :-
        atomic_formula(L),
        !,
        functor(L,P,A).
pred(L,P/A) :-
        L=(\+ B),
        functor(B,P,A).


%-------------------------------------------
% CONSTRAINT UTILITIES

% A and B
conjunction(A,B,C) :- append(A,B,C).

% MODE: is_truth_value(+ConstraintList,+Vars)
% TYPE: is_truth_value(list,list)
% SEMANTICS: test whether C evaluates to true
is_truth_value(L,Vars) :-
	copy_term((L,Vars),(L1,Vars1)),
	list_to_constraint(clpr,L1,C),
	C,
	dump(Vars1,_,[]).

% test whether C is satisfiable
% satisfiable(C) :-
% 	copy_term(C,Clist),
% 	list_to_constraint(clpr,Clist,Cons), Cons.

% MODE: complement_c(+C1,-C2)
% TYPE: complement_c(constraint,constraint)
% SEMANTICS: C2 is the negation of constraint C1
complement_c([N1=<N2],[N1>N2]) :- !.
complement_c([N1>=N2],[N1<N2]) :- !.
complement_c([N1>N2],[N1=<N2]) :- !.
complement_c([N1<N2],[N1>=N2]) :- !.
%complement_c([C1],[C2]) :- C1=.. [=:=,N1,N2], C2=.. [=\=,N1,N2].
complement_c([C1],[C2]) :- C1=.. [=\=,N1,N2], C2=.. [=:=,N1,N2].

% MODE: complement_eq(+C,-C1,-C2)
% TYPE: complement_eq(constraint,constraint,constraint)
% SEMANTICS: if C is an equality, the negation of C is a
% pair of constraints C1 and C2
complement_eq([C],[LHS>RHS],[LHS<RHS]) :-
	(functor(C,=,_);functor(C,=:=,_)),
	C=..[_,LHS,RHS].

is_equality([_=_]).
is_equality([_=:=_]).

%-------------------------------------------
% TERM UTILITIES

% no dot distinguish variable names and do not propagate unification
member_unifiable(A,[V|_]) :- copy_term((A,V),(A1,V1)), A1=V1.
member_unifiable(A,[_|Vs]) :- is_member(A,Vs).

% linking_variables(+Term1,+Term2,-Vars)
linking_variables(Term1,Term2,Vars) :-
	term_variables(Term1,Tv1),
	term_variables(Term2,Tv2),
	common_vars2(Tv1,Tv2,Vars).

% MODE: common_vars(+Vars1,+Vars2,-Vars)
% SEMANTICS: no variable is bound
common_vars2([],_,[]).
common_vars2([H|R],L,[H|R1]) :- is_member(H,L), common_vars2(R,L,R1).
common_vars2([H|R],L,R1) :- is_not_member(H,L), common_vars2(R,L,R1).
% 
test([V|VarsMore],Vars,Var) :-
	is_member(V,Vars),
	test(VarsMore,Vars,Var).
test([V|_],Vars,Var) :-
	is_not_member(V,Vars), Var=V.

%-------------------------------------------
% CLAUSE UTILITIES

% TYPE: find_unifiable_clause(atom,def,clause)
% MODE: find_unifiable_clause(+A,+Def,-Clause)
% SEMANTICS: find a clause in the definition whose head is unifiable with the given atom
find_unifiable_clause(A,(_,[(Head,bd(C,B))|_]),(Head,bd(C,B))) :- unifiable(A,Head).
find_unifiable_clause(A,(P,[_|Cls]),Clause) :- find_unifiable_clause(A,(P,Cls),Clause).


% SEMANTICS: Change body representation from (H,B) to (H,bd(C,L))
%            by separing the constraints C from the other literals L
split_bd([],bd([],[])).
split_bd([L|Ls],bd(Cns,Lits)) :-
	is_a_constraint(L) ->	
		(Cns=[L|Cns1], split_bd(Ls,bd(Cns1,Lits)));
		(Lits=[L|Lits1], split_bd(Ls,bd(Cns,Lits1))).

is_a_constraint(Cn) :-
	functor(Cn,F,_),
	(F=(=:=); F=(=); F=(>=); F=(=<); F=(>); F=(<); F=(=\=)).

%-------------------------------------------
% SET OPERATIONS

% A,B, and C are sets union(A,B,C) holds iff A U B = C
union([], A, A) :- !.
union([A|B], C, D) :-
        is_member(A, C), !,
        union(B, C, D).
union([A|B], C, [A|D]) :-
        union(B, C, D).

%% removed becouse already in "../basic_op.pl"
%% % A,B, and C are sets subtract(A,B,C) holds iff A - B = C
subtract2([], _, []).
subtract2([Element|Elements], Set, Dif) :-
	is_member(Element, Set), !, subtract2(Elements, Set, Dif).
subtract2([Element|Elements], Set, [Element|Dif]) :- subtract2(Elements, Set, Dif).

% A,B, and C are sets extract(A,B,C) holds iff A - {B} = C
% MODE: extract(+Set,-Element,-NewSet)
wrong_extract([A|As],A,As).%,0).
% extract([A|As],B,[A|Bs],M) :- M is N+1, extract(As,B,Bs,N).
wrong_extract([A|As],B,[A|Bs]) :- wrong_extract(As,B,Bs).

% MODE: extract(+List,-BeforeList,-Element,-AfterList)
% SEMANTICS: A,B and D are lists and C is an element of the list A
%   extract(A,B,C,D) holds iff A = B;[C];D where ; is the list
%   concatenation operator
extract([El|After],El,[],After).
extract([A|As],El,[A|Before],After) :- extract(As,El,Before,After).

% distinguish variable names (do not unify)
% is_member(+X,+Ys) :- there is a member Y of the list of terms Ys
% (possibly with variables) such that X==Y. No variable is bound.
is_member(A,[V|_]) :- A==V,!.
is_member(A,[V|Vs]) :- A\==V,
 		is_member(A,Vs).

v_sort([],[],[]).
v_sort(G,[L|Ls],[L1|Ls1]) :-
        del_elem(L1,G,G1),
        variant(L,L1),
        v_sort(G1,Ls,Ls1).


del_elem(X,[X|Xs],Xs).
del_elem(X,[Y|Xs],[Y|Ys]) :- del_elem(X,Xs,Ys).

% distinguish variable names (do not unify)
is_not_member(_,[]).
is_not_member(X,[A|L]):- X\==A, is_not_member(X,L).

% extract a sublist preserving ordering
sublist_noperm([],[]).
sublist_noperm([L|B],[L|SubBody]) :- sublist_noperm(B,SubBody).
sublist_noperm([_|B],SubBody) :- sublist_noperm(B,SubBody).


